""" COStar09.py
    based on COSTAR.py, which in turn
    adapts the TStar09.py module for collaborative problem solving.

    Last updated 22 February 2009.  S. Tanimoto

    Status of COSTAR features:
 * notion of component state spaces.
    This required adding a method for "fixing" the operators of each
    component state to work with their parts of the cartesian product state.
    Some functional programming saved the day here.    
    IMPLEMENTED
    
 * generic display method that aggregates component displays.
    This required re-organizing the painting methods in each component's
    Node class and/or State class to accept a desc. of a rectangle (center x
    and y, and width and height) and paint on the main canvas.    
    WORKING
    
    DONE: Fixed the desc field of each state so that it is a LIST of desc fields,
       then modified the postcondition function fixups to use these.
       If preconditions ever use the desc field of states, then
       then they will need to be adjusted to correctly access the correct
       component of the description.
    
 * controls for roles.
         Role = 0, 1, 2, ..., n
         n means all, where n = number of components
      DONE
      
 * communication layer.
      DONE

 * adjustment to the form of Annotation files, to facilitate merging of
     trees.
      DONE
      
 * representation for annotations (including free text and numeric),
    with each annotation having an association either with a particular state component
    or with the entire state.
      The representation should probably be a hash, mapping tuples (State-Path-Desc, Component No)
      to string values of the form:
      "PROPERTY-NAME: PROPERTY-VALUE (which can represent numbers, of course); AUTH: The Author; TIME: Date/Time"

 * display of annotations.
     OLD ANNOTATATIONS are shown.
     
 * More sophisticated controls for annotation
 (est time: 1 day)
     """

USE_ANNOTATIONS = True

import TStar09 as TStar
import Tkinter

INITIAL_DATA = None

def register_t_star_modifications():
  """This method sets various symbols within the TStar module
     to refer to their definitions related to the image processing
     application in this file. These definitions and given below
     this method."""
  TStar.__dict__.update({
    'TITLE':" CO-STAR, Version 0.9",
    'NODE_CLASS': COStar_Node,
    'create_initial_state': create_initial_costar_state,
    #'get_full_root_data': get_full_root_data,
    #'display_data_in_detail': display_data_in_detail,
    'add_app_specific_menus':add_collab_menu,
    'app_specific_tree_save':save_annotations,
    'app_specific_tree_load':load_annotations,
    #'load':load_img,
    'HEADLINE':"###COStar Version 0.9 Tree Description File",
    'TREE_FILE_TYPE':"mmt",
    'ROOT_DATA_FILE_TYPE':"jpg,gif,PNG",
    'ROOT_DATA_TYPE_DESCRIPTION':"Image files"
    })
  COStar_Node.one_time_setup()

NUM_SS_COMPONENTS = 0    
COMPONENT_NAMES = []
STATE_CLASSES = []
NODE_CLASSES = []
OP_SETS = []
STATE_DATA_INITIALIZERS = []
STATE_DESC_INITIALIZERS = []
ROLE = 0

def initialize():
  global NUM_SS_COMPONENTS, COMPONENT_NAMES, STATE_CLASSES, NODE_CLASSES
  global OP_SETS, STATE_DATA_INITIALIZERS, STATE_DESC_INITIALIZERS, ROLE
  NUM_SS_COMPONENTS = 0    
  COMPONENT_NAMES = []
  STATE_CLASSES = []
  NODE_CLASSES = []
  OP_SETS = []
  STATE_DATA_INITIALIZERS = []
  STATE_DESC_INITIALIZERS = []
  ROLE = 0
  

class COStarObject(TStar.State):
    def __init__(self):
        global NUM_SS_COMPONENTS, STATE_DATA_INITIALIZERS, STATE_DESC_INITIALIZERS
        TStar.State.__init__(self,[],None,None,desc=[])
        #self.data=[]
        n = NUM_SS_COMPONENTS
        if n>0:
          for i in range(n):
            #dummyInst = STATE_CLASSES[i](None,None,None,None)
            self.data.append(STATE_DATA_INITIALIZERS[i]())
            self.desc.append(STATE_DESC_INITIALIZERS[i])
        #print "In COStarObject.__init__: new object is: " + str(self.data)

    def copy(self):
        nmo = COStarObject()
        for i in range(NUM_SS_COMPONENTS):
          nmo.data[i] = self.data[i].copy()
          nmo.desc[i] = self.desc[i]
        return nmo
    def str(self):
        s = "COStarObject("
        for i in range(NUM_SS_COMPONENTS):
          s += str(self.data[i])
        s += ")"
        return s
    def initial_data(self):
        global NUM_SS_COMPONENTS, STATE_CLASSES
        nco = COStarObject()
        for i in range(NUM_SS_COMPONENTS):
          dummyInst = STATE_CLASSES[i](None,None,None,None)
          nco.data[i] = dummyInst.initial_data()
        return nco.data
        
ICONS = {}; ICONS05 = {}; ICONS03 = {}
ANICONS = {}; ANICONS05 = {}; ANICONS03 = {}
USER_ICONS = {}
ICON_NAMES = []
testkimage =0

import Image as PIL
import ImageTk

def load_annotation_icons():
    global ICON_NAMES
    ICON_NAMES = ['author','comment', 'lock', 'authors', 'weight']
    #testimage = PIL.open('COSTAR/author32.PNG')
    #global testtkimage
    #testtkimage = ImageTk.PhotoImage(testimage);
    global ICONS, ANICONS, ICONS05, ANICONS05
    for name in ICON_NAMES:
        ICONS[name] = PIL.open('COSTAR/'+name+'32.PNG');
        ANICONS[name] = ImageTk.PhotoImage(ICONS[name])
        ICONS05[name] = ICONS[name].copy();
        ICONS05[name].thumbnail((16,16))
        ANICONS05[name] = ImageTk.PhotoImage(ICONS05[name])
        ICONS03[name] = ICONS[name].copy();
        ICONS03[name].thumbnail((8,8))
        ANICONS03[name] = ImageTk.PhotoImage(ICONS03[name])

def get_user_icon(name, z=1.0):
    global USER_ICONS, ICONS, USER_ICONS05, ICONS05
    if (name==None or name==""): return None
    #print "Trying to load icon " + name
    #print str(USER_ICONS[name])
    if USER_ICONS.has_key(name):
        #print "Image " + name + " already loaded."
        if z==1.0: return USER_ICONS[name]
        if z==0.5: return USER_ICONS05[name]
    else:
        try: 
            #print "Trying to load " + name
            ICONS[name] = PIL.open('COSTAR/'+name)
            ICONS[name].thumbnail((32,32))
            #print "Loaded the image " + name
            USER_ICONS[name] = ImageTk.PhotoImage(ICONS[name])
            ICONS05[name] = ICONS[name].copy()
            ICONS05[name].thumbnail((16,16))
            USER_ICONS05[name] = ImageTk.PhotoImage(ICONS05[name])
                            
        except:
            print "Could not find the image: " + name
            return None
        if z==1.0: return USER_ICONS[name]
        if z==0.5: return USER_ICONS05[name]
    
class Annotation:
    count = 0
    def __init__(self,a_type):
        Annotation.count += 1
        self.type = a_type;
        self.id = Annotation.count
        self.val = 0
        if a_type=='author':
          self.text = TStar.USER
        else:
          self.text = "Test annotation"
        self.icon = ""
    def to_loadable_string(self):
        return "a=Annotation('"+self.type+"'); a.id="+str(self.id)+\
               "; a.val="+str(self.val)+"; a.text='"+self.text+"'; a.icon='"+self.icon+"';"
    
        
class Annotated_Node(TStar.Node):
    iw = 32
    bound = False
    def __init__(self, state):
        TStar.Node.__init__(self, state)
        self.annotations = [];
        # Here are 2 default annotations on new nodes, for demo purposes:
        # Note: Don't include author annotation automatically for the root.
        if self.s.p: 
          self.annotations.append(Annotation('author'))
        #self.annotations.append(Annotation('comment'))
        #TStar.SP_CANVAS.bind("<Button-1>", annotations_left_click)
        #if not Annotated_Node.bound:
        #    #print "Binding in Annotated_Node"
        #    TStar.SP_CANVAS.bind("<Button-1>", annotations_left_click)
        #    Annotated_Node.bound = True
        
    def paint_content_in_rectangle(self,a_canvas,x,y,w,h):
        if SHOWING_ANNOTATIONS:
          z = TStar.ZOOM_FACTOR
          if z==1.0: iw = 32
          elif z==0.5: iw = 16
          elif z==0.2: iw = 6.4
          else: iw = z*32
          #iw = z*Annotated_Node.iw #32 # icon width
          self.x0=z*self.x-w/2-1
          self.y0=z*self.y+h/2
          self.x1=z*self.x+w/2+1
          self.y1=z*self.y+h/2+iw+4
          a_canvas.create_rectangle(self.x0, self.y0, self.x1, self.y1, fill="black")
          #print "a_canvas = " + str(a_canvas)
          rx = 0
          for a in self.annotations:
              if z==1.5: theImage = ANICONS[a.type]
              if z==1.0: theImage = ANICONS[a.type]
              if z==0.5: theImage = ANICONS05[a.type]
              if z==0.2: theImage = ANICONS03[a.type] 
              if z==0.1: theImage = ANICONS03[a.type] # This should get changed to something the right scale.
              a_canvas.create_image(z*self.x-w/2+iw/2+1+rx , z*self.y+h/2+iw/2+2, image = theImage)
              rx += (iw + 1)
              name = a.icon
              if len(name)>0:
                  try:
                      icon = get_user_icon(name)
                      a_canvas.create_image(z*(self.x-w/2+iw/2+1+rx) , z*(self.y+h/2+iw/2+2), image = icon)
                      rx += (iw + 1)
                  except: print "Could not get icon from file: " + name

    def clicked_on_annotation(self,x,y):
        #def clicked_on(self, x,y):
        if self.s.hidden: return False
        if self!=TStar.TREE_DISP.focus_node: return
        #sn = self.s.space_needed()
        #z = TStar.ZOOM_FACTOR
        #sf = self.s.scale
        #w = self.width*sn*z*sf; h = self.height*sn*z*sf
        #return (self.x - w/2 < x < self.x + w/2 and
        #        self.y + h/2 < y < self.y + h/2 + Annotated_Node.iw)
        return (self.x0 < x < self.x1 and
                self.y0 < y < self.y1)
        
    def annotations_to_loadable_str(self):
        s = ""
        for a in self.annotations:
            s += a.to_loadable_string(); s += "\n"
            #s += "State_num_to_node_hash["+str(self.s.n)+"].annotations.append(a) \n"
            s += "affix(a, "+str(self.s.n)+")\n"
        return s;
    def write_annotations(self):
        s = self.annotations_to_loadable_str()
        global ANNOTATIONS_FILE
        ANNOTATIONS_FILE.write(s)
      
def annotations_left_click(event):
  #def left_click(event):
  """If the user clicked on a node,
     make it the selected node, and perform the
     current command on it, if legal."""
  x = TStar.SP_CANVAS.canvasx(event.x)
  y = TStar.SP_CANVAS.canvasy(event.y)
  for nd in TStar.TREE_DISP.all_nodes:
      if nd.clicked_on_annotation(x,y):
          print "Clicked on annotation area"
          TStar.update_node_focus(nd)
          edit_annotations()
          return
  #TStar.left_click(event)
    
def rev(lst):
  # Return a copy of a list, in reverse order."
  lst2 = lst[:]
  lst2.reverse()
  return lst2
    
import tkSimpleDialog
from Tkinter import Label, Entry, Listbox, Checkbutton, Canvas, END, IntVar
class EADialog(tkSimpleDialog.Dialog):
        
    def body(self, master):
        self.title("EDIT ANNOTATIONS")
        i = 1; iw = 32
        theNode = TStar.TREE_DISP.focus_node
        n = len(theNode.annotations)
        canvases = range(n)
        Label(master, text="type").grid(row=0,column=0)
        Label(master, text="id").grid(row=0,column=1)
        Label(master, text="value").grid(row=0,column=2)
        Label(master, text="description").grid(row=0,column=3)
        Label(master, text="user icon").grid(row=0,column=4)
        Label(master, text="Delete?").grid(row=0,column=5)
        self.vals = range(n)
        self.descs = range(n)
        self.deletes = range(n)
        self.usericon = range(n)
        for a in theNode.annotations:
            #print 'a.type is '+ a.type
            canvases[i-1] = Canvas(master, width=iw, height=iw)
            canvases[i-1].grid(row=i, column=0)
            canvases[i-1].create_image(iw/2+1 , iw/2+2, image = ANICONS[a.type])
            Label(master, text=str(a.id)).grid(row=i, column=1)
            e1=Entry(master, width=3)
            e1.grid(row=i, column=2)
            e1.insert(END,str(a.val))
            self.vals[i-1]=e1
            e2=Entry(master, width=30)
            e2.grid(row=i, column=3)
            e2.insert(END,a.text)
            self.descs[i-1]=e2
            e3=Entry(master, width=15)
            e3.grid(row=i, column=4)
            e3.insert(END,a.icon)
            self.usericon[i-1]=e3
            # Add checkbox here for deleting annotations:
            v = IntVar()
            c=Checkbutton(master, variable=v)
            c.grid(row=i,column=5)
            c.var = v
            self.deletes[i-1]=c
            i += 1            
        Label(master, text="Add new annotation:").grid(row=i, column=0, columnspan=3)
        Label(master, text="Add?").grid(row=i, column=5)
        ch = Listbox(master, height=5, exportselection=0)
        ch.grid(row=i+1, column=0)
        j=0
        for n in ICON_NAMES:
            ch.insert(j, ICON_NAMES[j])
            j += 1
        aae1 = Entry(master, width=3, text="5")
        aae1.grid(row=i+1, column=2)
        aae2 = Entry(master, width=30, text="description")
        aae2.grid(row=i+1, column=3)
        aae3 = Entry(master, width=15, text="user icon")
        aae3.grid(row=i+1, column=4)
        v = IntVar()
        aac1 = Checkbutton(master, variable=v)
        aac1.grid(row=i+1,column=5)
        aac1.var = v
        self.newtype = ch
        self.newval = aae1
        self.newdesc = aae2
        self.newusericon = aae3
        self.addButton=aac1
        if (n > 0):
            ai = theNode.annotations[0].icon
            if (not ai == None):
                top = show_large_user_image(master, ai)

    def apply(self):
        theNode = TStar.TREE_DISP.focus_node
        n = len(theNode.annotations)
        addsOrDeletesMade = False
        for i in rev(range(n)):
            #print "The variable value is "
            #print str((self.deletes[i]).var.get())
            # Delete the annotation if the delete box is checked:
            if self.deletes[i].var.get()==1:
                theNode.annotations[i:i+1]=[]
                addsOrDeletesMade = True
            else:
                # Update the annotation
                an = theNode.annotations[i]
                an.val = int(self.vals[i].get())
                an.text = self.descs[i].get()
                an.icon = self.usericon[i].get()
        # Add a new annotation, if the add box is checked
        needToAddNewAnnotation = self.addButton.var.get()==1
        if needToAddNewAnnotation:
          theSel = self.newtype.curselection()
          if len(theSel)>0:
            #print "theSel = " + str(theSel)
            theType = ICON_NAMES[int(theSel[0])]
          else:
            theType = "comment"
          an = Annotation(theType)
          try:
            an.val = int(self.newval.get())
          except: pass
          an.text = self.newdesc.get()
          an.icon = self.newusericon.get()
          theNode.annotations.append(an)
          addsOrDeletesMade = True
        TStar.redraw_tree()

tkim = None
def show_large_user_image(master, icon_fname):
    print "Icon f_name is " + icon_fname
    import re
    big_file = re.sub("\.jpg$", "_LRG.jpg", icon_fname)
    print "Trying to load " + big_file
    if big_file == '': return
    try:
      im = PIL.open('COSTAR/'+big_file)
      (w,h) = im.size
      sc = 5
      im2 = im.resize((w*sc, h*sc))
    except:
      print "Could not open file."
      return
    global tkim
    tkim = ImageTk.PhotoImage(im2)
    top = Toplevel(TStar.TK_ROOT)
    top.grid_rowconfigure(0, weight=1)
    top.grid_columnconfigure(0, weight=1)
    (cw, ch) = (800,700)
    can = Canvas(top, width = cw, height=ch)
    can.grid(row=0, column=0)
    can.create_image(cw/2, ch/2, image=tkim)
    top.wm_title(big_file)
    return top

def edit_annotations():
    """Bring up a window with the annotation editor"""
    TStar.log("edit_annotations")
    TStar.set_last_command(edit_annotations)
    theNode = TStar.TREE_DISP.focus_node
    #print "The selected state is: " + theNode.s.__str__()
    #try:
    eaframe = EADialog(TStar.TK_ROOT, theNode);
    #except Exception, inst:
    # print "caught exception from dialog box." + str(inst)
    
popup_option_added = False
def create_initial_costar_state():
    global popup_option_added
    if not popup_option_added:
      TStar.COMMANDS_POPUP.add_command(label="Edit Annotations...", command=edit_annotations)
      popup_option_added = True
    load_annotation_icons()
    global INITIAL_DATA
    INITIAL_DATA = COStarObject()
    #im = PIL.open(FILE_NAME)
    #im.thumbnail((IMAGE_WIDTH,IMAGE_HEIGHT))
    #INITIAL_DATA.image = im
    TStar.State.number = 0
    TStar.INITIAL_STATE = COStarObject()

# special menu for collaboration features:
from Tkinter import Menu
def add_collab_menu(menubar):
  collabMenu = Menu(menubar)
  menubar.add_cascade(label="Collaboration", menu=collabMenu)
  collabMenu.add_command(label="Show annotations", command=show_annotations)
  collabMenu.add_command(label="Hide annotations", command=hide_annotations)
  collabMenu.add_command(label="Configure Tree View", command=config_tree_view)
  #add_play_menu(menubar) # Include the music-playing menu
  create_role_menu(menubar)
  make_more_app_menus(menubar)

def make_more_app_menus(menubar): pass # Can be replaced by real app. program.
  
SHOWING_ANNOTATIONS=True

def show_annotations():
    TStar.log("show_annotations")
    global SHOWING_ANNOTATIONS
    SHOWING_ANNOTATIONS = True
    TStar.redraw_tree()
def hide_annotations():
    TStar.log("hide_annotations")
    global SHOWING_ANNOTATIONS
    SHOWING_ANNOTATIONS = False
    TStar.redraw_tree()
def config_tree_view():
    TStar.log("config_tree_view")
    ViewConfigurationFrame(TStar.TK_ROOT)
    #showing_annotations = True  # This being lower case is Probably an error
    #SHOWING_ANNOTATIONS = True

from Tkinter import Toplevel, Button    
class ViewConfigurationFrame:
    def __init__(self, parent):
        self.top = Toplevel(parent)
        top = self.top
        top.title("Configure the Tree View's dependence on Annotations")
        Label(top, text="Select nodes where").grid(row=0,column=0,columnspan=3)
        ac = Listbox(top,width=10,height=6, selectmode='MULTIPLE', exportselection=0)
        ac.grid(row=1,column=0)
        self.attrib = ac
        self.annotationChoices = ['author','authors','value','comment','lock','any']
        for item in self.annotationChoices:
            ac.insert(END, item)
        oc = Listbox(top,width=8,height=5, exportselection=0)
        oc.grid(row=1,column=1)
        self.op = oc
        self.options = ['exists','=','<','>','contains']
        for item in self.options:
            oc.insert(END, item)

        self.testValEntry = Entry(top)
        self.testValEntry.grid(row=1,column=2)
        b1 = Button(top, text="Update Tree", command=self.update)
        b1.grid(row=2, column=0, columnspan=2)
        b2 = Button(top, text="Close", command=self.ok)
        b2.grid(row=2, column=2)
        
    def update(self):
        print "We should update the tree now"
        nodes = TStar.TREE_DISP.all_nodes
        for n in nodes:
            if self.filter(n):
                n.s.display_version = 2 # normal
            else:
                n.s.display_version = 0 # minify
        TStar.redraw_tree()

    def filter(self, annotatedNode):
        """Takes an instance of Annotated_Node and returns True or False,
        depending on whether the node satisfies the configured view condition."""
        if (annotatedNode == None): return False
        annotations = annotatedNode.annotations
        op = self.op.curselection()
        if (op==None): return False
        operator = self.options[int(op[0])]
        print "operator is " + operator
        subj = self.attrib.curselection()
        print "subj is " + str(subj)
        if (subj==None): return False
        try:
         import string
         subjectNumList = map(string.atoi, subj)
        except ValueError: pass
        subjectList = map(lambda i,d=self.annotationChoices: d[i], subjectNumList)
        print "subjectList is " + str(subjectList)
        cosubject = self.testValEntry.get()
        print "cosubject is " + cosubject
        if (operator == 'exists'):
            if 'any' in subjectList:
                if len(annotations) > 0: return True
            for a in annotations:
                if a.type in subjectList: return True
            return False
        elif (operator == 'contains'):
            for a in annotations:
                if ('any' in subjectList) or (a.type in subjectList):
                    if a.text.__contains__(cosubject): return True
            return False
        else:
            print "We should test the values here, I guess"
            try: val2=int(cosubject)
            except ValueError: val2 = 0
            for a in annotations:
                if ('any' in subjectList) or (a.type in subjectList):
                    if (operator == '='):
                        if a.val == val2: return True
                    elif (operator == '<'):
                        if a.val < val2: return True
                    elif (operator == '>'):
                        if a.val > val2: return True
            return False
            
    def ok(self):
        print "Quitting the View Configuration Frame"
        self.top.destroy()
    def test(self):
        theNode = TStar.TREE_DISP.focus_node
        print str(self.filter(theNode))

ANNOTATIONS_FILE = None
def save_annotations(filename):
    """Create a file that can be read and each line evaluated to restore all
    the annotations in a tree. Modify filename to have extension .ann"""
    import re
    ann_file = re.sub("\.[^\.]*$", ".ann", filename)
    global ANNOTATIONS_FILE
    print "Opening file: " + ann_file
    ANNOTATIONS_FILE = open(ann_file, "w")
    # Walk the tree, and at each node, execute a save_node_annotations function
    for s in TStar.ALL_STATES:
        thisNode = TStar.TREE_DISP.state_node_hash[s]
        thisNode.write_annotations()
    ANNOTATIONS_FILE.close()

State_num_to_node_hash = {}
def load_annotations(filename, desc_string=None):
  if not desc_string:
    import re
    ann_file = re.sub("\.[^\.]*$", ".ann", filename)
    global ANNOTATIONS_FILE
    print "Opening file: " + ann_file
    try: 
        ANNOTATIONS_FILE = open(ann_file, "r")
    except:
        print "Could not read annotations file " + ann_file
        return
    # construct a hash table mapping state numbers to nodes.
    # Also, delete any default annotations.
  else:
    ANNOTATIONS_FILE = desc_string.split("\n")
  A = TStar.TREE_DISP.all_nodes
  global State_num_to_node_hash
  State_num_to_node_hash = {}
  for aNode in A:
      State_num_to_node_hash[aNode.s.n] = aNode
      if TStar.is_new_state(aNode.s): aNode.annotations = []
  # Now process each line of the file and evaluate it.
  for line in ANNOTATIONS_FILE:
    exec(line)
  TStar.redraw_tree()

def affix(annot, old_state_num):
  global State_num_to_node_hash
  if TStar.OLD_STATE_NEW_STATE_MAPPING!={}:
    try:
      #print "Mapping old state number "+str(old_state_num)
      new_state_num = TStar.OLD_STATE_NEW_STATE_MAPPING[old_state_num]
    except:
      new_state_num = old_state_num
  else:
    new_state_num = old_state_num
  try: 
    State_num_to_node_hash[new_state_num].annotations.append(annot)
  except:
    print "In COStar.affix, there was a problem (probably keyError) while"
    print "trying to affix annotations to state "+str(new_state_num)
    
class COStar_Node(Annotated_Node):
    #width = 320; height = 200  # Override dimensions of TStar.Node
    width = 400; height = 400
    #IMAGE_WIDTH; IMAGE_HEIGHT # image size
    def one_time_setup():
        TStar.State.tiny_size_factor = TStar.TINY_SIZE/(0.0+max(COStar_Node.width, COStar_Node.height))
    one_time_setup = staticmethod(one_time_setup)

                                        

    # The following 3 methods override the parent methods to
    # change the application domain to image processing.
    def domain_specific_init(self, state):
        """Performs a conversion of an image to the PhotoImage
           form, which can be used on a canvas within Tkinter."""
        self.display_elements = []

    def paint_content_in_rectangle(self,a_canvas,x,y,w,h):
        """Adds the node's displayable image to the canvas
           at the right place."""
        # Purple background rectangle for testing the geometry
        #marg = 5
        #a_canvas.create_rectangle(x-w/2-marg, y-h/2-marg, x+w/2+marg, y+h/2+marg, fill="purple")
        if USE_ANNOTATIONS:
          Annotated_Node.paint_content_in_rectangle(self,a_canvas,x,y,w,h)
        n = NUM_SS_COMPONENTS
        dummyNodes = []
        if n==0: return
        #print "data for this node is: "+str(self.s.data)
        if ROLE<n: # If not in ALL_ROLES mode, show only one substate.
          state_component = STATE_CLASSES[ROLE](self.s.data[ROLE], None, None)
          #print "data component "+str(i)+" is "+str(state_component)
          node = NODE_CLASSES[ROLE](state_component)
          node.paint_content_in_rectangle(a_canvas,x,y,w,h)
          return
        # Show all substates in a nice layout that depends on how many there are:
        for i in range(n): 
          state_component = STATE_CLASSES[i](self.s.data[i], None, None)
          #print "data component "+str(i)+" is "+str(state_component)
          node = NODE_CLASSES[i](state_component)
          dummyNodes.append(node)
          #print "This node is "+str(node)
        if n==1:
          dummyNodes[0].paint_content_in_rectangle(a_canvas,x,y,w,h)
          return
        if n==2:
          dummyNodes[0].paint_content_in_rectangle(a_canvas,x-w/4,y,w/2,h/2)
          dummyNodes[1].paint_content_in_rectangle(a_canvas,x+w/4,y,w/2,h/2)
          return
        if n==3:
          dummyNodes[0].paint_content_in_rectangle(a_canvas,x-w/3,y,w/3,h/3)
          dummyNodes[1].paint_content_in_rectangle(a_canvas,x,    y,w/3,h/3)
          dummyNodes[2].paint_content_in_rectangle(a_canvas,x+w/3,y,w/3,h/3)
          return
        if n==4:
          dummyNodes[0].paint_content_in_rectangle(a_canvas,x-w/4,y-h/4,w/2,h/2)
          dummyNodes[1].paint_content_in_rectangle(a_canvas,x+w/4,y-h/4,w/2,h/2)
          dummyNodes[2].paint_content_in_rectangle(a_canvas,x-w/4,y+h/4,w/2,h/2)
          dummyNodes[3].paint_content_in_rectangle(a_canvas,x+w/4,y+h/4,w/2,h/2)
          
    def unpaint_content(self,a_canvas):
        """removes the node's display elements from the canvas."""
        if self.has_painted_content:
            for e in self.display_elements:
              a_canvas.delete(e)
            self.has_painted_content=False
        return

def embed_function(f, multiObj_component):
  def newf(multiObj, param_string):
      old_component = multiObj.__dict__[multiObj_component]
      print "In embedded function, the old component is ", old_component
      new_component = f(old_component)
      new_mo = multiObj.copy()
      new_mo.__dict__.update({multiObj_component:new_component})
      print "the new multiObject is ", new_mo
      return new_mo
  return newf

#class COStarOperator(TStar.Operator):
  # Extension provides a new "component number" field, filled in at
  # binding time (between defining ops and starting the run).

class make_state_data_reader:
  def __init__(self, which):
    self.which_comp = which
  def __call__(self, d):
    return d[self.which_comp]

class make_state_data_writer:
  def __init__(self, which):
    self.which_comp = which
  def __call__(self, d, v):
    dc = costar_deep_copy(d)
    dc[self.which_comp] = v
    return dc

def costar_deep_copy(lst):
  lst2 = []
  for elt in lst:
    lst2.append(copy(elt))
  return lst2

def copy(elt):
  if type(elt)==type(0): return elt
  if type(elt)==type([]): return elt[:]
  else: return elt
    
class make_new_precond:
  def __init__(self, oldop, which):
    self.which_comp = which
    self.oldPre = oldop.precondition
    self.group = oldop.group
  def __call__(self, s):
    #print "in precondition, state is "+str(s)
    #print "  and full data is "+str(s.data)
    #print "  and component "+str(self.which_comp)+" of data is "+str(s.data[self.which_comp])
    val = self.oldPre(s.data[self.which_comp])
    #print "Precondition evaluates to "+str(val)+" before considering ROLE"
    global ROLE, NUM_SS_COMPONENTS, COMPONENT_NAMES
    if ROLE!=NUM_SS_COMPONENTS:
        val = val and self.group==COMPONENT_NAMES[ROLE]
    return val
  
class make_new_postcond:
  '''The postcondition should take a state and return a description object (list)
   that has the old description copied except for the current component desc,
   which should be updated using the old function.  We assume here that
   the function input is actually ignored. It still takes an input just
   for backwards compatibility with TStar.'''
  def __init__(self, oldPostcond, which):
    self.which_comp = which
    self.oldPost = oldPostcond
  def __call__(self, s):
    #print "in postcondition, state is "+str(s)
    #print "  and full data is "+str(s.data)
    #print "  and component "+str(self.which_comp)+" of data is "+str(s.data[self.which_comp])
    new_desc = s.desc[:]
    new_desc[self.which_comp] = self.oldPost(s)
    return new_desc

class make_new_initial_value_fn:
  def __init__(self, oldop, which):
    self.which_comp = which
    self.oldivf = oldop.initial_value_fn
  def __call__(self, s):
    #print "in initial_value_fn, state is "+str(s)
    #print "  and full data is "+str(s.data)
    #print "  and component "+str(self.which_comp)+" of data is "+str(s.data[self.which_comp])
    if self.oldivf:
        val = self.oldivf(s.data[self.which_comp])
    else: val = None
    return val
  
def fix_operator_for_costar(oldop, which_comp):
  '''Change the way the op gets and sets data from the data component.
  which_comp is in integer telling which component of the state data
  to use.'''
  reader = make_state_data_reader(which_comp)
  writer = make_state_data_writer(which_comp)
  new_precond = make_new_precond(oldop,which_comp)
  new_postcond = make_new_postcond(oldop.postcondition,which_comp)
  name = oldop.name
  if isinstance(oldop, TStar.OperatorSpecifiedWithDialog):
    new_fn = lambda d, stored, ps: writer(d, oldop.function(reader(d),stored,""))
    #print "new function for Op "+oldop.name+" is "+str(new_fn)
    new_initial_value_fn = make_new_initial_value_fn(oldop,which_comp)
    #new_initial_value_fn = lambda d: oldop.function(reader(d))
    newop = TStar.OperatorSpecifiedWithDialog(name,
                           new_fn, new_precond, new_postcond,
                           group=oldop.group, subgroup=oldop.subgroup,
                           subsubgroup=oldop.subsubgroup,
                           dialog_title=oldop.dialog_title,
                           prompt=oldop.prompt,
                           param_op_type=oldop.param_op_type,
                           filter_desc=oldop.filter_desc,
                           #initial_value_fn=oldop.initial_value_fn)
                           initial_value_fn=new_initial_value_fn)
  else:
    new_fn = lambda d,ps: writer(d, oldop.function(reader(d),""))
    #print "new function for Op "+oldop.name+" is "+str(new_fn)
    newop = TStar.Operator(name, new_fn, new_precond, new_postcond,
                           group=oldop.group, subgroup=oldop.subgroup,
                           subsubgroup=oldop.subsubgroup)
  return newop

def register_state_space_component(name,newStateClass, 
                                   newNodeClass, newOpSet, state_data_initializer,
                                   state_desc_initializer=""):
  global NUM_SS_COMPONENTS, ROLE, COMPONENT_NAMES, STATE_CLASSES,\
      NODE_CLASSES, OP_SETS, STATE_DATA_INITIALIZERS, STATE_DESC_INITIALIZERS
  COMPONENT_NAMES.append(name)
  STATE_CLASSES.append(newStateClass)
  NODE_CLASSES.append(newNodeClass)
  OP_SETS.append(newOpSet)
  STATE_DATA_INITIALIZERS.append(state_data_initializer)
  STATE_DESC_INITIALIZERS.append(state_desc_initializer)
  NUM_SS_COMPONENTS += 1
  ROLE = NUM_SS_COMPONENTS # default role view is 'all roles'
  return ROLE - 1 # role number for this component.

    
register_t_star_modifications()

# Here is an example of using COStar to create a state space that is the
# cartesian product of two component spaces. 


class NumState(TStar.State):
  def __init__(self,data, parent, last_operator, desc=""):
      super(NumState,self).__init__(data, parent, last_operator, desc)
  def __str__(self):
      return "### "+str(self.data)+" ###"

class NumNode(TStar.Node): pass

from math import sin
opA1 = TStar.Operator("Sine", lambda s, ps: NumState(sin(s.data+0.0),None,None))
opA2 = TStar.Operator("* 100", lambda s, ps: NumState(100 * s.data, None, None))
NumOps = [opA1, opA2]

def initializer1(): return NumState(1, None, None)


class StringState(TStar.State):
  def __init__(self,data, parent, last_operator, desc=""):
      super(StringState,self).__init__(data, parent, last_operator, desc)
  def __str__(self):
      return "~~~ "+self.data+" ~~~"

class StringNode(TStar.Node): pass

opB1 = TStar.Operator("Drop a char", lambda s, ps: StringState(s.data[:-1],None,None))
opB2 = TStar.Operator("Add a char", lambda s, ps: StringState("A"+s.data[:],None,None))
StringOps = [opB1, opB2]

def initializer2(): return StringState("Start!", None, None)
    
MENUBAR = None
roleMenu = None
def create_role_menu(menubar):
  from Tkinter import Menu
  global MENUBAR, roleMenu
  MENUBAR = menubar
  roleMenu = Menu(menubar)
  menubar.add_cascade(label="Select role", menu=roleMenu)
  for role in range(NUM_SS_COMPONENTS):
     selectrole = make_role_selector(role)
     roleMenu.add_command(label=COMPONENT_NAMES[role], command=selectrole)
  selectrole = make_role_selector(NUM_SS_COMPONENTS)
  roleMenu.add_command(label="All roles", command=selectrole)

class make_role_selector:
  def __init__(self, role):
    self.r = role
  def __call__(self):
    global ROLE, NUM_SS_COMPONENTS
    ROLE = self.r
    if ROLE < NUM_SS_COMPONENTS:
      role_name = COMPONENT_NAMES[self.r]
    else: role_name = "All roles"
    msg = "Role is now Role "+str(self.r)+" ("+role_name+")"
    print msg
    TStar.log(msg)
    TStar.PROPERTIES['ROLE']=ROLE
    TStar.redraw_tree()

def go(app_menu_creator=None):
  global INITIAL_DATA, NUM_SS_COMPONENTS, OP_SETS
  n = NUM_SS_COMPONENTS
  #print "Number of components is " + str(n)
  dummyCOStarObject = COStarObject()
  TStar.INITIAL_DATA = dummyCOStarObject.initial_data()
  #print "INITIAL_DATA is now: "+str(TStar.INITIAL_DATA)
  # Now register all COStar operators as TStar operators.
  TStar.register_operators([])
  #print "NUM_SS_COMPONENTS = "+str(NUM_SS_COMPONENTS)
  for comp in range(NUM_SS_COMPONENTS):
    #print "Length of OP_SETS for this component is "+str(len(OP_SETS[comp]))
    for op in OP_SETS[comp]:
      newop = fix_operator_for_costar(op, comp)
      #print "Fixed operator: " + newop.name
      TStar.register_more_operators([newop])
  TStar.setup_and_run()
  print "The session is finished."
  
if __name__ == '__main__':
  #global INITIAL_DATA
  print "Setting up COStar..."
  register_state_space_component("Number Component", NumState, NumNode, NumOps, initializer1)
  register_state_space_component("String Component", StringState, StringNode, StringOps, initializer2)
  go()
  
#  dummyCOStarObject = COStarObject()
#  TStar.INITIAL_DATA = dummyCOStarObject.initial_data()
#  print "INITIAL_DATA is now: "+str(TStar.INITIAL_DATA)
#  TStar.setup_and_run()
#  print "The session is finished."




