''' GameTester.py
  First version operational on 16 June 2007

  Updated 15 August 2007, 14 September 2007
   
The program supports RUDIMENTARY game playing in that
the player can move through the maze.
The player can sort of see and solve puzzles,

Saying magic phrases may cause things to happen and
the inventory of magic phrases is diminished.
Door opening via rule actions is working.

NEXT STEP IN IMPLEMENTATION:

Additional conditions to support rules:
  Predicate="Solved" Arg="2"

  etc.
  
Then it will be necessary to implement more elaborate logic of the game
in terms of SPECIFIED doors unlocking in response to saying COMBINATIONS
of magic phrases.

By serving as an importable module, this program supports the
program GraphicGameTester, which is under development in another file.

'''
print "Loading GameTester.py"
from PRIMEClasses import *
import PRIMEClasses

def error(msg): print msg
debugging = True
def diagnostic(msg):
    if debugging: print msg

def tell_user(msg): print msg
    
ROOM_VISITS = {} # Keeps track of how many times a room has been visited.

# This method might not be needed, but there might be a desire in the
# future to allow restarting a game without reloading or restarting the
# Python shell.  If such a need arises, this function will probably have
# to be fixed up somewhat to meet that need, resetting all doors to their
# initial states, and resetting puzzles to their original states.
def initialize_game():
    global CURRENT_ROOM_INDEX, MAGIC_PHRASE_INVENTORY, MAGIC_PHRASE_PUZZLE_NUM
    global CURRENT_PUZZLE, SCORE, TURNS
    global CURRENT_FACING_DIRECTION, ROOM_VISITS, INITIAL_MAGIC_PHRASES
    CURRENT_ROOM_INDEX = 0
    #MAGIC_PHRASE_INVENTORY = INITIAL_MAGIC_PHRASES
    MAGIC_PHRASE_INVENTORY = []
    MAGIC_PHRASE_PUZZLE_NUM = []
    CURRENT_PUZZLE = None
    SCORE = 0
    TURNS = 0
    CURRENT_FACING_DIRECTION = 'E'
    ROOM_VISITS = {}

def do_welcome():
    if not update_current_room():
        tell_user("This game is not playable, because room number 0 has not been defined.")
        return
    tell_user("Welcome to...")
    tell_user("    "+get_name() + "(version "+get_version()+", of "+get_design_date()+")")
    tell_user("               by "+get_authors()+"           Serial No. "+get_serial_no())
    tell_user("                        created with "+get_design_tools())
    tell_user("")

# Here is the main method of the game engine.    
def run_game():
    do_welcome()
    done = False
    while not done:
        tell_user('''Make a move (H for help).''')
        c = raw_input("  ENTER COMMAND: ")
        done = do_one_iteration(c)
    tell_user("Thanks for playing.")

def do_one_iteration(c):
    #global CURRENT_FACING_DIRECTION, DIRECTION_MAP
    c= c.upper()
    if c=='Q': return True
    if c=='D':
        describe_current_room()
        return False
    if c in ['N','E','W','S']:
        PRIMEClasses.CURRENT_FACING_DIRECTION = c
        tell_user("You are now facing " + DIRECTION_MAP[c] + ".")
        return False
    if c=='G':
        handle_go()
        return False
    if c=='M':
        maybe_utter()
        return False
    if c=='P':
        do_puzzle_activity()
        return False
    if c=='H': tell_user('''
   D for description of current room,
   N, E, S, or W to face north, east, south, or west,
   G to go in the current direction,
   P to work on the puzzle in front of you, if any,
   M to try to utter some magic words,
   Q to quit.'''); return False
    tell_user("I don't recognize the command: " + c)
    return False

def describe_current_room():
    #global CURRENT_ROOM_INDEX, PRIME_ROOMS
    tell_user("You are in a nice room, with number "+str(PRIMEClasses.CURRENT_ROOM_INDEX))
    current_room = PRIME_ROOMS[str(PRIMEClasses.CURRENT_ROOM_INDEX)]
    name = current_room.name
    if name != None: tell_user("The name of the room is: " + name)
    #print "Walls here are: " + str(current_room.walls)
    # We invoke the get_description method of the PRIME_wall class to handle
    # the details to describing itself.
    for w in current_room.walls.values():
        desc = w.get_description()
        if (desc != ""): tell_user(desc)

def change_room(new_room): pass # Overridden in GraphicGameTester.py

def maybe_utter():
    if len(MAGIC_PHRASE_INVENTORY)==0:
        tell_user("You need to know magic to utter magic words.")
        return
    tell_user("You have the following magic phrases available:")
    counter = 0
    for mp in MAGIC_PHRASE_INVENTORY:
        tell_user(str(counter)+': '+MAGIC_PHRASE_INVENTORY[counter])
        counter += 1
    tell_user("Enter the number for the magic phrase you with to say, or -1 to cancel.")
    #numstring = raw_input("? ")
    numstring = string_input("? ")
    n = -1
    try:
        n = int(numstring)
    except:
        tell_user("That's not a number I recognize. Try again!")
        maybe_utter()
        return
    if n<0:
        tell_user("Maybe you'll be more ready to engage in magic next time.")
        return
    if n>=len(MAGIC_PHRASE_INVENTORY):
        tell_user("That number is too big. Try again.")
        maybe_utter()
        return
    mp = MAGIC_PHRASE_INVENTORY[n]
    tell_user("You have chosen to utter the phrase:")
    tell_user("   "+mp)
    num = MAGIC_PHRASE_PUZZLE_NUM(mp)
    print "The Puzzle number of the magic phrase you uttered is:"+str(num)
    handle_event("Uttering", str(num))
    tell_user("There might or might NOT be magic happening.")
    MAGIC_PHRASE_INVENTORY.remove(mp)
    MAGIC_PHRASE_PUZZLE_NUM.remove(num)

def do_puzzle_activity():
    ''' This procedure will handle puzzle activity by the player.
    First, there must be a puzzle in front of the player.
    If there is a puzzle, then we acknowledge this and welcome the
    user to the puzzle (or back to the puzzle, if some work has
    already been done on it).
    Both music puzzles and image puzzles are handled the same way,
    except for a few details.
    '''
    puzzle_number = get_any_puzzle_here()
    if puzzle_number < 0:
        tell_user("There's no puzzle in front of you to work on.")
        return
    CURRENT_PUZZLE = get_puzzle(puzzle_number)
    tell_user("You are now working on puzzle "+CURRENT_PUZZLE.name)
    CURRENT_PUZZLE.describe()
    solution = list(eval(CURRENT_PUZZLE.correct_sequence))
    while(True):
        tell_user("The state of your work on this puzzle is represented by ")
        cos = CURRENT_PUZZLE.current_op_sequence
        tell_user("  the operator sequence: " + str(cos))
        if CURRENT_PUZZLE.solved:
            tell_user("You have already solved this puzzle. (Remember?)")
            return
        tell_user("Here are your choices:")
        ops = CURRENT_PUZZLE.operators.values()
        n = len(ops)
        for op in ops:
            tell_user("   "+op.to_string())
        tell_user("   -1: undo last operator")
        tell_user("   L:  Leave puzzle in its current state")
        good_choice = False
        while (not good_choice):
            c = raw_input("? ")
            if c=='L' or c=='l': return
            try:
                k = int(c)
                if k > -1 and k < n: good_choice = True; break
                else:
                    if k==-1:
                        if len(cos)>0: good_choice = True; break
                        else:
                            tell_user("There is nothing to undo.")
            except: None
            tell_user("Your input: "+c)
            tell_user("  is not a valid command here.")
        if k==-1:
            CURRENT_PUZZLE.current_op_sequence.pop()
            continue
        cos.append(k)
        if cos==solution:
            tell_user("Congratulations! You solved the puzzle.")
            handle_event("Solved puzzle", str(puzzle_number))
            tell_user("Here is a new magic phrase for you:")
            tell_user("    "+CURRENT_PUZZLE.message)
            CURRENT_PUZZLE.solved = True
            MAGIC_PHRASE_INVENTORY.append(CURRENT_PUZZLE.message)
            return
        
    
def get_any_puzzle_here():
    # Retrieve the wall that the player is facing.
    cw = get_current_wall()
    return cw.puzzle_num

def get_current_wall():
    global CURRENT_ROOM
    cw = CURRENT_ROOM.walls[PRIMEClasses.CURRENT_FACING_DIRECTION]
    return cw

def handle_go():
    # Is there a door in front of the player?
    wall = get_current_wall()
    #diagnostic("Current wall is "+str(wall))
    door = wall.door
    if (door==None):
        tell_user("There is no door here to walk through.")
        return
    # For now, assume the door is open.
    if (not door.open):
        tell_user("You can't walk through a door that is closed.")
        tell_user("You'll probably have to use some magic to open it.")
        return
    tell_user("OK, you are walking through the door.")
    global CURRENT_ROOM
    #handle_event("leaving", str(CURRENT_ROOM_INDEX))
    CRI = PRIMEClasses.CURRENT_ROOM_INDEX
    CRI = door.getOtherRoomNo(CRI)
    PRIMEClasses.CURRENT_ROOM_INDEX = CRI
    update_current_room()
    k=register_visit()
    if k==1:
        handle_event("Reached", str(CRI+1)) # Special case of first entry to this room.
        tell_user("You have just reached room number "+str(CURRENT_ROOM.num)+" ("+\
            CURRENT_ROOM.name+") for the first time.")
    else:
        tell_user("You are now in room number "+str(CURRENT_ROOM.num)+" ("+\
            CURRENT_ROOM.name+").")
    change_room(CURRENT_ROOM) # In GraphicGameTester, permits playing background music
                              # specific to this room.
    handle_event("Entering", str(CRI+1)) # Respond to any entry to this room.
    
def update_current_room():
    global CURRENT_ROOM, PRIME_ROOMS
    CRI = PRIMEClasses.CURRENT_ROOM_INDEX
    try:
        CURRENT_ROOM = PRIME_ROOMS[str(CRI)]
        return True
    except:
        tell_user("In this game, there is no room defined with number "+str(CRI)+".")
        tell_user("However, you are trying to go into it.")
        return None
        
COND_EVENTS_NEEDING_ARGS = ["Entering","Reached","Solved","Uttering"]
#["Entered specific room", "Solved specific puzzle","Uttered"]

FLAGS = {}
def handle_event(event_type, arg):
    global COND_EVENTS_NEEDING_ARGS, SCORE, FLAGS
    diagnostic( "Handling event with type: "+event_type+", and arg: "+arg)
    # Give each rule a chance to fire.
    for r in PRIME_RULES.values():
        diagnostic( "Trying rule "+str(r.num)+" called: "+r.name)
        if r.condition_type == "AND":  # The alternative here is OR
            AND = True;
            satisfied = True
        else:
            AND = False
            satisfied = False
        for c in r.conditions:
            # Generally a rule may have any number of conditions,
            # and each condition will have either a predicate or an event description.
            subresult = False
            if (c.predicate!=None and c.predicate!=""):
                diagnostic( "  Trying condition predicate: "+c.predicate+" with arg: "+c.arg)
                if c.predicate=="In":
                    if PRIMEClasses.CURRENT_ROOM_INDEX==int(c.arg):
                        subresult = True
                        diagnostic( "    Yes, we're in the right room.")
                    else:
                        diagnostic( "    We're not in room "+str(c.arg))
                        subresult = False;
                if c.predicate=="Reached":
                    if event_type=="Reached":
                        if arg==c.arg:
                            subresult = True
                            continue
                if c.predicate=="DoorClosed":
                    try: 
                        thisDoor = DOORS[int(c.arg)]
                    except:
                        tell_user("There is no door number "+c.arg+". Game Terminating.")
                    if not thisDoor.open:
                        subresult = True
                        diagnostic( "    OK, Door "+c.arg+" is in fact closed.")
                    else:
                        diagnostic( "    No, Door "+c.arg+" is open; condition fails.")
                if c.predicate=="DoorOpen":
                    try: 
                        thisDoor = DOORS[int(c.arg)]
                    except:
                        tell_user("There is no door number "+c.arg+". Game Terminating.")
                    if thisDoor.open:
                        subresult = True
                        diagnostic( "    OK, Door "+c.arg+" is in fact open.")
                    else:
                        diagnostic( "    No, Door "+c.arg+" is closed; condition fails.")
                if c.predicate=="ScoreBelow":
                    try: 
                        test_score = int(c.arg)
                    except:
                        tell_user("There is no test score provided.")
                    if SCORE < test_score:
                        subresult = True
                        diagnostic( "    OK, current score ("+str(SCORE)+") is less than "+c.arg+".")
                    else:
                        diagnostic( "    No, current score ("+str(SCORE)+") is NOT less than "+c.arg+".")
                if c.predicate=="ScoreAbove":
                    try: 
                        test_score = int(c.arg)
                    except:
                        tell_user("There is no test score provided.")
                    if SCORE > test_score:
                        subresult = True
                        diagnostic( "    OK, current score ("+str(SCORE)+") is greater than "+c.arg+".")
                    else:
                        diagnostic( "    No, current score ("+str(SCORE)+") is NOT greater than "+c.arg+".")
                if c.predicate=="FlagSet":
                    try: 
                        flag = c.arg
                    except:
                        tell_user("There is no flag mentioned.")
                    try:
                        subresult = FLAGS[flag]
                        if subresult: diagnostic( "    OK, flag "+c.arg+" is True.")
                        else: diagnostic( "    No, flag "+c.arg+" is False.")
                    except:
                        subresult = False
                        diagnostic( "    Note: flag "+c.arg+" is not registered and considered False.")
                if c.predicate=="FlagNotSet":
                    try: 
                        flag = c.arg
                    except:
                        tell_user("There is no flag mentioned.")
                    try:
                        subresult = not FLAGS[flag]
                        if subresult: diagnostic( "    OK, flag "+c.arg+" is False.")
                        else: diagnostic( "    No, flag "+c.arg+" is True.")
                    except:
                        subresult = True
                        diagnostic( "    Note: flag "+c.arg+" is not registered and considered False.")
                if not c.predicate in ["In", "DoorClosed", "DoorOpen", "ScoreBelow", "ScoreAbove",
                                       "FlagSet", "FlagNotSet"]:
                    error( "Unknown predicate, "+c.predicate+", in condition "+r.name)
            if (c.event!=None and c.event!=""):
                diagnostic( "  Trying condition event: "+c.event+": "+c.arg)
                if c.event==event_type:
                    subresult = True
                    if c.event in COND_EVENTS_NEEDING_ARGS:
                        #need to test argument, too."
                        print "c.arg = "+c.arg+"; arg = "+str(arg)
                        if c.arg!=str(arg):
                            subresult = False
                            diagnostic( "   Argument "+c.arg+" fails test.")
                        #else: subresult = True
                            
                            #if AND: satisfied = False
                    if subresult: diagnostic( "  This condition is satisfied.")
                    if not AND: satisfied = True; break
                else:
                    diagnostic( "  Condition not satisfied; next rule?")
                    subresult = False
                    #if AND: satisfied = False
            if AND: satisfied = satisfied and subresult
            else: satisfied = satisfied or subresult
            if AND and not satisfied: break
            if not AND and satisfied: break
        if not satisfied: continue
        diagnostic( "This rule fires!")
        diagnostic( "Actions are: "+str(r.actions))
        for a in r.actions:
            diagnostic("This action is: "+str(a))
            diagnostic("Event is: "+a.event)
            diagnostic("Arg is: "+a.arg)
            if a.event=="Score":
                adjust_score(int(a.arg)); continue
            if a.event=="Announce":
                announce(a.arg); continue
            if a.event=="Open":
                open_door(a.arg); continue
            if a.event=="Close":
                close_door(a.arg); continue
            if a.event=="Set":
                FLAGS[a.arg]=True; refresh_flag_display(); continue
            if a.event=="Clear":
                FLAGS[a.arg]=False; refresh_flag_display(); continue
            if a.event=="Play":
                full_filename = PRIMEClasses.FOLDER_NAME+"/"+a.arg
                play_audio_file(full_filename)
                continue
            diagnostic("We should perform action "+a.event+": "+a.arg)

# Override this in graphic game engines:
def refresh_flag_display():
    pass

def play_audio_file(file):
    print "Imagine the audio: "+file

def register_visit():
    cri = PRIMEClasses.CURRENT_ROOM_INDEX
    k = 0
    try:
        k = ROOM_VISITS[cri]
    except: None
    ROOM_VISITS[cri]=k+1
    return k+1

# Here are methods that implement rule actions:
def adjust_score(amount):
    global SCORE
    SCORE += amount
    tell_user("Your score is now: "+str(SCORE))

def announce(arg):
    tell_user("*** "+arg+" ***")

def string_input(prompt):
    txt = raw_input(prompt)
    return txt

def open_door(arg):
    n = int(arg)
    door = DOORS[n]
    if door.open: tell_user("Door "+arg+" is already open.")
    else: door.open = True; tell_user("Door "+arg+" is now open.")

def close_door(arg):
    n = int(arg)
    door = DOORS[n]
    if not door.open: tell_user("Door "+arg+" is already closed.")
    else: door.open = False; tell_user("Door "+arg+" is now closed.")

# If this module is loaded as the main module, load and run a sample game.    
if __name__ == '__main__':
    print "running the module GameTester"
    if load_game("samplePRIME"):
    #if load_game("corruptedTestFile.xml"):
        print "Game loaded."
        run_game()
