'''prime-image-ops.py
Implementations of image operators needed for creating and
solving PRIME game puzzles.

S. Tanimoto, August 29, 2007.
'''

import Image as PIL
from Tkinter import *
from ImageTk import PhotoImage

DEFAULT_IMAGE = None
def set_up_tests():
    global DEFAULT_IMAGE
    DEFAULT_IMAGE = "../mona-rgb.jpg"
    im = PIL.open(DEFAULT_IMAGE)
    im.thumbnail((256,256))
    TK_ROOT = Tk(className="prime-image-ops test")
    CANVAS = Canvas(TK_ROOT, width=512, height=512)
    CANVAS.pack()
    p = im.load()
    #pim = PhotoImage(shuffle_rows_twice(im))
    pim = PhotoImage(pixel_crossover(im))
    #pim = PhotoImage(pixel_crossover(pixel_crossover(pixel_crossover(pixel_crossover(im)))))
    CANVAS.create_image(256,256,image=pim)
    TK_ROOT.mainloop()


def horiz_flip(im):
    pix1 = im.load()
    im2 = im.copy()
    pix2 = im2.load()
    s = im.size
    w = s[0]; h = s[1]
    for y in range(h):
        for x in range(w):
            pix2[x,y] = pix1[w-1-x, y]
    return im2

def odd(x): return x%2==1

def rotate_180(im):
    pix1 = im.load(); im2 = im.copy(); pix2 = im2.load()
    s = im.size; w = s[0]; h = s[1]
    for x in range(w):
        for y in range(h):
            pix2[x,y] = pix1[w-1-x, h-1-y]
    return im2
def shuffle_rows(im):
    pix1 = im.load(); im2 = im.copy(); pix2 = im2.load()
    s = im.size; w = s[0]; h = s[1]
    for x in range(w):
        x0 = 2*x
        if x>=w/2: x0 = w-1 - (2*(w-1-x))
        for y in range(h):
            pix2[x,y] = pix1[x0, y]
    return im2

def shuffle_columns(im):
    pix1 = im.load(); im2 = im.copy(); pix2 = im2.load()
    s = im.size; w = s[0]; h = s[1]
    for y in range(h):
        y0 = 2*y
        if y>=h/2: y0 = h-1 - (2*(h-1-y))
        for x in range(w):
            pix2[x,y] = pix1[x, y0]
    return im2

def pixel_crossover(im):
    # Each pixel component in (R,G,B) is considered a byte made up of
    # two nibbles N1, N0, of four bits each.
    # For every pixel, a "partner" is found by adding w/2 to x and
    # h/2 to y.  The two pixels are  N1, N0 and P1, P0.
    # The first pixel value is replaced by P0, N1 which is
    # the right byte of P followed by the left byte of the original pixel,
    # and the partner by N0, P1.  This scrambles things up well.
    # This transformation generates a cyclic group of order 4, provided
    # the width and height are even numbers.    
    pix1 = im.load()
    im2 = im.copy()
    pix2 = im2.load()
    s = im.size; w = s[0]; h = s[1]; w2=int(w/2); h2=int(h/2)
    for y in range(h):
        yp = (y+h2) % h # Partner's y coord.
        for x in range(w):
            xp = (x+w2) % w # Partner's x coord.
            N = pix1[x, y]   # Original home pixel.
            P = pix1[xp, yp] # Partner pixel.
            new_pixel_values = [0,0,0] # List holding color vals temporarily.
            for i in range(3):
                N1 = N[i] / 16 # Left nibble of home pixel
                N0 = N[i] % 16 # Right nibble
                P1 = P[i] / 16 # Left nibble of partner
                P0 = P[i] % 16 # Right nibble
                new_pixel_values[i] = P0*16+N1 # Result of crossover.
            pix2[x,y] = tuple(new_pixel_values) # Store result.
    return im2

def shuffle_columns_twice(im):
  return(shuffle_columns(shuffle_columns(im)))

def shuffle_rows_twice(im):
  return(shuffle_rows(shuffle_rows(im)))

set_up_tests()

