''' PRIMEClasses.py

This module provides both class definitions to represent PRIME
games in memory and some code for reading an XML file that
describes a PRIME game.

Written by S. Tanimoto, 16 June 2007.
Modified 16 August 2007.
Status: Working

Provides support classes to GameTester.py, GraphicGameTester.py
and possibly to the future PRIME Game Designer program.

At that time, it might be desirable to add methods to CREATE
the kinds of XML files that this module currently reads.

'''
#import pprint
 
import xml.dom.minidom
from xml.dom.minidom import Node

# global variables to hold game data:
NAME = "DEFAULT GAME"
VERSION = "0.000"
AUTHORS = "Anonymous"
DESIGN_TOOLS = "unknown design tool"
DESIGN_DATE = "1000 BC"
SERIAL_NO = "0"
GAME   = None
LAYOUT = None
#ROOMS  = []
IMAGE_PUZZLES = {}
MUSIC_PUZZLES = {}
RULES  = {}
 # The following phrases are arbitrary initial phrases,
 # and a future enhancement is to load these from the XML file.
INITIAL_MAGIC_PHRASES = ['Abracadabra', 'Open Sesame']
DIRECTION_MAP = {'N':'north', 'E':'east', 'W':'west', 'S':'south'}

# Define some global variables for game playing:

CURRENT_ROOM_INDEX = 0
CURRENT_ROOM = None
#MAGIC_PHRASE_INVENTORY = INITIAL_MAGIC_PHRASES
MAGIC_PHRASE_INVENTORY = []
MAGIC_PHRASE_PUZZLE_NUM = []

CURRENT_PUZZLE = None
SCORE = 0
TURNS = 0
CURRENT_FACING_DIRECTION = 'E'
FOLDER_NAME = ""

def load_game(foldername, filename=None):
    global GAME, LAYOUT, NAME, AUTHORS, VERSION,\
           DESIGN_TOOLS, DESIGN_DATE, SERIAL_NO,\
           FOLDER_NAME
    FOLDER_NAME = foldername
    if filename==None:
        filename = FOLDER_NAME+"/"+FOLDER_NAME+".xml"
    try:
        GAME = xml.dom.minidom.parse(filename)
    except:
        print "The file \""+filename+"\" could not be parsed as an XML file."
        return
    try:
        primegame = GAME.getElementsByTagName("prime_game")[0]
    except:
        #if primegame==None:
        print "Although "+filename+" is an XML file, no PRIME_GAME tag was found."
        return
    NAME = primegame.getAttribute("name")
    VERSION = primegame.getAttribute("version")
    AUTHORS = primegame.getAttribute("authors")
    DESIGN_TOOLS = primegame.getAttribute("design_tools")
    DESIGN_DATE = primegame.getAttribute("design_date")
    SERIAL_NO = primegame.getAttribute("design_serial_no")
    try:
        LAYOUT = GAME.getElementsByTagName("layout")[0]
    except:
        print "Error: No layout in this game file."
        return None
    domRooms = LAYOUT.getElementsByTagName("room")
    for room in domRooms: build_room(room)
    #pprint.pprint(ROOMS)
    domImagePuzzles = GAME.getElementsByTagName("image_puzzle")
    #pprint.pprint(domImagePuzzles)
    for puzzle in domImagePuzzles: build_puzzle(puzzle, "image")
    domMusicPuzzles = GAME.getElementsByTagName("music_puzzle")
    #pprint.pprint(domMusicPuzzles)
    for puzzle in domMusicPuzzles: build_puzzle(puzzle, "music")
    build_rulebase()
    return primegame

def get_name(): return NAME
def get_authors(): return AUTHORS
def get_version(): return VERSION
def get_design_tools(): return DESIGN_TOOLS
def get_design_date(): return DESIGN_DATE
def get_serial_no(): return SERIAL_NO

class PRIME_room:
    def __init__(self, num, name):
        self.num = num # Critical for maze layout method:
         # row = num / 3; col = num % 3
        self.name = name
        self.background_audio = None
        self.walls = {}
        for loc in ['N','E','W','S']:
            self.walls[loc] = PRIME_wall(loc)

class PRIME_wall:
    def __init__(self, loc):
        self.wallpaper = None
        self.text = None
        self.puzzle_num = None
        self.door = None
        self.loc = loc
        self.room_num = None
    def get_description(self):
        global direction_map
        locname = DIRECTION_MAP[self.loc]
        desc = ""
        if self.wallpaper != None:
            desc += "The wallpaper on the "+locname+" wall is " + str(self.wallpaper)+"\n"
        if self.text != None:
            desc += "There is text: " + self.text + "\n"
        if self.puzzle_num != None:
            desc += "There is a puzzle on the "+locname+" wall with number: "+str(self.puzzle_num)+"\n"
        if self.door != None:
            to = self.door.getOtherRoomNo(self.room_num)
            desc += "There is a door in the "+locname+" wall leading to room "+str(to)+"\n"
            if self.door.open: state="open"
            else: state="closed"
            desc += "That door is "+state+".\n"
        return desc

            

class PRIME_door:
    def __init__(self, num, fromRoomNo, toRoomNo):
        self.num = num
        self.fromRoomNo = fromRoomNo
        self.toRoomNo = toRoomNo
        self.open = False
    def getOtherRoomNo(self, givenRoomNo):
        #print "In getOtherRoomNo, the number of this door is "+str(self.num)
        if givenRoomNo==self.fromRoomNo: return self.toRoomNo
        if givenRoomNo==self.toRoomNo: return self.fromRoomNo
        #print "The types are :" + str(type(self.toRoomNo)) +", "+str(type(self.fromRoomNo))+", "+str(type(givenRoomNo))
        #print "The values are: "+str(self.toRoomNo)+", "+str(self.fromRoomNo)+", "+str(givenRoomNo)
        print "This door doesn't connect from room "+str(givenRoomNo)
        return givenRoomNo

class PRIME_puzzle:
    def __init__(self,num):
        self.name = "Puzzle "+str(num)
        self.num = num
        self.operators = {}
        self.correct_sequence = ()
        self.message = ""
        self.current_op_sequence = [] # saved as a list.
        self.proc_images = {}
        self.proc_music = {}
        self.note_seq = {}
        self.solved = False
    def get_description(self):
        desc = ""
        puzzle_type=self.get_puzzle_type()
        desc += "This is a "+puzzle_type+" puzzle named \""+self.name+"\".\n"
        desc += "The operators are:\n"
        for op in self.operators.values():
            desc += op.to_string()+"\n"
        if puzzle_type=="image":
            desc += "The processed images are:\n"
            for pif in self.proc_images.values():
                desc += pif + "\n"
        return desc
    def describe(self):
        print self.get_description()
    def get_puzzle_type(self):
        puzzle_type="image"
        if self.__class__==PRIME_music_puzzle:
            puzzle_type="music"
        return puzzle_type


class PRIME_image_puzzle(PRIME_puzzle):
    def __init__(self,num):
        PRIME_puzzle.__init__(self,num)

class PRIME_music_puzzle(PRIME_puzzle):
    def __init__(self,num):
        PRIME_puzzle.__init__(self,num)
        #self.proc_music = {}
        #self.note_seq = {}
    
class PRIME_operator:
    def __init__(self, role_num, name):
        # role_num is like number, and there will be only one operator
        # with any given role number in a given puzzle, but
        # the role numbers can be re-used from one puzzle to another.
        self.role_num = role_num
        self.name = name
    def to_string(self):
        return str(self.role_num)+": "+self.name

class PRIME_rule:
    def __init__(self,num,name):
        self.name = name
        self.num = num
        self.conditions = []
        self.actions = []
        self.firings = 0
    def describe(self):
        print "This is rule: "+str(self.num)
        for cond in self.conditions:
            print str(cond)
        for action in self.actions:
            print str(action)

class PRIME_condition:
    None

class PRIME_action:
    None

PRIME_ROOMS = {}
DOORS = {} # keyed by "num"
PRIME_RULES = {}


def build_room(room):
    global PRIME_ROOMS
    name = room.getAttribute("name")
    background_audio = room.getAttribute("background_audio")
    string_num = room.getAttribute("num")
    num = int(string_num)
    thisRoom = PRIME_room(num, name)
    thisRoom.background_audio = background_audio
    PRIME_ROOMS[string_num] = thisRoom
    domWalls = room.getElementsByTagName("wall")
    for domWall in domWalls:
        w = build_wall(num, domWall)
        thisRoom.walls[w.loc] = w
        
def build_wall(room_num, wall):
    loc = wall.getAttribute("loc")
    w = PRIME_wall(loc)
    wallpapers = wall.getElementsByTagName("wallpaper")
    if len(wallpapers)>0:
        w.wallpaper = wallpapers[0].getAttribute("name")
    text = wall.getElementsByTagName("text")
    if len(text)>0:
        w.text = text[0].getAttribute("contents")
    puzzles = wall.getElementsByTagName("puzzle_inst")
    if len(puzzles)>0:
        w.puzzle_num = puzzles[0].getAttribute("puzzle_num")
    doors = wall.getElementsByTagName("door")
    if len(doors)>0:
        w.door = build_door(room_num, doors[0])
        #print "Just assigned to wall.door: " + str(w.door)
    w.room_num = room_num
    return w

def build_door(room_num, domDoor):
    #print "Calling build_door with room_num = "+str(room_num)
    global DOORS
    num = int(domDoor.getAttribute("num"))
    toRoomNo = int(domDoor.getAttribute("into"))
    # if a door already exists connecting these two rooms
    # because it was defined in the other room, return that
    # door without creating a new one.
    possible = None
    try:
        possible = DOORS[num]
    except: None
    if possible!=None: return possible
    # Otherwise, create the new door:
    door = PRIME_door(num, room_num, toRoomNo);
    state = domDoor.getAttribute("state")
    if state=="open": door.open = True
    # Register it:
    #print "Registering door number "+str(num)
    DOORS[num] = door
    return door
    
def build_puzzle(domPuzzle, puzzle_type):
    #print "Calling build_puzzle for ("+puzzle_type+") puzzle = "+str(domPuzzle)
    global IMAGE_PUZZLES, MUSIC_PUZZLES
    num = int(domPuzzle.getAttribute("num"))
    #print "This is puzzle number " + str(num)
    if puzzle_type=="image":
        puzzle = PRIME_image_puzzle(num)
        IMAGE_PUZZLES[num] = puzzle
    if puzzle_type=="music":
        puzzle = PRIME_music_puzzle(num)
        MUSIC_PUZZLES[num] = puzzle
    domProcImages = domPuzzle.getElementsByTagName("proc_image")
    for domProcImage in domProcImages:
        pif = domProcImage.getAttribute("filename")
        raw_pis = domProcImage.getAttribute("seq")
        try:
            pis = eval(raw_pis)
        except:
            pis = raw_pis
        #print "pis = "+str(pis)
        puzzle.proc_images[pis] = pif
    if puzzle_type=="music":
        domProcMusics = domPuzzle.getElementsByTagName("proc_music")
        for domProcMusic in domProcMusics:
            pmf = domProcMusic.getAttribute("filename")
            try:
              pms = domProcMusic.getAttribute("seq") # operator sequence
              pms = eval(pms) # operator sequence
            except:
              pms = "default"
            msg = domProcMusic.getAttribute("msg") # scrambled note sequence
            #print "In build_puzzle, msg="+msg
            #print "seq="+pms
            puzzle.proc_music[pms] = pmf
            puzzle.note_seq[pms] = msg
        domSources = domPuzzle.getElementsByTagName("source")
        puzzle.background = domSources[0].getAttribute("file")
    domOperators = domPuzzle.getElementsByTagName("operator")
    #print "domOperators="+str(domOperators)
    for domOperator in domOperators:
        op_role = domOperator.getAttribute("role")
        op_name = domOperator.getAttribute("name")
        puzzle.operators[op_role] = PRIME_operator(op_role, op_name)
    correct_sequences = domPuzzle.getElementsByTagName("correct_sequence")
    puzzle.correct_sequence = correct_sequences[0].getAttribute("seq")
    #print "Correct sequence is "+str(puzzle.correct_sequence)
    messages = domPuzzle.getElementsByTagName("message")
    puzzle.message = messages[0].getAttribute("text")
    puzzle.name = domPuzzle.getAttribute("name")
    #print "Hidden message is "+str(puzzle.message)

def get_puzzle(num):
    n = int(num)
    global IMAGE_PUZZLES, MUSIC_PUZZLES
    try: p = IMAGE_PUZZLES[n]; return p
    except:
        try: p = MUSIC_PUZZLES[n]; return p
        except:
            print "There is no puzzle with puzzle number "+str(num)
            return None        

   
def build_rule(domRule):
    global PRIME_RULES
    name = domRule.getAttribute("name")
    string_num = domRule.getAttribute("num")
    num = int(string_num)
    thisRule = PRIME_rule(num, name)
    PRIME_RULES[string_num] = thisRule
    thisRule.condition_type = domRule.getAttribute("condition_type")
    domConds = domRule.getElementsByTagName("condition")
    for domCond in domConds:
        c = build_condition(domCond)
        thisRule.conditions.append(c)
    domActions = domRule.getElementsByTagName("action")
    for domAction in domActions:
        a = build_action(domAction)
        thisRule.actions.append(a)

def build_condition(domCond):
    c = PRIME_condition()
    c.predicate = domCond.getAttribute("predicate")
    c.event = domCond.getAttribute("event")
    c.arg = domCond.getAttribute("arg")
    return c

def build_action(domAction):
    a = PRIME_action()
    a.event = domAction.getAttribute("event")
    a.arg = domAction.getAttribute("arg")
    return a
    
def build_rulebase():
    try:
        control_program = GAME.getElementsByTagName("control_program")[0]
    except:
        print "This game file does not contain a CONTROL_PROGRAM section."
        print "So it is not playable. Terminating now."
        import os
        os._exit(1)
    domRules = control_program.getElementsByTagName("rule")
    for domRule in domRules: build_rule(domRule)
    

if __name__ == '__main__':
    print "running the module PRIMEClasses"
