'''Michael Lange <klappnase at 8ung.at>
   2 The Meter class provides a simple progress bar widget for Tkinter.
   3 
   4 INITIALIZATION OPTIONS:
   5 The widget accepts all options of a Tkinter.Frame plus the following:
   6 
   7     fillcolor -- the color that is used to indicate the progress of the
   8                  corresponding process; default is "orchid1".
   9     value -- a float value between 0.0 and 1.0 (corresponding to 0% - 100%)
  10              that represents the current status of the process; values higher
  11              than 1.0 (lower than 0.0) are automagically set to 1.0 (0.0); default is 0.0 .
  12     text -- the text that is displayed inside the widget; if set to None the widget
  13             displays its value as percentage; if you don't want any text, use text="";
  14             default is None.
  15     font -- the font to use for the widget's text; the default is system specific.
  16     textcolor -- the color to use for the widget's text; default is "black".
  17 
  18 WIDGET METHODS:
  19 All methods of a Tkinter.Frame can be used; additionally there are two widget specific methods:
  20 
  21     get() -- returns a tuple of the form (value, text)
  22     set(value, text) -- updates the widget's value and the displayed text;
  23                         if value is omitted it defaults to 0.0 , text defaults to None .
  24 '''
   
import Tkinter

class Meter(Tkinter.Frame):
    def __init__(self, master, width=300, height=20, bg='white', fillcolor='orchid1',\
             value=0.0, text=None, font=None, textcolor='black', *args, **kw):
        Tkinter.Frame.__init__(self, master, bg=bg, width=width, height=height, *args, **kw)
        self._value = value

        self._canv = Tkinter.Canvas(self, bg=self['bg'], width=self['width'], height=self['height'],\
                                  highlightthickness=0, relief='flat', bd=0)
        #self._canv.pack(fill='both', expand=1)
        self._canv.grid(row=0, column=0)
        self._rect = self._canv.create_rectangle(0, 0, 0, self._canv.winfo_reqheight(), fill=fillcolor,\
                                               width=0)
        self._text = self._canv.create_text(self._canv.winfo_reqwidth()/2, self._canv.winfo_reqheight()/2,\
                                          text='', fill=textcolor)
        if font:
            self._canv.itemconfigure(self._text, font=font)

        self.set(value, text)
        self.bind('<Configure>', self._update_coords)

    def _update_coords(self, event):
        '''Updates the position of the text and rectangle inside the canvas when the size of
        the widget gets changed.'''
        # looks like we have to call update_idletasks() twice to make sure
        # to get the results we expect
        self._canv.update_idletasks()
        self._canv.coords(self._text, self._canv.winfo_width()/2, self._canv.winfo_height()/2)
        self._canv.coords(self._rect, 0, 0, self._canv.winfo_width()*self._value, self._canv.winfo_height())
        self._canv.update_idletasks()

    def get(self):
        return self._value, self._canv.itemcget(self._text, 'text')

    def set(self, value=0.0, text=None):
        #make the value failsafe:
        print "ProgressMeter setting value to "+str(value)
        if value < 0.0:
            value = 0.0
        elif value > 1.0:
            value = 1.0
        self._value = value
        if text == None:
            #if no text is specified use the default percentage string:
            text = str(int(round(100 * value))) + ' %'
        self._canv.coords(self._rect, 0, 0, self._canv.winfo_width()*value, self._canv.winfo_height())
        self._canv.itemconfigure(self._text, text=text)
        self._canv.update_idletasks()

##-------------demo code--------------------------------------------##

def _demo(meter, value):
    meter.set(value)
    if value < 1.0:
        value = value + 0.005
        meter.after(50, lambda: _demo(meter, value))
    else:
        meter.set(value, 'Demo successfully finished')

if __name__ == '__main__':
    root = Tkinter.Tk(className='meter demo')
    m = Meter(root, relief='ridge', bd=3)
    m.pack(fill='x')
    m.set(0.0, 'Starting demo...')
    m.after(1000, lambda: _demo(m, 0.0))
    root.mainloop()
