'''RuleBaseDesigner05.py
  This is a component state-space for PRIMEDesigner05.py.
  Working, but needs display of an actual image.
  Needs a group of operators for selecting an initial image and
   secret message.  The precondition should be that the currently
   selected puzzle has nothing in it.
  Change the precondition on adding transformations to make sure
   that the selected puzzle has at least one text line in it already.
'''

import TStar09 as TStar
import COStar09 as COStar
from Tkinter import Canvas
import TextListListDesigner05 as TextListListDesigner

GROUP_NAME = None
OBJECT_NAME = None

class RuleBState(TextListListDesigner.TextListListState):
  def __init__(self,data, parent, last_operator, desc=""):
      super(RuleBState,self).__init__(data, parent, last_operator, desc)
      self.obj_name = OBJECT_NAME
  def get_background_color(self): return "light green"

class RuleBNode(TextListListDesigner.TextListListNode):
  pass

def ok_to_highlight(rblws, ps=""):
  if rblws.selected < 0: return False
  ln = len(rblws.tll[rblws.selected]) # num. lines in selected rule.
  if ln<4: return False
  if ln==4 and rblws.lineToHighlight>-1: return False
  return True # There are is at least one unlighted line
    # that is not IF:, THEN:, or END.

def has_highlight(rblws, ps=""):
  return rblws.selected >= 0 and  rblws.lineToHighlight>=0
  
def delete_highlighted(rblws, ps=""):
  rblws2 = TextListListDesigner.TextListListWithSelection()
  rblws2.tll = rblws.tll[:] # copy elements, not reference
  sel = rblws.selected
  rblws2.selected = sel
  rule = rblws.tll[sel]
  k = rblws.lineToHighlight
  new_rule = rule[:k]+rule[k+1:]
  rblws2.tll[sel]=new_rule
  k -= 1
  if k<1:
    rblws2.lineToHighlight = -1 # no highlight now
  else:
    line = new_rule[k]
    if line=="THEN:":
      k-=1
      if k<1: rblws2.lineToHighlight = -1 # no highlight now
    else:
      rblws2.lineToHighlight = k  
  return rblws2

def highlight_next(rblws, ps=""):
  rblws2 = TextListListDesigner.TextListListWithSelection()
  rblws2.tll = rblws.tll
  rblws2.selected = rblws.selected
  k = rblws.lineToHighlight + 1  
  ln = len(rblws.tll[rblws.selected]) # num. lines in selected rule.
  if k>=ln-1: k=1
  line = rblws.tll[rblws.selected][k]
  if line=="THEN:":
    k+=1
    if k>ln-2: k=1
  rblws2.lineToHighlight = k  
  return rblws2

def highlight_previous(rblws, ps=""):
  rblws2 = TextListListDesigner.TextListListWithSelection()
  rblws2.tll = rblws.tll
  rblws2.selected = rblws.selected
  k = rblws.lineToHighlight - 1  
  ln = len(rblws.tll[rblws.selected]) # num. lines in selected rule.
  if k<1: k=ln-2
  line = rblws.tll[rblws.selected][k]
  if line=="THEN:":
    k-=1
    if k<1: k=ln-1
  rblws2.lineToHighlight = k  
  return rblws2

def add_condition_to_selected(tllws, ps="A string"):
  "Add one condition to the currently selected rule."
  tllws2 = TextListListDesigner.TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  loc = newtl.index("THEN:")
  newtl.insert(loc,ps)
  tllws2.lineToHighlight = loc
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2

def add_action_to_selected(tllws, ps="A string"):
  "Add one condition to the currently selected rule."
  tllws2 = TextListListDesigner.TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  loc = newtl.index("END.")
  newtl.insert(loc,ps)
  tllws2.lineToHighlight = loc
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2

def add_condition(rblws, ps=""):
  rblws2 = add_condition_to_selected(rblws,"(Condition:) "+ps)
  return rblws2

def add_condition_in(rblws, ps=""):
  return add_condition(rblws, "In room ")
  
def add_condition_entering(rblws, ps=""):
  return add_condition(rblws, "Entering room: ")
  
def add_condition_uttering(rblws, ps=""):
  return add_condition(rblws, "Uttering phrase: ")
  
def add_condition_door_open(rblws, ps=""):
  return add_condition(rblws, "Door is open: ")

def add_condition_door_closed(rblws, ps=""):
  return add_condition(rblws, "Door is closed: ")

def add_condition_score_less_than(rblws, ps=""):
  return add_condition(rblws, "Score is less than: ")

def add_condition_score_greater_than(rblws, ps=""):
  return add_condition(rblws, "Score is greater than: ")

def add_condition_flag_not_set(rblws, ps=""):
  return add_condition(rblws, "Flag is NOT set. Its name is: ")

def add_condition_flag_is_set(rblws, ps=""):
  return add_condition(rblws, "Flag IS set. Its name is: ")

def add_condition_reached(rblws, ps=""):
  return add_condition(rblws, "Reached room for 1st time: ")

def add_action(rblws, ps=""):
  rblws2 = add_action_to_selected(rblws,"(Action:) "+ps)
  return rblws2

def add_action_open_door(rblws, ps=""):
  return add_action(rblws, "Open door: ")

def add_action_close_door(rblws, ps=""):
  return add_action(rblws, "Close door: ")

def add_action_score(rblws, ps=""):
  return add_action(rblws, "Increase score by: ")

def add_action_announce(rblws, ps=""):
  return add_action(rblws, "Announce: ")

def add_action_set_flag(rblws, ps=""):
  return add_action(rblws, "Set flag whose name is: ")

def add_action_clear_flag(rblws, ps=""):
  return add_action(rblws, "Clear flag whose name is: ")

def setup():
  global GROUP_NAME, OBJECT_NAME
  GROUP_NAME = "Rule Base Designer"
  OBJECT_NAME = "rule"
  basic_ops = TextListListDesigner.make_basic_tll_ops(obj_name=OBJECT_NAME,
                                                    group_name=GROUP_NAME,
                                                    tll_type="Rule")
  #image_op1 = TStar.Operator("Add condition", add_condition,
  #                         group=GROUP_NAME,
  #                         precondition=TextListListDesigner.need_selection)
  cond_op1 = TStar.OperatorSpecifiedWithDialog("Add condition: 'In room'",
               lambda b,rm,p: add_condition(b,"In room: "+str(rm)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the room number")
  cond_op1a = TStar.Operator("Add condition: 'In room'", add_condition_in,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op2 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Entering room'",
               lambda b,rm,p: add_condition(b,"Entering room: "+str(rm)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the room number")
  cond_op2a = TStar.Operator("Add condition: 'Entering room'", add_condition_entering,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op3 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Uttering phrase'",
               lambda b,phrase_num,p: add_condition(b,"Uttering phrase: "+str(phrase_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the phrase (puzzle) number")
  cond_op3a = TStar.Operator("Add condition: 'Uttering phrase'", add_condition_uttering,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op4 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Door closed'",
               lambda b,door_num,p: add_condition(b,"Door closed: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the door number")
  cond_op4a = TStar.Operator("Add condition: 'Door closed'", add_condition_door_closed,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op5 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Door open'",
               lambda b,door_num,p: add_condition(b,"Door open: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the door number")
  cond_op5a = TStar.Operator("Add condition: 'Door open'", add_condition_door_open,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op6 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Flag set'",
               lambda b,door_num,p: add_condition(b,"Flag set: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               prompt="Enter the flag name")
  cond_op6a = TStar.Operator("Add condition: 'Flag set'", add_condition_flag_is_set,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op7 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Flag not set'",
               lambda b,door_num,p: add_condition(b,"Flag not set: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               prompt="Enter the flag name")
  cond_op7a = TStar.Operator("Add condition: 'Flag not set'", add_condition_flag_not_set,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op8 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Score is below'",
               lambda b,score,p: add_condition(b,"Score is below: "+str(score)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the score amount")
  cond_op8a = TStar.Operator("Add condition: 'Score is below'", add_condition_score_less_than,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op9 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Score is above'",
               lambda b,score,p: add_condition(b,"Score is above: "+str(score)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the score amount")
  cond_op9a = TStar.Operator("Add condition: 'Score is above'", add_condition_score_greater_than,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  cond_op10 = TStar.OperatorSpecifiedWithDialog("Add condition: 'Reached room for 1st time'",
               lambda b,rm,p: add_condition(b,"Reached room for 1st time: "+str(rm)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the room number")
  cond_op10a = TStar.Operator("Add condition: 'Reached room'", add_condition_reached,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  action_op1 = TStar.OperatorSpecifiedWithDialog("Add action: 'Open door'",
               lambda b,door_num,p: add_action(b,"Open door: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the door number")
  action_op1a = TStar.Operator("Add action: 'Open door'", add_action_open_door,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  action_op2 = TStar.OperatorSpecifiedWithDialog("Add action: 'Close door'",
               lambda b,door_num,p: add_action(b,"Close door: "+str(door_num)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the door number")
  action_op2a = TStar.Operator("Add action: 'Open door'", add_action_open_door,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  action_op3 = TStar.OperatorSpecifiedWithDialog("Add action: 'Increase score'",
               lambda b,amount,p: add_action(b,"Increase score: "+str(amount)),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               param_op_type='Int',
               prompt="Enter the number of points")
  action_op3a = TStar.Operator("Add action: 'Increase score'", add_action_score,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  action_op4 = TStar.OperatorSpecifiedWithDialog("Add action: 'Announce...'",
               lambda b,msg,p: add_action(b,"Announce: "+msg),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               prompt="Enter the announcement message")
  action_op4b = TStar.OperatorSpecifiedWithDialog("Add action: 'Play audio file..'",
               lambda b,msg,p: add_action(b,"Play audio file: "+msg),
               precondition=TextListListDesigner.need_selection,
               param_op_type='File',
               group=GROUP_NAME,
               prompt="Choose the audio file to play")
  action_op5 = TStar.OperatorSpecifiedWithDialog("Add action: 'Set flag'",
               lambda b,name,p: add_action(b,"Set flag: "+name),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               prompt="Enter the name of the flag")
  action_op5a = TStar.Operator("Add action: 'Set flag'", add_action_set_flag,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  action_op6 = TStar.OperatorSpecifiedWithDialog("Add action: 'Clear flag'",
               lambda b,name,p: add_action(b,"Clear flag: "+name),
               precondition=TextListListDesigner.need_selection,
               group=GROUP_NAME,
               prompt="Enter the name of the flag")
  action_op6a = TStar.Operator("Add action: 'Set flag'", add_action_set_flag,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  h_op1 = TStar.Operator("Highlight next line", highlight_next,
                           group=GROUP_NAME,
                           precondition=ok_to_highlight)
  h_op2 = TStar.Operator("Highlight previous line", highlight_previous,
                           group=GROUP_NAME,
                           #precondition=TextListListDesigner.need_selection)
                           precondition=ok_to_highlight)
  h_op3 = TStar.Operator("Delete highlighted line", delete_highlighted,
                           group=GROUP_NAME,
                           precondition=has_highlight)

  basic_ops += [cond_op1, cond_op2, cond_op3, cond_op4, cond_op5, cond_op6,
                cond_op7, cond_op8, cond_op9, cond_op10,
                action_op1, action_op2, action_op3, action_op4, action_op4b, action_op5, action_op6,
                h_op1, h_op2, h_op3]

  COStar.register_state_space_component(GROUP_NAME, RuleBState, RuleBNode,
                                      basic_ops,
                                      TextListListDesigner.initializer)





