""" TStar09.py       Steve Tanimoto, 7 Feb 2008.
    (updated Aug. 31, 2007)
    (last updated 21 April, 2009)
    
    The T* module, Version 0.9.  It builds on version 0.81.
    V 0.9 adds a dummy logging method.  This can be overridden by
        application-specific logging in app files such as PRIME Designer.
      Also adds 1-level UNDO for tree-buiilding operations.
      And also: option to use absolute operator numbers (needed
          for collaborative applications where tree merging will be done).
          Right now, TStar is set to load either mode of file but
          always write in the mode controlled by PREFER_ABSOLUTE_OPERATOR_NOS,
          which is False by default but can be overridden by apps like PRIME Designer.
      Added options to grow subtrees of either depth 2 or 3.
    V 0.8.1 adds: Option to use relative paths for file parameters
        in parameterized operators. (Needed for sharing of design files
        involving image files, sound files, etc.)
    V 0.8 adds:
        1. Adjustments to support COStar08.py, such as having
           group and subgroup designations for each operator.
           DONE: *REPRESENTATION NOW CREATED*
            1.Add automatic cascading menu if there is more
             than one group present.
            2.Add group properties to the example system in
             order to test the new feature.
              (NOTE: nothing is done yet with subgroup and subsubgroup features)
             
        2. Node display handled in a way that delegates to
           a method that receives a canvas and a rectangle.
           DONE
           
        3. Experimental version of parameterized operators of
           these types:
           A. file name+path (operator itself can contain a "filter"
              that limits the folders and the extensions of files
              to choose from.
           B. Integers, normally used to reference PRIME components
              by their IDs (Door 3, Puzzle 7, etc.)
           C. Free text.

             a. A dialog-op capability for operators,
                subclassing Operator.
                FACILITY TO FURTHER INSTANTIATE THESE PARAMETERIZED OPERATORS AT
                PARAMETER DIALOG TIME.
                   WE CREATE A NEW OPERATOR
                   INSTANCE AT DIALOG TIME. THIS OPERATOR is
                   an instance of class OperatorSpecifiedWithDialog, but it is
                   not put on any menus, and it will only be referenced
                   in two places: (a) each state's "last operator" field (DONE),
                   and (b) in tree files to explain how states are generated. (TO DO)
                   A special format will be required for this kind of operator,
                   including ID code for the OperatorSpecifiedWithDialog instance,
                   and the parameter value.
                
             b. Developed a test that creates 3 new parameterized operators
                that put the data from dialogs into the state data
                as a "string" or int.
             c. NOT DONE (too app specific): Add another operator that can be 
                used to edit the existing data.  It's like the first, but it
                primes the buffer with the existing data.

        Other recent work: adjustments to the tree-file writing and reading
        to accommodate the new kind of operators.  Optional time-stamping of states.
        Use of a global variable PROPERTIES for application-specific info that should
        be written and read in tree files.

        Implemented "path extraction" as one way to clean up design explorations.
         However, using it may go against the spirit of TStar if people don't save
         their full trees before using this.
        Implemented "selection" which is separate from "focus".  A selection can hold
         an arbitrary number of nodes.  So far, there is only one planned use of
         selection -- to select two nodes for path merging.  But there are many
         other possibilities, such as setting attributes on all selected nodes at once.
         Selections could also work with the annotation node-selection mechanism.

        Implemented tree combining via the File menu.  This works. However, it
        duplicates states -- instead of reusing an existing state, it queries the
        user to reapply the operator. If user clicks Cancel, the whole operation
        stops.  So... Try to fix this by reusing an existing state instead of
        creating the new one. [And later on, find a way to merge annotations for them.]
          There seems to be an issue with state numbering. When loading in the second
        tree on top of an existing one, the old state numbers from the file are used,
        resulting in two State-1's, for example.  Probably it would be best to reassign
        state numbers to the overlaid nodes.  But (a) if the state is a duplicate of
        an existing state, it should not be created, but its annotations mapped to
        the state from the tree in memory.

        WORKING.
        
        To Do:
        1. have the code in open_tree_file, handle merging trees more thoroughly by
            creating a hash that maps from old state numbers to new state numbers
            for the tree that is being added.  Then COStar.py will be able to
            translate state numbers to put the annotations on the right states.

        2. Means to "Open" a node for a more detailed level of
           display and interaction.
           MENU ITEM and DEFAULT METHOD (no-op) ARE THERE
        3. Low-level logging hooks with upgrade path to INFACT logging.
           ADD LATER.
    

    PURPOSE OF THIS MODULE:
        
    This module supports what we can call "interactive state
    space search".  Whereas traditional search algorithms in the
    context of artificial intelligence work completely automatically,
    this module lets the user make the moves.  It provides support
    to the user in terms of computing new states, displaying that
    portion of the state space that the user has embodied, and
    providing controls to permit the user to adapt the presentation
    to his or her needs.  This type of tool could ultimately be a
    powerful problem solving tool, useful in several different
    modes of use: interactive processing of individual objects,
    programming by demonstration (the path from the root to any
    other node in the state space represents a way of processing
    any object similar in structure to that of the root object.)
    By showing any number of states simultaneously, the user has
    the ability to compare, relatively easily, a large number of
    alternative versions (or derived states), and so it becomes
    convenient to compare large numbers of variations on processing
    sequences.

    The default domain here is simple (some trivial arithmetic)

    To use this code for a different application domain, certain
    methods and values need to be redefined. This can all be done
    in another file.  For an example of the use of T* in simple
    image processing, see the file TRAIPSE01.py.
      

    """
from Tkinter import *

# Note: It is very important that ImageTk be imported AFTER Tkinter
# Otherwise the constructor ImageTk.PhotoImage will not work.
from math import log

# Create some global variables:
TK_ROOT = None; SP_CANVAS = None; TREE_DISP = None
X_SCROLLBAR = None; Y_SCROLLBAR = None
COMMANDS_POPUP = None; COMMANDS_POPUP_FRAME = None
OPS_POPUP = None; OPS_POPUP_FRAME = None; LAST_EVENT=None
RESIZE_POPUP = None; RESIZE_POPUP_FRAME = None
VISIBILITY_POPUP = None; VISIBILITY_POPUP_FRAME = None
INITIAL_STATE=None; ALL_STATES=[]
OPERATORS=[]
PRODUCED_OBJECT=None
TINY_SIZE = 6
ZOOM_FACTOR = 1.0
VIEW = "Canonical" # Can be set for specific user roles by applications.
AUTO_FOCUS_LAST_NODE = True
RELATIVE_PATHS = True # new in Version 0.8.1

# The following 5 globals should normally be modified within
# an importing application of TStar:
TITLE=" T* Transparent STate-space search ARchitecture (Version 0-8)" # "." not allowed by Tkinter
HEADLINE = "###TStar Version 09 Tree Description File"
TREE_FILE_TYPE="tst"
ROOT_DATA_FILE_TYPE="int"
ROOT_DATA_TYPE_DESCRIPTION="Files containing one integer"
PROPERTIES = {} # Used by applications to read and set items that should be saved
# and read from tree files.  T-Star will save and restore this hash to and from
# tree files.
USE_ABSOLUTE_OPERATOR_NOS = False
PROPERTIES["USE_ABSOLUTE_OPERATOR_NOS"] = USE_ABSOLUTE_OPERATOR_NOS
PREFER_ABSOLUTE_OPERATORS = False # Can be overridden by an app.

USER="Unknown" # Used in the headers of any tree files that are saved.

ADVANCED = True # Provide all menus and items
TIMESTAMPS = True # If true, put a timestamp on each state, for use in path merging.
AUTOFOCUS_BooleanVar = None
DIRTY = False # Becomes true when edits are made and false when saves are done.
FIRST_NEW_STATE_NUM = 0

# Info. for possible undoing of last operator(s) applic.
UNDO_ENABLED = True

def setup_and_run():
  global TK_ROOT, SP_CANVAS, TREE_DISP, INITIAL_STATE,\
         X_SCROLLBAR, Y_SCROLLBAR
  TK_ROOT = Tk(className=TITLE)
  TK_ROOT.grid_rowconfigure(0, weight=1)
  TK_ROOT.grid_columnconfigure(0, weight=1)

  # add scrollbars
  X_SCROLLBAR = Scrollbar(TK_ROOT, orient=HORIZONTAL)
  Y_SCROLLBAR = Scrollbar(TK_ROOT)
  X_SCROLLBAR.grid(row=1, column=0, sticky='ew')
  Y_SCROLLBAR.grid(row=0, column=1, sticky='ns')

  # Set up the canvas for drawing the tree diagrams
  SP_CANVAS = Canvas(TK_ROOT, xscrollcommand=X_SCROLLBAR.set,
                     yscrollcommand=Y_SCROLLBAR.set, width=550, height=600)
  SP_CANVAS.grid(row=0,column=0,sticky='nesw')
  SP_CANVAS.bind("<Button-1>", left_click)
  SP_CANVAS.bind("<Shift-Button-1>", left_shift_click)
  SP_CANVAS.bind("<Button-3>", right_click)
  X_SCROLLBAR.config(command=SP_CANVAS.xview)
  Y_SCROLLBAR.config(command=SP_CANVAS.yview)
  X_SCROLLBAR.set(0.1, 0.5)
  Y_SCROLLBAR.set(0.1, 0.5)
  global AUTOFOCUS_BooleanVar, AUTO_FOCUS_LAST_NODE
  AUTOFOCUS_BooleanVar = BooleanVar()
  AUTOFOCUS_BooleanVar.set(AUTO_FOCUS_LAST_NODE)

  create_menus()
  create_commands_popup() # Used when the user right-clicks on a node.
  create_initial_state()  # Create the initial state and display it.
  initialize_states_and_display()
  TK_ROOT.mainloop()

def create_menus():
  global ADVANCED, UNDO_ENABLED
  menu_bar = Menu(TK_ROOT)
  TK_ROOT.config(menu=menu_bar)
  fileMenu = Menu(menu_bar)
  menu_bar.add_cascade(label="File", menu=fileMenu)
  fileMenu.add_command(label="New", command=new_tree)
  fileMenu.add_command(label="Open tree file...", command=open_tree_file)
  fileMenu.add_command(label="Save tree As...", command=save_tree_file)
  fileMenu.add_command(label="Open and combine into tree...", command=union_from_file)
  fileMenu.add_separator()
  fileMenu.add_command(label="Exit", command=exit_TStar)

  stateMenu = Menu(menu_bar)
  menu_bar.add_cascade(label="Tree", menu=stateMenu)
  if UNDO_ENABLED: stateMenu.add_command(label="Undo", command=undo)
  if ADVANCED: stateMenu.add_command(label="Load initial state data...", command=load)
  stateMenu.add_checkbutton(label="Automatically put focus on newly created nodes", command=autofocus, variable=AUTOFOCUS_BooleanVar)
  stateMenu.add_command(label="Scroll now to focused node", command=scroll_to_focused)
  stateMenu.add_command(label="Scroll to root", command=scroll_to_root)
  stateMenu.add_command(label="Clear selection", command=clear_selection)
  stateMenu.add_command(label="Extract path", command=extract_path)
  stateMenu.add_command(label="Merge ancestor paths for selected nodes", command=merge_paths)
  if ADVANCED: stateMenu.add_command(label="Generate subtree of depth 2", command=gen_subtree_depth2)
  if ADVANCED: stateMenu.add_command(label="Generate subtree of depth 3", command=gen_subtree_depth3)

  viewMenu = Menu(menu_bar)
  menu_bar.add_cascade(label="View", menu=viewMenu)
  viewMenu.add_command(label="Zoom_Factor = 0.1", command=zoom01)
  viewMenu.add_command(label="Zoom_Factor = 0.2", command=zoom02)
  viewMenu.add_command(label="Zoom_Factor = 0.5", command=zoom05)
  viewMenu.add_command(label="Zoom_Factor = 1.0", command=zoom10)
  viewMenu.add_command(label="Zoom_Factor = 1.5", command=zoom15)

  add_app_specific_menus(menu_bar)
  
  helpMenu = Menu(menu_bar)
  menu_bar.add_cascade(label="Help", menu=helpMenu)
  helpMenu.add_command(label="Instructions", command=show_instructions)
  helpMenu.add_command(label="About", command=show_about)

def exit_TStar():
  print "Terminating T-Star session."
  log("Exiting")
  TK_ROOT.quit()
  TK_ROOT.destroy()
  
def add_app_specific_menus(menubar): pass # Redefine this in your application.

def new_tree():
  log("new_tree")
  global DIRTY, INITIAL_STATE
  if DIRTY:
    import tkMessageBox
    result = tkMessageBox.askokcancel("Abandon changes?", "You will lose your work on the current tree.")
    if not result: return
  create_initial_state()
  initialize_states_and_display()
  current_state = INITIAL_STATE
  redraw_tree()
  DIRTY = False

def union_from_file():
  log("union_from_file")
  open_tree_file(remove_existing=False)

FIRST_NEW_STATE_NUM = None    
DEBUG_TREE_FILE = False
LAST_PARAM_NAME = None # Used in debugging tree file problems.
SUPPRESS_TIMESTAMPING = False
OLD_STATE_NEW_STATE_MAPPING = {} # Used when states need to be renumbered,
 # esp. by COStar code that loads annotations when merging trees.

def open_tree_file(remove_existing=True, desc_string=None):
  log("open_tree_file")
  global SUPPRESS_TIMESTAMPING, DEBUG_TREE_FILE, LAST_PARAM_NAME, DIRTY
  global OLD_STATE_NEW_STATE_MAPPING
  import tkFileDialog
  if DIRTY:
    import tkMessageBox
    result = tkMessageBox.askokcancel("Abandon changes?", "You will lose your work on the current tree.")
    if not result: return
  if not desc_string:  
    FILE_NAME = tkFileDialog.askopenfilename(
      filetypes=[("T-Star Tree Files", TREE_FILE_TYPE)])
    if FILE_NAME=='': return
    try:
      log("Trying to open: "+FILE_NAME)
      f = open(FILE_NAME, "r")
      print "Reading TStar tree file: "+FILE_NAME
      temp_headline = f.readline()[:-1]
      if temp_headline[:16] != HEADLINE[:16]:
          print "This file is incompatible with this version of TStar"
          return None
      temp_body = f.read()
      f.close()
    except:
      log("Could not open "+FILE_NAME)
      print "Could not load new data from file: ", FILE_NAME
  else:
    temp_body = desc_string
  OLD_STATE_NEW_STATE_MAPPING = {} 
  global FIRST_NEW_STATE_NUM
  if remove_existing:
    State.number = 0
    create_initial_state()
    initialize_states_and_display()
    FIRST_NEW_STATE_NUM = 0
  else:
    FIRST_NEW_STATE_NUM = State.number + 1
  current_state = INITIAL_STATE
  SUPPRESS_TIMESTAMPING = True
  tree_file_pre_operation() # No-op except in COStar applications (sets ROLE to ALL_ROLES)
  try:
    import string
    line_num = 0
    global PROPERTIES, USE_ABSOLUTE_OPERATOR_NOS
    USE_ABSOLUTE_OPERATOR_NOS = False
    for line in string.split(temp_body, "\n"):
      line_num += 1
      if DEBUG_TREE_FILE: print "Processing line "+str(line_num)+": "+line
      if line_num < 6 and line.startswith("###PROPERTIES="):
          NEW_PROPERTIES = eval(line[14:]) # should set PROPERTIES
          print "NEW_PROPERTIES is: "+str(NEW_PROPERTIES)
          try: 
            if NEW_PROPERTIES["USE_ABSOLUTE_OPERATOR_NOS"]:
              USE_ABSOLUTE_OPERATOR_NOS = True
              print "USE_ABSOLUTE_OPERATOR_NOS is True. ---------------------"
          except: pass
          if remove_existing:
            PROPERTIES = NEW_PROPERTIES
      if len(line)==0 or line[0]=="#": continue
      if line[0]=="v":
          line = line[2:]
          opidx = get_int_param(line,"op")
          applicables = []
          if not USE_ABSOLUTE_OPERATOR_NOS:
            applicables = applicable_ops(current_state)
          if DEBUG_TREE_FILE:
            print "current_state: "+str(current_state)
            print "opidx="+str(opidx)
            print "applicables="+str(map(lambda op:op.name, applicables))
          try:
            if USE_ABSOLUTE_OPERATOR_NOS:
              op = OPERATORS[opidx]
            else:
              op = applicables[opidx]
          except:
            print "ERROR -- Specified operator index out of range of applicables."
            print "While reading file: opidx = "+str(opidx)
            print "  and the applicables are "+str(map(lambda op:op.name, applicables))
            raise
          ps = get_long_param_string(line)
          if isinstance(op, OperatorSpecifiedWithDialog):
            #print "Parameter string is: "+ps
            newop = op.clone_with_encapsulated_param(ps)
            op = newop
          current_state = TREE_DISP.try_to_create_new_child(current_state, op, ps)
          current_state.param_string = ps
          continue
      if line[0]=="^":
          if current_state.p:
            current_state = current_state.p
          continue
      try:
        old_state_num = get_int_param(line,"N")
        if remove_existing:
          current_state.n = old_state_num
          State.number = max(State.number, old_state_num)
        else:
          OLD_STATE_NEW_STATE_MAPPING[old_state_num]=current_state.n
        h = get_int_param(line,"G")
        if h: current_state.hidden = True
        current_state.display_version = get_int_param(line,"D")
        current_state.renderingMode = get_int_param(line, "R")
        current_state.mainChild = get_int_param(line, "M")
        current_state.importance = get_int_param(line, "I")
        current_state.horizBuffer = get_int_param(line, "B")
        current_state.subtreeWidth = get_int_param(line, "W")
        current_state.subtreeHeight = get_int_param(line, "H")
        current_state.lineWidth = get_int_param(line, "L")
        current_state.color = get_int_param(line, "C")
        current_state.triangle = get_int_param(line, "T")
        if TIMESTAMPS:
          current_state.timestamp = get_string_param(line, "t")
      except:
        print "could not find rendering spec ("+LAST_PARAM_NAME+") for state in the line "+str(line_num)+":" + line
      try:
        current_state.scale = get_float_param(line,"S")
      except:
        current_state.scale = 1.0
  except:
    print "Error while reading tree file in line "+str(line_num)+": " + line
    raise
  tree_file_post_operation() # No-op except in COStar applications (restores ROLE)
  SUPPRESS_TIMESTAMPING = False
  redraw_tree()
  if (desc_string==None): app_specific_tree_load(FILE_NAME)
  if remove_existing: DIRTY=False
  return temp_body

def is_new_state(s):
  global FIRST_NEW_STATE_NUM
  return s.n >= FIRST_NEW_STATE_NUM

def tree_file_pre_operation(): pass
def tree_file_post_operation(): pass

def undo():
  global UNDO_ENABLED
  if not UNDO_ENABLED: return
  TREE_DISP.undo()
  redraw_tree()

  
def get_int_param(line, param_name):
    """Retrieve an integer parameter in the given line of text,
    directly following the first occurrence of param_name and the
    equals sign (=)."""
    return int(get_string_param(line, param_name))
  
def get_float_param(line, param_name):
    """Retrieve a floating point parameter in the given line of text,
    directly following the first occurrence of param_name and the
    equals sign (=)."""
    return float(get_string_param(line, param_name))
  
def get_string_param(line, param_name):
    """Retrieve a restricted string (no double quotes in it) """
    global LAST_PARAM_NAME; LAST_PARAM_NAME = param_name
    j = line.index(param_name + "=")
    start = j + len(param_name) + 1
    end = line.index(";", start)
    strparam = line[start: end]
    return strparam

def get_long_param_string(line):
  """Retrieve the rest of the line, following "P="
  This can contain double quotes or any other text than newlines."""
  global LAST_PARAM_NAME; LAST_PARAM_NAME = "P"
  j = line.index(";P=")        
  ps = line[j+3:]
  return ps

def save_tree_file():
  """Brings up a file dialog box, so the user can choose a
  file for saving the current tree configuration. Then saves it."""
  global TK_ROOT, USER, USE_ABSOLUTE_OPERATOR_NOS, PREFER_ABSOLUTE_OPERATORS
  import tkFileDialog
  savefile = tkFileDialog.asksaveasfilename(
    filetypes=[("T-Star Tree Files", TREE_FILE_TYPE)],
    defaultextension=TREE_FILE_TYPE)
  log("save_tree_file: "+savefile)
  if savefile:
    try:
      USE_ABSOLUTE_OPERATOR_NOS = PREFER_ABSOLUTE_OPERATORS 
      tree_file_pre_operation() # No-op except in COStar applications (sets ROLE to ALL_ROLES)
      tree_desc_str = INITIAL_STATE.get_tree_desc_str()
      tree_file_post_operation() # (Restores ROLE)
      f = open(savefile, "w")
      f.write(HEADLINE+"\n")
      f.write("###USER="+USER+"\n")
      import time
      f.write("###TIME="+str(time.asctime())+"\n")
      f.write("###VIEW="+VIEW+"\n")
      PROPERTIES["USE_ABSOLUTE_OPERATOR_NOS"] = PREFER_ABSOLUTE_OPERATORS
      f.write("###PROPERTIES="+str(PROPERTIES)+"\n")
      f.write(tree_desc_str)
      f.close()
      app_specific_tree_save(savefile)
      global DIRTY; DIRTY = False
      log("save succeeded")
    except:
        print "Couldn't write the file: ", savefile
        log("save failed")
        raise

def app_specific_tree_save(fn): pass
def app_specific_tree_load(fn): pass

def extract_path():
  log("extract_path")
  global DIRTY, INITIAL_STATE
  sn = TREE_DISP.focus_node
  if not sn:
    import tkMessageBox
    result = tkMessageBox.askokcancel("Need focus node", "You must first focus on a node. Then retry this.")
    return
  if DIRTY:
    import tkMessageBox
    result = tkMessageBox.askokcancel("Abandon changes?", "You may lose much of your work on the current tree. Proceed anyway?")
    if not result: return
  # Find the chain of ancestors of the selected node, and then
  # (a) relink these to avoid links to all others, and
  # (b) remove all others.
  #result = tkMessageBox.askokcancel("Not yet implemented", "Not yet implemented.")
  path = get_ancestors(sn.s)
  print "Path is: "+str(path)
  # Do relinking:
  this_child = None
  for s in path:
    s.first_child = this_child
    s.right_sibling = None
    this_child = s
  # Maybe delete other states and nodes ??
  # Let's assume for now that they are simply garbage-collected.
  # Note that annotations might be kept around in a hash.
  # Someday it might be good to provide code to explicitly delete all
  # objects that are parts of or that correspond to deleted states.
  redraw_tree()

def get_ancestors(s):
    path = [s]    
    done=False
    while not done:
      parent = s.p
      if parent==None: done=True
      else: path.append(parent); s=parent
    return path
  
def merge_paths():
  log("merge_paths")
  global DIRTY, INITIAL_STATE
  sel = TREE_DISP.selection
  if len(sel)<2:
    import tkMessageBox
    result = tkMessageBox.askokcancel("You must first select at least two nodes. Then retry this.")
    return
  if DIRTY:
    import tkMessageBox
    result = tkMessageBox.askokcancel("Abandon changes?", "You may lose much of your work on the current tree. Proceed anyway?")
    if not result: return
  #print "First node is "+str(sel[0])
  #print "Second node is "+str(sel[1])
  print "Preparing to build combined path using "+str(len(sel))+" selected nodes."
  overall_path = []
  for node in sel:
    path = get_ancestors(node.s)
    path.reverse()
    overall_path += path
  #path1 = get_ancestors(sel[0].s)
  #path2 = get_ancestors(sel[1].s)
  #print "The two paths are: "+str(path1)+' and \n'+str(path2)
  # Create merged list:
  #concatenated = path1 + path2
  #overall_path.sort(key=lambda s: s.timestamp)

  #print "The merged list timestamps are: "+str(map(lambda s:s.timestamp, overall_path))
  # Now build new path, in which the operators used in the old paths are applied in
  # the new order.  If at any point an operator is no longer applicable to its starting
  # stop the process there and add the new path at its current length.
  root = overall_path[0]
  print "Number of nodes in merged list: "+str(len(overall_path))
  current = root
  k=0
  for s in overall_path:
    if s==root: continue
    if s.n==0: continue # added to eliminate bug that occurred during some merges.
    op = s.last_operator
    if op==None:
      print "last_operator field is None for state "+str(s.n)
    #print "This operator is "+str(op)
    #if op==None: continue
    # Try to apply op to current 
    current = TREE_DISP.try_to_create_new_child(current, op)
    if not current:
      print "Path terminates early because of a conflict between state and next precondition."
      print "Operator name is: "+op.name
      if is_instance(op, OperatorSpecifiedByDialog):
        print "  Parameter: "+op.param
      break
    k += 1
  print "Number of nodes in new path: "+str(k)
  redraw_tree()
    
     
    
  
def clear_selection():
  TREE_DISP.selection = []
  redraw_tree()

def show_instructions():
  log("show_instructions")
  import tkMessageBox
  tkMessageBox.showinfo("Help","Instructions:\n\n"
        "Right-clicking on a node opens a pop-up menu of node options.\n"
        "Left-clicking re-applies the most recent command to the node clicked on.\n\n"
        "T-Star's main purpose is to facilitate user exploration of a space "
        'of possible "states" of a design or problem. '
        "It keeps track of what states have been explored, and it displays this "
        "information in a flexible way on the screen. "
        "A tree can be saved to or loaded from a disk file, in order to "
        "checkpoint a session or to prepare a demonstration.")

def show_about():
  log("show_about")
  import tkMessageBox
  tkMessageBox.showinfo("About","TStar Version 0.8\nCopyright 2007 Steven Tanimoto\n"
        "University of Washington, Seattle, WA\n\n"
        "T-Star is an experimental design and problem-solving tool.")

def initialize_states_and_display():
    global ALL_STATES, TREE_DISP
    ALL_STATES = [INITIAL_STATE]
    initial_node = NODE_CLASS(INITIAL_STATE)
    TREE_DISP = tree_display(SP_CANVAS,NODE_CLASS)
    TREE_DISP.state_node_hash = {}
    TREE_DISP.state_node_hash[INITIAL_STATE] = initial_node
    TREE_DISP.all_nodes = [initial_node]
    redraw_tree()

def left_click(event):
  """If the user clicked on a node,
     make it the focus node, and perform the
     current command on it, if legal."""
  global LAST_EVENT
  LAST_EVENT = event
  x = SP_CANVAS.canvasx(event.x)
  y = SP_CANVAS.canvasy(event.y)
  for nd in TREE_DISP.all_nodes:
      if nd.clicked_on(x,y):
          update_node_focus(nd)
          log("left_click -- updated node focus ")
          #perform_last_command()
          apply_one_op()
          break

def left_shift_click(event):
  """If the user clicked on a node,
     make it the focus node, and perform the
     current command on it, if legal."""
  global LAST_EVENT
  LAST_EVENT = event
  x = SP_CANVAS.canvasx(event.x)
  y = SP_CANVAS.canvasy(event.y)
  for nd in TREE_DISP.all_nodes:
      if nd.clicked_on(x,y):
          update_node_focus(nd)
          log("left_shift_click")
          perform_last_command()
          break

def autofocus():
  global AUTO_FOCUS_LAST_NODE, AUTOFOCUS_BooleanVar
  AUTO_FOCUS_LAST_NODE = not AUTO_FOCUS_LAST_NODE
  AUTOFOCUS_BooleanVar.set(AUTO_FOCUS_LAST_NODE)
  
def update_node_focus(nd):
  """Modify the display to highlight the correct
     focus node."""
  sn = TREE_DISP.focus_node
  if sn:
    sn.unfocus()
    if (not sn.s.hidden) and (sn.s.renderingMode!=1):
      sn.unpaint()
      sn.paint()
  TREE_DISP.focus_node = nd
  nd.focus()
  nd.unpaint()
  nd.paint()

def right_click(event):
  """If the user right-clicked on a node, focus on it,
     and bring up the popup menu"""
  x = SP_CANVAS.canvasx(event.x)
  y = SP_CANVAS.canvasy(event.y)
  global LAST_EVENT
  LAST_EVENT = event
  log("right_click")
  for nd in TREE_DISP.all_nodes:
      if nd.clicked_on(x,y):
          update_node_focus(nd)
          show_popup_menu(COMMANDS_POPUP,event)
          break

def create_commands_popup():
  global COMMANDS_POPUP, COMMANDS_POPUP_FRAME, TK_ROOT, ADVANCED
  COMMANDS_POPUP_FRAME = Frame(TK_ROOT, width=200, height=200)
  COMMANDS_POPUP = Menu(COMMANDS_POPUP_FRAME, tearoff=0)
  P = COMMANDS_POPUP
  if ADVANCED: P.add_command(label="Open detailed view", command=open_detailed_view)
  P.add_command(label="Apply one operator...", command=apply_one_op)
  P.add_command(label="Add to/Remove from selection", command=toggle_in_selection)
  if ADVANCED: P.add_command(label="Grow all children", command=make_children)
  if ADVANCED: P.add_command(label="Resize node's display...", command=resize_menu)
  if ADVANCED: P.add_command(label="Visibility...", command=visibility_menu)
  #P.add_command(label="Make Tiny", command=make_tiny)
  #P.add_command(label="Make Mini", command=make_mini)
  #P.add_command(label="Make Normal", command=make_normal)
  #P.add_command(label="Make Descendants Tiny", command=make_descendants_tiny)
  #P.add_command(label="Make Descendants Mini", command=make_descendants_mini)
  #P.add_command(label="Make Descendants Normal", command=make_descendants_normal)
  #P.add_command(label="Make Ancestors Normal", command=make_ancestors_normal)
  #P.add_command(label="Hide Subtree", command=hide_subtree)
  #P.add_command(label="Show Details", command=show_details)
  #P.add_command(label="Unhide Children", command=unhide_children)   

LAST_EVENT = None
def show_popup_menu(pm, event):
    global LAST_EVENT
    LAST_EVENT = event
    pm.post(event.x, event.y)

def open_detailed_view():
    print "No detailed view defined for this search problem."
    
class temp_op_callable:
    """Objects in this class are used as temporary functions,
       similar to closures in functional languages. They are used
       as the 'command' argument to entries in the popup menu for
       applicable operators.  This list is regenerated each time
       the user wants to apply a single operator to the focus
       state."""
    def __init__(self, op, s):
        self.op = op; self.s = s
    def __call__(self):
        # If this operator requires further specification by the user,
        # bring up the dialog now.
        #print "Applying operator: "+str(self.op)
        #print "Operator type is: "+str(type(self.op))
        if isinstance(self.op, OperatorSpecifiedWithDialog):
          #print "Now show dialog..."
          successful = self.op.show_dialog(self.s)
          if not successful: return 
          # Now create a new instance of the operator that encapsulates
          # the user-supplied parameter, and use that operator
          # in generating the successor.
          op_to_use = self.op.clone_with_encapsulated_param(self.op.param)
          log("User parameter: "+str(self.op.param))
          #print "Created new instance of OperatorSpecifiedWithDialog to encapsulate the data:"
          #print param
          #print "This operator is: "+str(op_to_use)
          #print "The static, mother operator is: "+str(self.op)
        else: op_to_use = self.op
        # Now go ahead and apply the operator to get the new state.
        log("Applying operator: "+op_to_use.name) 
        if TREE_DISP.try_to_create_new_child(self.s, op_to_use):
            redraw_tree()
            global DIRTY; DIRTY = True
            global LAST_NODE
            if AUTO_FOCUS_LAST_NODE and LAST_NODE!=None:
              update_node_focus(LAST_NODE)
              scroll_to_focused()
        else: print "No new state created."

def scroll_to_focused():  # Brings the focus node on-screen, usually near the middle. 
    sn = TREE_DISP.focus_node
    scroll_to_node(sn)

def scroll_to_node(sn):
    if sn!=None:
      sx = sn.x; sy = sn.y
      w = sn.s.space_needed()
      rect = SP_CANVAS.bbox(ALL)
      maxx = rect[2]; maxy = rect[3]
      window_width = TK_ROOT.winfo_width(); window_height = TK_ROOT.winfo_height()
      xfrac = (ZOOM_FACTOR*sx - window_width/2 - w)  / (0.0+maxx)
      yfrac = (ZOOM_FACTOR*sy - window_height/2) / (0.0+maxy)
      SP_CANVAS.xview_moveto(xfrac); SP_CANVAS.yview_moveto(yfrac)

def scroll_to_root():
    global INITIAL_STATE
    root_node = TREE_DISP.state_node_hash[INITIAL_STATE]
    scroll_to_node(root_node )

def toggle_in_selection():
    log("toggle_in_selection")
    set_last_command(toggle_in_selection)
    sn = TREE_DISP.focus_node
    sel = TREE_DISP.selection
    if sn in sel: sel.remove(sn)
    else: sel.append(sn)
    redraw_tree()


POPUPS_FOR_GROUPS = None        
def create_operators_popup():
  global OPS_POPUP, OPS_POPUP_FRAME, TK_ROOT, POPUPS_FOR_GROUPS
  OPS_POPUP_FRAME = Frame(TK_ROOT, width=200, height=200)
  OPS_POPUP = Menu(OPS_POPUP_FRAME, tearoff=0)
  count = 0; temp_func = []
  applicables = applicable_ops(TREE_DISP.focus_node.s)
  groupsize = {}
  ngroups = 0
  grouplist = []
  POPUPS_FOR_GROUPS = {}
  # Compute the numbers of operators in each group,
  # and make a list of the different groups.
  for op in applicables:
      #print "In create_operators_popup, group of op is "+op.group
      try:
        s = groupsize[op.group]
        groupsize[op.group] = s+1
      except:
        ngroups += 1
        groupsize[op.group] = 1
        grouplist.append(op.group)
  #print "ngroups = " + str(ngroups)
  # If there is only one group (or none), create a popup menu
  # of the operators without showing the group name.
  if ngroups < 2:    
      for op in applicables:
          temp = temp_op_callable(op, TREE_DISP.focus_node.s)
          OPS_POPUP.add_command(label=op.name, command=temp )
  else:
      for g in grouplist:
          #s = groupsize[g]
          #print "For group "+g+", size = "+str(s)
          #if s==1:
          #    temp = temp_op_callable(op, TREE_DISP.focus_node.s)
          #    OPS_POPUP.add_command(label=op.name, command=temp )
          #else:
          pfg = create_popup_for_group(g, applicables)
          POPUPS_FOR_GROUPS[g] = pfg
          temp = temp_group_callable(g)
          OPS_POPUP.add_command(label=g, command=temp )

GRP_POPUP_FRAME = None
def create_popup_for_group(grp, ops):
  global GRP_POPUP_FRAME, TK_ROOT
  GRP_POPUP_FRAME = Frame(TK_ROOT, width=200, height=200)
  grp_popup = Menu(GRP_POPUP_FRAME, tearoff=0)
  for op in ops:
      if op.group != grp: continue
      temp = temp_op_callable(op, TREE_DISP.focus_node.s)
      grp_popup.add_command(label=op.name, command=temp )
  return grp_popup

class temp_group_callable:
    """These objects serve as temporary functions that are used
    as callbacks to display popup menus with groups of operators."""
    def __init__(self, grp):
        self.g = grp
    def __call__(self):
        global POPUPS_FOR_GROUPS
        pfg = POPUPS_FOR_GROUPS[self.g]
        x = LAST_EVENT.x + 30
        y = LAST_EVENT.y + 20
        pfg.post(x, y) # use event.x and event.y from the
        # most recent mouse click event, offset slightly,
        # in order to give the "cascading popup" effect.
  
def perform_last_command():
    """This is invoked when the user clicks the left mouse button
       on a node."""
    last_command()

def set_last_command(the_last_command):
    global last_command
    last_command = the_last_command

def redraw_tree():
    #TREE_DISP.draw_tree_inorder(INITIAL_STATE)
    TREE_DISP.draw_tree_using_layout_params(INITIAL_STATE)
    sn = TREE_DISP.focus_node
    SP_CANVAS.config(scrollregion=SP_CANVAS.bbox(ALL))
    scroll_to_focused()
    
def make_children():
    """Generate the children of the focus node. If any
       children already exist, only generate those children
       using other operators than those used to create the
       existing children."""
    set_last_command(make_children)
    sn = TREE_DISP.focus_node
    TREE_DISP.grow_successors_and_their_nodes(sn)
    redraw_tree()
    global DIRTY; DIRTY = True
      
def apply_one_op():
    """Populate a popup menu with the names of currently applicable
       operators, and let the user choose which one to apply."""
    log("apply_one_op")
    set_last_command(apply_one_op)
    currently_applicable_ops = applicable_ops(TREE_DISP.focus_node.s)
    #print "Applicable operators: ",\
    #    map(lambda o: o.name, currently_applicable_ops)
    global OPS_POPUP
    create_operators_popup()
    show_popup_menu(OPS_POPUP, LAST_EVENT)

#last_command = make_children # Default command for left mouse button.
last_command = apply_one_op # Default command for left mouse button.

def resize_menu():
    set_last_command(resize_menu)
    global RESIZE_POPUP
    if RESIZE_POPUP==None: create_resize_popup()
    show_popup_menu(RESIZE_POPUP, LAST_EVENT)

def visibility_menu():
    set_last_command(visibility_menu)
    global VISIBILITY_POPUP
    if VISIBILITY_POPUP==None: create_visibility_popup()
    show_popup_menu(VISIBILITY_POPUP, LAST_EVENT)

def create_resize_popup():
  global RESIZE_POPUP, RESIZE_POPUP_FRAME, TK_ROOT
  RESIZE_POPUP_FRAME = Frame(TK_ROOT, width=200, height=200)
  RESIZE_POPUP = Menu(RESIZE_POPUP_FRAME, tearoff=0)
  count = 0; temp_func = []
  all_sizing_ops = [['Make Tiny',make_tiny],
             ['Make Mini',make_mini],
             ['Make Normal',make_normal],
             ['Make Descendants Tiny', make_descendants_tiny],
             ['Make Descendants Mini',make_descendants_mini],
             ['Make Descendants Normal',make_descendants_normal],
             ['Make Ancestors Normal', make_ancestors_normal],
             ["Set node's scale factor",set_node_scale],
             ["Set subtree's scale factor",set_subtree_scale]]
  if ADVANCED: ops_to_use = all_sizing_ops
  else: ops_to_use = all_sizing_ops[0,2,3,5]
  for op in ops_to_use:
      RESIZE_POPUP.add_command(label=op[0], command=op[1])
  
def create_visibility_popup():
  global VISIBILITY_POPUP, VISIBILITY_POPUP_FRAME, TK_ROOT
  VISIBILITY_POPUP_FRAME = Frame(TK_ROOT, width=200, height=200)
  VISIBILITY_POPUP = Menu(VISIBILITY_POPUP_FRAME, tearoff=0)
  count = 0; temp_func = []
  for op in [['Hide Subtree',hide_subtree], ['Unhide Children',unhide_children],
             ['Put node into Ellipsis',enter_ellipsis],
             ['Shrink Ellipsis to Parent',shrink_ellipsis_to_parent],
             ['Take Parent Out of Ellipsis',parent_out_of_ellipsis],
             ['Take Children Out of Ellipsis',children_out_of_ellipsis],
             ['Show Details',show_details]]:
      VISIBILITY_POPUP.add_command(label=op[0], command=op[1])

def enter_ellipsis():
    log("enter_ellipsis")
    set_last_command(enter_ellipsis)
    if TREE_DISP.focus_node.s == INITIAL_STATE:
        print "Cannot put the root into ellipsis,",\
              "because there would be no way to bring it out again."
    else:
      TREE_DISP.focus_node.s.renderingMode = 1
      these_children = children(TREE_DISP.focus_node.s)
      map(lambda s: s.end_ellipsis(), these_children)
      redraw_tree()
      global DIRTY; DIRTY = True

def parent_out_of_ellipsis():
    log("parent_out_of_ellipsis")
    global DIRTY
    set_last_command(parent_out_of_ellipsis)
    s = TREE_DISP.focus_node.s
    #print "Entering parent_out_of_ellipsis"
    if s.renderingMode!=2:
      print "Current node does not terminate an ellipsis.";
      return
    parent = s.p
    if parent==None:
      print "Node has no parent to take out of ellipsis.";
      return
    if parent.renderingMode!=1:
      print "Parent was not in ellipsis.  (Strange -- should not happen)";
      return
    s.renderingMode=0 ; #print "current node is now normal."
    gp = parent.p
    if gp==None:
      print "No grandparent found."
      parent.renderingMode=0; redraw_tree(); return
    if gp.renderingMode==1:
      parent.renderingMode=2 # parent now terminates the ellipsis
      #print "Parent now terminates the ellipsis."
      redraw_tree()
      DIRTY = True
      return
    if gp.renderingMode==0:
      parent.renderingMode=0 # Ellipsis is gone
      #print "Ellipsis is now gone"
      redraw_tree()
      DIRTY = True
      return
    print "Unknown renderingMode configs in parent_out_of_ellipsis."
    print "s.renderingMode = "+str(s.renderingMode)
    print "parent.renderingMode = "+str(parent.renderingMode)
    print "grandparent.renderingMode = "+str(gp.renderingMode)
    redraw_tree()

def children_out_of_ellipsis():
    log("children_out_of_ellipsis")
    set_last_command(children_out_of_ellipsis)
    these_children = children(TREE_DISP.focus_node.s)
    #for c in these_children:
    #  print "Child is: "+str(c.n)
    map(lambda s2: s2.out_of_ellipsis(), these_children)
    redraw_tree()
    global DIRTY
    DIRTY = True

def shrink_ellipsis_to_parent():
    # This command reduces the amount of vertical space taken up by
    # the ellipsis above the focus node.
    log("shrink_ellipsis_to_parent")
    set_last_command(shrink_ellipsis_to_parent)
    s = TREE_DISP.focus_node.s
    if s.renderingMode != 2: return # Focus node must be end of an ellipsis
    parent = s.p
    rescale = 0.5
    while parent.renderingMode==1:
      parent.scale *= rescale
      parent = parent.p
    redraw_tree()
    global DIRTY
    DIRTY = True
  
def make_tiny():
    """Make the focus node as small as possible."""
    log("make_tiny")
    set_last_command(make_tiny)
    TREE_DISP.focus_node.s.display_version = 0
    redraw_tree()
    global DIRTY
    DIRTY = True

def make_mini():
    """Make the focus node relatively small."""
    log("make_mini")
    set_last_command(make_mini)
    TREE_DISP.focus_node.s.display_version = 1
    redraw_tree()

def make_normal():
    """Restore the focus node to normal size."""
    log("make_normal")
    set_last_command(make_normal)
    TREE_DISP.focus_node.s.display_version = 2
    redraw_tree()
    global DIRTY
    DIRTY = True

def make_descendants_tiny():
    """Minimize the focus node and all of its descendants."""
    log("make_descendants_tiny")
    set_last_command(make_descendants_tiny)
    descendants = preorder_collect(TREE_DISP.focus_node.s, root=True)
    map(lambda s: s.set_display_version(0), descendants)
    redraw_tree()
    global DIRTY
    DIRTY = True

def make_descendants_mini():
    """Set the focus node and all of its descendants to the middle size."""
    log("make_descendants_mini")
    set_last_command(make_descendants_mini)
    descendants = preorder_collect(TREE_DISP.focus_node.s, root=True)
    map(lambda s: s.set_display_version(1), descendants)
    redraw_tree()
    global DIRTY
    DIRTY = True

def make_descendants_normal():
    """Restore the focus node and all of its descendants
       to normal size, but do not change the hidden status
       of any nodes."""
    log("make_descendants_normal")
    set_last_command(make_descendants_normal)
    descendants = preorder_collect(TREE_DISP.focus_node.s, root=True)
    map(lambda s: s.set_display_version(2), descendants)
    redraw_tree()
    global DIRTY; DIRTY = True

def make_ancestors_normal():
    """Restore the focus node and those on the path from it to
       the root."""
    log("make_ancestors_normal")
    set_last_command(make_ancestors_normal)
    ancestors = collect_ancestors(TREE_DISP.focus_node.s)
    map(lambda s: s.set_display_version(2), ancestors)
    redraw_tree()
    global DIRTY
    DIRTY = True

SCALE_DIALOG = None
SCALE_INPUT = 1.0

def set_node_scale():
    global SCALE_DIALOG
    log("set_node_scale")
    SCALE_DIALOG = ScaleDialog(TK_ROOT,"Enter a scale factor in the range 0.01 to 5.0 for the focus node.")
    SCALE_DIALOG.e.insert(0,str(SCALE_INPUT))
    TK_ROOT.wait_window(SCALE_DIALOG.top)
    TREE_DISP.focus_node.s.scale = SCALE_INPUT
    redraw_tree()
    global DIRTY
    DIRTY = True
  
def set_subtree_scale():
    global SCALE_DIALOG
    log("set_subtree_scale")
    SCALE_DIALOG = ScaleDialog(TK_ROOT,"Enter a scale factor in the range 0.01 to 5.0 for the focus subtree.")
    SCALE_DIALOG.e.insert(0,str(SCALE_INPUT))
    TK_ROOT.wait_window(SCALE_DIALOG.top)
    set_node_scale_in_whole_subtree(TREE_DISP.focus_node.s, SCALE_INPUT)
    redraw_tree()
    global DIRTY
    DIRTY = True

def set_node_scale_in_whole_subtree(state, scale):
    state.scale = scale
    these_children = children(state)
    for c in these_children:
      set_node_scale_in_whole_subtree(c, scale)
    global DIRTY
    DIRTY = True
    
def hide_subtree():
    """Hide the focus node (provided it is not the root node).
       The node will not be displayed when its status is 'hidden'.
       Its descendants will not be displayed, either, although they
       do not necessarily have the hidden status directly."""
    log("hide_subtree")
    set_last_command(hide_subtree)
    if TREE_DISP.focus_node.s == INITIAL_STATE:
        print "Cannot hide the root,",\
              "because there would be no way to unhide it."
    else:
      TREE_DISP.focus_node.s.hidden = True
      redraw_tree()
      global DIRTY
      DIRTY = True

def unhide_children():
    log("unhide_children")
    set_last_command(unhide_children)
    these_children = children(TREE_DISP.focus_node.s)
    map(lambda s: s.unhide(), these_children)
    redraw_tree()
    global DIRTY
    DIRTY = True

# The following is a domain-dependent display generator.
def show_details():
    log("show_details")
    set_last_command(show_details)
    this_state = TREE_DISP.focus_node.s
    op_sequence = this_state.get_op_sequence()
    function_sequence = this_state.get_fn_sequence()
    op_seq_desc = map(lambda o: o.name, op_sequence)
    description = "Operator sequence for this state is: \n" + \
          str(op_seq_desc)
    top = Toplevel(TK_ROOT)
    top.grid_rowconfigure(0, weight=1)
    top.grid_columnconfigure(0, weight=1)
    temp_object = get_full_root_data()
    for fn in function_sequence:
       temp_object = fn(temp_object)
    global PRODUCED_OBJECT # make available in case user wants to save it.
    PRODUCED_OBJECT = temp_object
    detail_canvas = Canvas(top, width=600, height=500)
    detail_canvas.grid(row=0,column=0,sticky='nesw')
    info_label = Label(top, text="State type = "+this_state.desc)
    info_label.grid(row=1,column=0)
    save_button = Button(top, text="Save object as...", command=do_save)
    save_button.grid(row=2,column=0)
    description += "\n\nState data: " + str(temp_object)
    vis_info = "\n renderingMode="+str(this_state.renderingMode)+"; scale="+str(this_state.scale)+"\n\n"
    display_data_in_detail(detail_canvas, description + vis_info, op_seq_desc,
                           temp_object, this_state)
    #detail_canvas.create_image(0,0,image=new_image,anchor=NW)
    #new_image.width = new_image.width # Dummy op that makes image show up.

def load():
    log("load")
    global FILE_NAME, ROOT_DATA_FILE_TYPE, ROOT_DATA_TYPE_DESCRIPTION
    import tkFileDialog
    FILE_NAME = tkFileDialog.askopenfilename(
        filetypes=[(ROOT_DATA_TYPE_DESCRIPTION, ROOT_DATA_FILE_TYPE)])
    try:
        temp = read_data_from_file(FILE_NAME)
    except:
        print "Could not load new data from file: ", FILE_NAME
    else:
        global INITIAL_DATA
        INITIAL_DATA = build_data_object(temp)
        create_initial_state()
        initialize_states_and_display()

def do_save():
    log("do_save")
    global TK_ROOT, PRODUCED_OBJECT
    import FileDialog
    sfd = FileDialog.SaveFileDialog(TK_ROOT)
    savefile = sfd.go(key="test")
    if savefile:
        PRODUCED_OBJECT.save(savefile)
        log("filename is "+savefile)

def zoom01(): zoomTo(0.1); log("zoom01")
def zoom02(): zoomTo(0.2); log("zoom02")
def zoom05(): zoomTo(0.5); log("zoom05")
def zoom10(): zoomTo(1.0); log("zoom10")
def zoom15(): zoomTo(1.5); log("zoom15")
def zoomTo(z):
  log("zoomTo")
  global ZOOM_FACTOR
  ZOOM_FACTOR = z
  redraw_tree()

class ScaleDialog:
  def __init__(self, parent, prompt):
    top = self.top = Toplevel(parent)
    Label(top, text=prompt).pack()
    self.e = Entry(top)
    self.e.pack(padx=5)
    b=Button(top, text="OK", command=self.ok)
    b.pack(pady=5)
  def ok(self):
    global SCALE_INPUT
    try:
      SCALE_INPUT = float(self.e.get())
    except: SCALE_INPUT = 1.0
    #print "SCALE_INPUT = "+str(SCALE_INPUT)
    self.top.destroy()
    
def get_State_number(): return State.number
def set_State_number(n): State.number = n
class State(object):
    """Encapsulates the state-space element, together with
    any possible annotations, strategy or history info.
    By subclassing 'object', we force State to be a
    'new class' in Python.  This permits its methods to be
    called properly via 'super' in methods of child classes,
    such as TextListListState in TextListList.py.
    """
    number = 0
    tiny_size_factor = 0.12 # used by the size_needed method
    
    def __init__(self, data, parent, last_operator, desc=""):
        self.p = parent
        self.last_operator = last_operator # This is used to
        # keep track of how this state was obtained from its parent.
        self.first_child = None
        self.last_child = None
        self.right_sibling = None

        left_sibling = None # temporary assignment.
        # Figure out the depth of the state and its left sibling,
        # if any, using the parent's last_child reference:
        if not parent:
            self.d = 0
        else:
            self.d = parent.d + 1
            left_sibling = parent.last_child
            parent.last_child = self
        # Let the left sibling know about its new right sibling:
        if left_sibling:
            left_sibling.right_sibling = self
        else:
            if parent:
                # If it has no left sibling, then it is a first
                # child. If it also has a parent, then the parent
                # should be updated:
                parent.first_child = self
        # Assign this state a unique number:
        self.n = State.number
        State.number += 1
        self.hidden = False
        self.display_version=2 # 0=tiny, 1=miniature, 2=normal, 3=expanded.
        self.renderingMode=0 # 0=inherited, 1=stop, 2=resume
        self.mainChild=0
        self.importance=4
        self.horizBuffer=1
        self.subtreeWidth=0
        self.subtreeHeight=0
        self.lineWidth=1
        self.color=0 # 0=inherited
        self.triangle=0 # no triangle
        self.scale=1.0
        self.param_string=""
        global TIMESTAMPS, SUPPRESS_TIMESTAMPING
        if TIMESTAMPS and not SUPPRESS_TIMESTAMPING: self.register_timestamp()
    # Here is the domain dependent component of the state:
        self.data = data
        self.desc = desc
    def __str__(self):
        return "[state %d]" % self.n
    def space_needed(self):  # Horizontal space needed to display this state, in "units".
        if self.display_version==0: #return 0.12 # "tiny" # NOTE: NO LONGER USED
            return State.tiny_size_factor
        if self.display_version==1: return 0.3 # "mini"
        if self.display_version==2: return 1.0 # "normal"
        if self.display_version==3: return 2.0 # "double" (not in use as of Nov.14)
        else: return 1
    def getVSpaceNeeded(self):
        if self.display_version==0: #return 0.12 # "tiny" # NOTE: NO LONGER USED
            return State.tiny_size_factor
        if self.display_version==1: return 0.3 # "mini"
        if self.display_version==2: return 1.0 # "normal"
        if self.display_version==3: return 2.0 # "double" (not in use as of Nov.14)
        else: return 1
    def set_display_version(self,dv):
        self.display_version = dv;
    def unhide(self):
        self.hidden = False
    def get_op_sequence(self):
        if not self.p: return []
        else: return self.p.get_op_sequence() + [self.last_operator]
    def get_fn_sequence(self):
        if not self.p: return []
        else: return self.p.get_fn_sequence() + [self.last_operator.function]
    def has_hidden_children(self):
        this_child = self.first_child
        while this_child:
            if this_child.hidden: return True
            this_child = this_child.right_sibling
        return False
    def get_tree_desc_str(self):
        global USE_ABSOLUTE_OPERATOR_NOS
        USE_ABSOLUTE_OPERATOR_NOS = True
        #print "In State.get_tree_desc_str, USE_ABSOLUTE_OPERATOR_NOS = "+str(USE_ABSOLUTE_OPERATOR_NOS)
        tds = "N="+str(self.n)
        tds += ";G="
        if self.hidden: tds += '1'
        else: tds += '0'
        tds += ";D="+str(self.display_version)
        tds += ";R="+str(self.renderingMode)
        tds += ";M="+str(self.mainChild)
        tds += ";I="+str(self.importance)
        tds += ";B="+str(self.horizBuffer)
        tds += ";W="+str(self.subtreeWidth)
        tds += ";H="+str(self.subtreeHeight)
        tds += ";L="+str(self.lineWidth)
        tds += ";C="+str(self.color)
        tds += ";T="+str(self.triangle)
        tds += ";S="+str(self.scale)+";"
        if TIMESTAMPS:
          try:
            tds += "t="+self.timestamp+";"
          except: pass
        this_child = self.first_child
        while this_child:
            applicables = applicable_ops(self)
            #print "applicable ops for state "+str(self)+" are:", applicables
            
            op = this_child.last_operator
            if isinstance(op, OperatorSpecifiedWithDialog):
              PS = op.parameter
            else:
              PS = this_child.param_string  # Note: original field for params is not being used (yet)
            if USE_ABSOLUTE_OPERATOR_NOS:
              i = get_absolute_operator_index(this_child.last_operator) 
            else:
              i = get_operator_index_from_applicables(applicables, this_child.last_operator)
            tds += "\nv:" + "op=" + str(i)+";P="+str(PS)+"\n"
            tds += this_child.get_tree_desc_str()
            this_child = this_child.right_sibling
        tds += "\n^"
        return tds
    def out_of_ellipsis(self):
        print "Node "+str(self.n)+" going out of ellipsis"
        self.renderingMode = 0
        these_children = children(self)
        for c in these_children:
          c.un_end_ellipsis()
          
    def end_ellipsis(self):
        if self.renderingMode==0: self.renderingMode=2
        #self.renderingMode=2

    def un_end_ellipsis(self):
        if self.renderingMode==2: self.renderingMode=0

    def register_timestamp(self, time_string=None):
      import time
      if not time_string: time_string = time.strftime("%Y-%m-%d-%H:%M:%S")
      self.timestamp = time_string

    def initial_data(self):
        return Wrapped_int(1)

def get_operator_index_from_applicables(applicables, operator):
  # Returns an integer indicating which of the applicable operators was
  # used to create this state.
  # This can handle the case where the operator is a "derived instance"
  # and is not on the list of applicables. In such a case it returns the
  # index of the "mother" operator.  It gets the mother op's name by
  # taking a prefix (up to the double colon) of the derived op's name.
  if isinstance(operator, OperatorSpecifiedWithDialog):
    # Handle the special kind of operator...
    # Transform applicables to a list of operator names.
    names = map(lambda op: op.name, applicables)
    # Figure out the mother operator's name (the opname_root).
    opname_plus = operator.name+"::"
    k = opname_plus.index('::')
    opname_root = opname_plus[:k]
    #print "In get_operator_index_from_applicables, names="+str(names)
    #print "opname_root="+str(opname_root)
    # Find the index of the mother operator in the list.
    i = names.index(opname_root)
  else:
    # Normal kind operator -- find it directly.
    i = applicables.index(operator)
  return i

def get_absolute_operator_index(operator):
  if isinstance(operator, OperatorSpecifiedWithDialog):
    absopidx = operator.mother.absolute_op_num
  else:
    absopidx = operator.absolute_op_num
  return absopidx

def default_precondition(s): return True
def default_postcondition(s): return s.desc

class Operator(object):
    """encapsulates a transformation on states.
    """
    def __init__(self, name, data_function,
                 precondition=default_precondition,
                 postcondition=default_postcondition,
                 group=None, subgroup=None, subsubgroup=None):
        self.name = name
        self.function = data_function
        self.precondition = precondition
        self.postcondition = postcondition
        self.group = group
        self.subgroup = subgroup
        self.subsubgroup = subsubgroup
        self.absolute_op_num = -1 # to be set by register_operators.
        
    def __str__(self):
        s = "<operator: %s>" % (self.name)        
        return s
    
    def is_applicable(self, old_state):
        return apply(self.precondition, [old_state])
    
    def _apply(self, old_state, param_string):
        #print "Applying operator named ", self.name," with param_string: ", param_string
        return State(apply(self.function, [old_state.data, param_string]),
                     old_state,
                     self,
                     desc=apply(self.postcondition, [old_state])
                     )

class make_callable():
  def __init__(self, fn, param):
    self.function = fn
    self.parameter = param
    #print "Creating callable with param = "+str(param)
  def __call__(self, data, applic_time_params=""):
    fn = self.function
    new_data = fn(data, self.parameter, applic_time_params)
    return new_data
    

import tkFileDialog    
from tkSimpleDialog import askstring, askinteger

class OperatorSpecifiedWithDialog(Operator):
    """Although an normal Operator can take a parameter, there is no automatic
       presentation of a dialog, upon operator-selection, to get the value of this
       parameter.  The OperatorSpecifiedWithDialog class adds this feature and
       captures the user-input parameter to result in a normal Operator that
       has a non-empty-string parameter to it.
       There are several subtleties to this new class:
         1. Instances are used in two ways. Each instance serves one and only
            one of these two purposes.
            a. as permanent operators that bring up dialog boxes each time they
               are manually selected.
            b. as dynamically created closure-like transformations that can
               be applied to a state to get a new state.  Each time a user
               invokes a permanent operator and enters a parameter, one of
               these closure-like operators is created.  This closure has
               a new name, constructed using the derived instance count.
               The 'clone' method defined here helps create the new instance.
        2. The closure also closes in a function that takes 3 args instead
               of the Operator's usual 2.  The new arg (in second position),
               receives the stored parameter value from the dialog.
               When creating one of these operators, don't forget to pass in
               a reference to a function that will receive this argument.
               See the example at the end of this file.
        3. The permanent operator specifies info about how to create the dialog.
           This includes a param_op_type of "Free Text", "Int", or "File".
           There is also a dialog_title and a prompt that are passed to
           the dialog creating library functions.

       """
    def __init__(self, name, data_function,
                 precondition=default_precondition,
                 postcondition=default_postcondition,
                 group=None, subgroup=None, subsubgroup=None,
                 param_op_type='Free Text',
                 dialog_title='Enter information',
                 prompt='Type parameter here:',
                 filter_desc=[('All files','*')],
                 initial_value_fn=None):
      #print "dialog_title is "+str(dialog_title)
      super(OperatorSpecifiedWithDialog,self).__init__(name, data_function,
                 precondition, postcondition, group=group,
                 subgroup=subgroup, subsubgroup=subsubgroup)
      self.param_op_type = param_op_type # Either 'Free Text', 'Int', or 'File'.
      self.dialog_title=dialog_title
      self.prompt=prompt
      self.filter_desc = filter_desc      # Spec for what values are OK.
      self.initial_value_fn=initial_value_fn
      self.derivedInstanceCount=0
      # Initialize the Parameter -- maybe not nec., since Operator has params already.
      if param_op_type=="Int": self.parameter = "0"
      else: self.parameter = ""
      #print "In new OperatorSpecifiedWithDialog, self.dialog_title_string = "+str(self.dialog_title_string)
    def clone_with_encapsulated_param(self, param):
      self.derivedInstanceCount+=1
      newName = self.name+'::'+str(self.derivedInstanceCount)
      clone = OperatorSpecifiedWithDialog(newName, None,
                                          precondition=self.precondition,
                                          postcondition=self.postcondition)
      clone.parameter = param # For printing debugging message
      clone.function = make_callable(self.function, param)
      clone.mother = self # added 16 Apr 09 to facilitate finding absolute operator num. 
      return clone
    
    '''
    def __call__(self, state):
      new_state = data_function(state, self.param)
      # Maybe apply the postcondition, too. TO DO: ADD THIS CODE LATER
      return new_state
    '''
    # Overriding application method to incorporate the stored parameter:
    def _apply(self, old_state, applic_time_param_string=""):
        #print "Applying operator (specified with dialog) named ", self.name,\
        #      " with user-spec-parameter: (BEGIN)\n",self.parameter,\
        #      "\n(END), and applic-time param_string: ", applic_time_param_string,"\n"
        return State(apply(self.function, [old_state.data, applic_time_param_string]),
                     old_state,
                     self,
                     desc=apply(self.postcondition, [old_state]))
      
    def show_dialog(self,state):
      # Display a dialog to obtain the value of the parameter interactively.
      if self.param_op_type=='Free Text':
        return self.show_free_text_dialog(state)
      if self.param_op_type=='Int':
        return self.show_int_dialog(state)
      if self.param_op_type=='File':
        return self.show_file_dialog()
                     
    def show_free_text_dialog(self,state):
      if self.initial_value_fn:
        val = self.initial_value_fn(state)
        self.param = askstring(self.dialog_title,self.prompt,initialvalue=val)
      else:
        self.param = askstring(self.dialog_title,self.prompt)
      if self.param==None: return False
      else: return True
                      
    def show_int_dialog(self,state):
      if self.initial_value_fn:
        val = self.initial_value_fn(state)
        self.param = askinteger(self.dialog_title,self.prompt,initialvalue=val)
      else:
        self.param = askinteger(self.dialog_title,self.prompt)
      if self.param==None: return False
      else: return True

    def show_file_dialog(self):
      combined_title=self.dialog_title+" -- "+self.prompt
      filename = tkFileDialog.Open(filetypes=self.filter_desc,
                                   title=combined_title).show()
      if filename:
        '''
        if RELATIVE_PATHS:
          print "Trying to compute relative path..."
          # check for path outside of regular folders, and if so,
          # copy the file to a regular folder and change the path to
          # point to the copy.
          # The file types and folders are particular to the needs of PRIME Designer,
          # but additional categories and folders for resources could be set up.
          if filetypes[0][0]=="JPG Files": regs = ["common-images", "new-images"]
          if filetypes[0][0]=="WAV Files": regs = ["common-audio", "new-audio"]
          if file_outside(filename, regs):
            print "File is outside regular folders for images and audio... copying..."
            from os import copy
            relative_filename = regs[1]+"/"+basename(filename)
            copy(filename, "./"+relative_filename) # copy to new-images or new-audio
          else:
            # File is already in a regular folder. Compute shortened (relative) path
            print "File is already inside a regular folder."
            path_parts = filename.split("/")
            relative_filename = path_parts[-2]+"/"+path_parts[-1]
          filename = relative_filename
'''
        self.param = filename;
        return True
      else: print "No file selected."; return False
                     
def file_outside(filename, regs):
  path_parts = filename.split("/")
  deepest_folder = path_parts[-2]
  if deepest_folder in regs: return False
  else: return True
  
# The node object consists of a state plus some display coordinates.
# It's sort of the graphical embodiment of the state.
class Node(object):
    width=64; height=64 # Dimensions of the node (actually the image)
    def __init__(self, state):
        self.s = state
        self.focused = False
        self.rect = None
        self.text = None
        self.qin = 0 # quasi-inorder number for inorder display
        self.has_painted_content=False
        self.domain_specific_init(state)
        self.x = 0

    def set_coords(self, x, y):
        self.x = x
        self.y = y
    def paint(self):
        z = ZOOM_FACTOR # global variable for now
        if self.s.display_version==0:
            w=TINY_SIZE; h=TINY_SIZE
        else:
            sn = self.s.space_needed()
            w = self.width*sn; h = self.height*sn
        scale = self.s.scale
        w = int(scale*w); h = int(scale*h)
        # Display a box (which will appear behind the state's image):
        if TREE_DISP.node_in_selection(self):
          if z*scale < 0.8: margin = 7
          else: margin = 10
          self.rect = \
            SP_CANVAS.create_rectangle(z*(self.x-w/2-margin),z*(self.y-h/2-margin),
               z*(self.x+w/2+margin),z*(self.y+h/2+margin),fill="purple")
          
        node_color = "light blue"
        if self.s.has_hidden_children(): node_color = "light gray"
        if self.focused:
          if z*scale < 0.4: margin = 3
          else: margin = 10
          self.rect = \
            SP_CANVAS.create_rectangle(z*(self.x-w/2-margin),z*(self.y-h/2-margin),
               z*(self.x+w/2+margin),z*(self.y+h/2+margin),fill="light green")
        else:
          self.rect = \
            SP_CANVAS.create_rectangle(z*(self.x-w/2-1),z*(self.y-h/2-1),
              z*(self.x+w/2+1),z*(self.y+h/2+1),fill=node_color)
        # Display the image itself:
        if self.s.display_version > 1:
          if self.s.scale*ZOOM_FACTOR > 0.2:
            self.paint_label(z*w,z*h)
          self.paint_content(z*w,z*h,SP_CANVAS)
        
    def paint_label(self,w,h):
        # Display the state identifying string:
        z = ZOOM_FACTOR 
        self.text=SP_CANVAS.create_text(z*self.x,z*self.y-h/2-12*ZOOM_FACTOR,
                                        text=str(self.s))
    def unpaint(self):
        if self.rect: 
          SP_CANVAS.delete(self.rect)
        if self.text:
          SP_CANVAS.delete(self.text)
        self.unpaint_content(SP_CANVAS)

    def clicked_on(self, x,y):
        if self.s.hidden: return False
        z = ZOOM_FACTOR
        sn = self.s.space_needed()
        scale = self.s.scale
        w = self.width*sn*scale; h = self.height*sn*scale
        return (self.x - w/2 < x/z < self.x + w/2 and
                self.y - h/2 < y/z < self.y + h/2)
    def focus(self):
        self.focused = True
    def unfocus(self):
        self.focused = False
    def is_leaf(self):
        if self.s.first_child: return False
        else: return True

    def paint_content(self,w,h,a_canvas):
        z = ZOOM_FACTOR
        self.has_painted_content=True
        self.paint_content_in_rectangle(a_canvas,z*self.x,z*self.y,w,h)
    def unpaint_content(self,a_canvas):
        if self.has_painted_content:
            for item in self.p: a_canvas.delete(item)
        self.has_painted_content=False

    # Override the following 2 methods in a child class to
    # change the application domain for the program.
    def domain_specific_init(self, state):
        self.p = []

    def paint_content_in_rectangle(self,a_canvas,xcenter,ycenter,w,h):
        self.p.append(a_canvas.create_text(xcenter,ycenter,
                      text=str(self.s.data)))
      
NODE_CLASS=Node
LAST_NODE = None
# A kind of object to represent a tree display.
# It references a Tkinter canvas.
# It provide a method to draw a tree from a list of states.
LAST_INCR = 0


class tree_display:
    left_margin = 50
    top_margin = 50
    #deltax = 0  # horiz. scaling factor for values from quasi-inorder-tr.
    #deltay = 0 # vertical dist. betw. node centers, level-to-level.
    def __init__(self, acanvas, node_class_to_use):
        self.tdcanvas = acanvas
        self.state_node_hash = {}
        self.all_nodes = [];
        self.focus_node = None
        self.selection = []
        self.running_x = 0
        self.node_class = node_class_to_use
        # UNDO stuff (added 12 April 2009)
        self.nodes_to_undo = []
        self.selection_for_undo = []
        self.focus_for_undo = None
        self.parent_for_undo = None
        self.parent_last_child_for_undo = None
        self.parent_first_child_for_undo = None
        self.state_no_for_undo = None

    def grow_successors_and_their_nodes(self, a_node):
        these_children = successors(a_node.s)
        for c in these_children:
            self.register_state_and_node(c)

#    def register_state_and_node(self, s):
#        ALL_STATES.append(s)
#        new_node = self.node_class(s)
#        self.state_node_hash[s] = new_node
#        self.all_nodes.append(new_node)

    def node_in_selection(self, a_node):
        return a_node in self.selection
        
    def try_to_create_new_child(self, a_state, an_operator, param_string=""):
        if an_operator==None: return None
        existing_child = a_state.first_child
        op_is_redundant = False
        while existing_child:
            if existing_child.last_operator == an_operator:
                op_is_redundant = True; break
            existing_child = existing_child.right_sibling
        if op_is_redundant:
            user_response_num = warn_about_redundancy()
            if user_response_num == 1: return None
        # Save info for possible UNDO operation:
        self.state_no_for_undo = get_State_number()
        self.parent_first_child_for_undo = a_state.first_child
        self.parent_for_undo = a_state
        lc = a_state.last_child
        self.parent_last_child_for_undo = lc
        self.parent_lc_rs_for_undo = None
        if (lc != None): self.parent_lc_rs_for_undo = lc.right_sibling
        self.nodes_to_undo = []
        self.selection_for_undo = []
        self.focus_for_undo = TREE_DISP.focus_node
        # Done saving for possible UNDO.
        
        # Now generate the new state:
        this_child = an_operator._apply(a_state, param_string)
        if not this_child: return None
        self.register_state_and_node(this_child)
        return this_child

    def draw_tree_using_layout_params(self, root):
        self.tdcanvas.delete(ALL) # remove any existing display. 
        # Assign x coordinates to the nodes by performing 
        # a quasi-inorder traversal of the tree.
        root_node = self.state_node_hash[root]
        tree_display.deltax = int((root_node.width + 2)*1.2)
        tree_display.deltay = int((root_node.height + 2)*1.2)+10
        self.compute_subtree_bounding_boxes(root)
        self.running_x = 0
        self.compute_node_coordinates(root)
        nodes_to_draw = self.get_nodes(root,[],first_state=True)
        for n in nodes_to_draw:
          n.x = tree_display.left_margin + n.x*tree_display.deltax
          n.y = tree_display.top_margin + n.y*tree_display.deltay
          #n.set_coords(n.x,n.y)
        z = ZOOM_FACTOR
        for n in nodes_to_draw:
          if n.s.renderingMode==1: continue
          if n.s.d > 0:
              parent_node = self.state_node_hash[n.s.p]
              if n.s.renderingMode==2:
                while parent_node.s.renderingMode==1: # draw over "ellipsis"
                  parent_node = self.state_node_hash[parent_node.s.p]
                self.tdcanvas.create_line(
                  z*n.x, z*n.y, z*parent_node.x, z*parent_node.y, fill="dark green", stipple="gray50")
              else:
                self.tdcanvas.create_line(
                  z*n.x, z*n.y, z*parent_node.x, z*parent_node.y)
        # Finally, paint all the nodes:
        for n in nodes_to_draw:
          if n.s.renderingMode==1: continue
          n.paint()

    def compute_subtree_bounding_boxes(self, a_state, parent_in_ellipsis=False):
      """This performs something like inorder traversal starting at
         the given state. If the state is hidden, skip it.
         Compute the width and height of the bounding box around the
         subtree rooted at this state."""
      scale = a_state.scale
      a_state.bbwidth = 0; 
      if a_state.hidden: a_state.bbheight = 0; return
      #global LAST_INCR
      a_state.tempx = 0
      a_node = self.state_node_hash[a_state] # get tree node assoc. with this state.
      epsilon = 0.02 # enough to separate nodes in tiny mode.
      a_node.w2 = (a_state.space_needed()/2.0)+epsilon # half the display width of the node.
      unhidden_children_states = unhidden_children(a_state) # get the list of children.
      n_unhidden_children = len(unhidden_children_states)
      #thisNodeHeight = tree_display.deltay # Probably needs to be fixed. SLT
      in_ellipsis = (parent_in_ellipsis and a_state.renderingMode == 0) or \
                    a_state.renderingMode == 1
      provisional_height = a_state.getVSpaceNeeded()*scale             
      if in_ellipsis:
        a_state.bbheight = provisional_height*0.25
      else:
        a_state.bbheight = provisional_height
      if n_unhidden_children==0:        # if a_state has no unhidden children...
        a_state.tempx += a_node.w2*2    # add  the display width to running_x.
        if in_ellipsis:
          a_state.bbwidth = a_state.tempx*scale #*0.25    Reducing the width here doesn't work right.
        else:
          a_state.bbwidth = a_state.tempx*scale
        #print "Bounding box of subtree for state "+str(a_state.n)+" is "+str(a_state.bbwidth)+", "+str(a_state.bbheight)
        #print "State "+str(a_state.n)+" has no unhidden children, it seems."
        return 
      else:
        next_child = a_state.first_child            # prepare to iterate through the left kids.
        #a_state.starting_x = 0
        a_state.children_width=0
        for i in range(n_unhidden_children):
            if not next_child:
                print "Error, expected another unhidden child in state",\
                      a_state; break
            if next_child.hidden: next_child = next_child.right_sibling
            #LAST_INCR = next_child.space_needed() # amount of horiz space most recently used for a child at this level. 
            self.compute_subtree_bounding_boxes(next_child,parent_in_ellipsis=in_ellipsis) # recursive call on the child.
            last_child = next_child
            # Take the max height and the sums of the widths
            a_state.children_width += last_child.bbwidth #+ tree_display.deltax
            a_state.bbheight = max(a_state.bbheight, last_child.bbheight)
            #last_incr = next_child.space_needed() # amount of horiz space most recently used for a child at this level. 
            next_child = next_child.right_sibling
        a_state.bbwidth = max(a_state.bbwidth, a_state.children_width)
        #if self.running_x < starting_x + a_node.w2:
        #    self.running_x = starting_x + a_node.w2
      #print "Bounding box of subtree for state "+str(a_state.n)+" is "+str(a_state.bbwidth)+", "+str(a_state.bbheight)
    
    def compute_node_coordinates(self, a_state, y=0):
      """This performs something like inorder traversal starting at a_state.
        If the state is hidden, skip it."""
      if a_state.hidden: return # no coords needed.
      global LAST_INCR
      #self.running_x = 0
      scale = a_state.scale
      a_node = self.state_node_hash[a_state] # get tree node assoc. with this state.
      state_height = a_state.getVSpaceNeeded()*scale  # Does not take ZOOM_FACTOR into consid.
      
      a_node.set_coords(self.running_x+a_state.bbwidth/2, y+state_height/2)
      a_node.qin = self.running_x + a_state.bbwidth/2
      #print "Coords of state "+str(a_state.n)+" are "+str(a_node.x)+", "+str(a_node.y)
      epsilon = 0.02 # enough to separate nodes in tiny mode.
      a_node.w2 = scale*(a_state.space_needed()/2.0)+epsilon # half the display width of the node.
      unhidden_children_states = unhidden_children(a_state) # get the list of children.
      n_unhidden_children = len(unhidden_children_states)
      #thisNodeHeight = tree_display.deltay # Probably needs to be fixed. SLT
      #a_state.bbheight = thisNodeHeight
      if n_unhidden_children==0:        # if a_state has no unhidden children...
        self.running_x += a_node.w2     # add half the display width to running_x.
        #a_node.qin = self.running_x     # the node's quasi-inorder position is saved.
        self.running_x += a_node.w2     # the other half of the node-width is added to running_x
        #a_state.bbwidth = self.running_x 
        #print "Bounding box of subtree for state "+str(a_state.n)+" is "+str(a_state.bbwidth)+", "+str(a_state.bbheight)
        #print "State "+str(a_state.n)+" has no unhidden children, it seems."
        return 
      else:
        next_child = a_state.first_child  # prepare to iterate through the kids.
        #starting_x = self.running_x
        next_y = y + 1.2*state_height  #+ tree_display.deltay # Needs to be fixed
        for i in range(n_unhidden_children):
            if not next_child:
                print "Error, expected another unhidden child in state",\
                      a_state; break
            if next_child.hidden: next_child = next_child.right_sibling
            #LAST_INCR = next_child.space_needed() # amount of horiz space most recently used for a child at this level. 
            self.compute_node_coordinates(next_child, next_y) # recursive call on the child.
            last_child = next_child
            next_child = next_child.right_sibling
        #if self.running_x < starting_x + a_node.w2:
        #    self.running_x = starting_x + a_node.w2
      #print "Bounding box of subtree for state "+str(a_state.n)+" is "+str(a_state.bbwidth)+", "+str(a_state.bbheight)
    

    def get_nodes(self, a_state, nodelist_so_far,
                                first_state = False):
      """This performs something like inorder traversal starting at
         the given state. If the state is hidden, skip it."""
      if a_state.hidden: return nodelist_so_far
      unhidden_children_states = unhidden_children(a_state) # get the list of children.
      n_unhidden_children = len(unhidden_children_states)
      a_node = self.state_node_hash[a_state] # get tree node assoc. with this state.
      if n_unhidden_children==0:        # if a_state has no unhidden children...
        return nodelist_so_far + [a_node] # return, with this node appended to the given node list.
      else:
        next_child = a_state.first_child            # prepare to iterate through the left kids.
        for i in range(n_unhidden_children):
            if not next_child:
                print "Error, expected another unhidden child in state",\
                      a_state; break
            if next_child.hidden: next_child = next_child.right_sibling
            temp = self.get_nodes(next_child,[]) # recursive call on the child.
            nodelist_so_far += temp
            last_child = next_child
            next_child = next_child.right_sibling
        nodelist_so_far.append(a_node)
      return nodelist_so_far

    def grow_successors_and_their_nodes(self, a_node):
        # register stuff for undo:
        self.nodes_to_undo = []
        self.selection_for_undo = self.selection[:]
        self.focus_for_undo = self.focus_node
        self.parent_for_undo = a_node.s
        self.parent_first_child_for_undo = a_node.s.first_child
        lc = a_node.s.last_child
        self.parent_last_child_for_undo = lc
        self.parent_lc_rs_for_undo = None
        if (lc != None): self.parent_lc_rs_for_undo = lc.right_sibling
        self.state_no_for_undo = get_State_number()
        
        these_children = successors(a_node.s)
        for c in these_children:
            self.register_state_and_node(c)
    def register_state_and_node(self, s):
        ALL_STATES.append(s)
        new_node = self.node_class(s)
        self.state_node_hash[s] = new_node
        self.all_nodes.append(new_node)
        global LAST_NODE
        LAST_NODE = new_node
        #print "Recording last node"
        self.nodes_to_undo.append(new_node)
    def undo(self):
        p = self.parent_for_undo
        if p == None:
          print "In tree_display.undo: No undo-able operation is available for undo."; return
        fc = self.parent_first_child_for_undo
        p.first_child = fc
        #print "restoring so that first child is "+str(fc)
        lc = self.parent_last_child_for_undo
        p.last_child = lc
        set_State_number(self.state_no_for_undo)
        if (lc != None):
          if lc.right_sibling != None:
            lc.right_sibling = self.parent_lc_rs_for_undo
            #print "Undoing link from left sibling."
        for anode in self.nodes_to_undo:
          #print "Removing state from the hash:" + str(anode.s)
          del self.state_node_hash[anode.s]
          ALL_STATES.remove(anode.s)
          self.all_nodes.remove(anode)
        self.nodes_to_undo = []
        self.parent_for_undo = None
        #print "state_node_hasn is: "+ str(self.state_node_hash)

def warn_about_redundancy():
    import Dialog
    d = Dialog.Dialog(None, {'title': 'Duplicate An Existing State?',
        'text': 'The operator you have chosen has already been applied to'
                ' the focus state. Really create an identical twin?',
        'bitmap': 'questhead',
        'default': 0,
        'strings': ('Create twin',
                    'Cancel reapplication')})
    return d.num

def preorder_collect(r, root=False):
    """Traverses the subtree rooted at state r with preorder,
       and places each state reached into the list of states to
       be returned.  When calling from top level (non-recursively),
       be sure to use root=True, so that the method will not try
       to access its right sibling, which would take it out of its
       own subtree."""
    states_found = [r]
    if r.first_child:
        states_found += preorder_collect(r.first_child)
    if not root:
        if r.right_sibling:
            states_found += preorder_collect(r.right_sibling)
    return states_found

def collect_ancestors(a_state):
    if not a_state.p:
        return [a_state]
    else:
        return collect_ancestors(a_state.p)+[a_state]


def children(state):
    """Returns the existing children of the given state by
       traversing links."""
    if state.first_child:
        return [state.first_child] + right_siblings(state.first_child)
    else:
        return []

def unhidden_children(state):
    return filter(lambda s: not s.hidden, children(state))

def right_siblings(state):
    """Returns the siblings to the right of the given state by
       traversing links."""
    if state.right_sibling:
        return [state.right_sibling] + right_siblings(state.right_sibling)
    else:
        return []

def applicable_ops(s):
    """Returns the subset of OPERATORS whose preconditions are
       satisfied by the state s."""
    return filter(lambda o: o.is_applicable(s), OPERATORS)
    
def successors(s):
    """Generates and returns the successors of the state s by
       applying to it all those operators in the list OPERATORS
       that are applicable to the state s."""
    ops_already_applied_here = map(lambda c: c.last_operator, children(s))
    nonredundant_ops = filter(lambda op: not op in ops_already_applied_here,
                              applicable_ops(s))
    noninteractive_ops = filter(lambda op: not isinstance(op, OperatorSpecifiedWithDialog),
                                nonredundant_ops)
    return map(lambda o: o._apply(s,""), noninteractive_ops)

def register_operators(ops):
    """Takes the given list ops of operators and assigns it to
       the global variable OPERATORS. Also computes the length
       of the list and assigned it to the global BRANCHING_FACTOR."""
    global OPERATORS, BRANCHING_FACTOR
    OPERATORS = ops
    BRANCHING_FACTOR = len(OPERATORS)
    for i in range(len(ops)):
      OPERATORS[i].absolute_op_num = i
      # print "absolute_op_num for Operator "+OPERATORS[i].name + " = "+str(i)

def register_more_operators(ops):
    """Appends more operators onto OPERATORS and
       updates BRANCHING_FACTOR."""
    global OPERATORS, BRANCHING_FACTOR
    oldn = len(OPERATORS)
    OPERATORS += ops
    newn = len(OPERATORS)
    BRANCHING_FACTOR = len(OPERATORS)
    for i in range(oldn, newn):
      OPERATORS[i].absolute_op_num = i
      # print "absolute_op_num for Operator "+OPERATORS[i].name + " = "+str(i)

def gen_subtree_depth2():
    gen_subtree(2)

def gen_subtree_depth3():
    gen_subtree(3)
    
def gen_subtree(n_new_levels):
    print "Generating a subtree of depth 2"
    log("gen_subtree")
    # Get focus node.
    # Get number of levels requested (default: 2) and display mode (default: tiny)
    # Use Breadth-First approach to generating the requested subtree,
    #   automatically avoiding duplicates.
    # Set the display mode on all the new states (but not on subtree components
    #   that existed before the command was executed).
    sn = TREE_DISP.focus_node
    if not sn:
        print "No state has been chosen as the focus, so we will use the root state."
        sn = TREE_DISP.state_node_hash[INITIAL_STATE]
    #n_new_levels = 2 
    new_states_display_mode = 0 # hardcoded for now
    pending = [sn] # list of nodes whose successors must be generated.
    #prepare for possible UNDO:
    TREE_DISP.nodes_to_undo = []
    TREE_DISP.selection_for_undo = TREE_DISP.selection[:]
    TREE_DISP.focus_for_undo = TREE_DISP.focus_node
    TREE_DISP.parent_for_undo = sn.s
    TREE_DISP.parent_first_child_for_undo = sn.s.first_child
    lc = sn.s.last_child
    TREE_DISP.parent_last_child_for_undo = lc
    TREE_DISP.parent_lc_rs_for_undo = None
    if (lc != None): TREE_DISP.parent_lc_rs_for_undo = lc.right_sibling
    TREE_DISP.state_no_for_undo = get_State_number()
    # generate the states and nodes:
    for i in range(n_new_levels):
        new_pending = []
        for node in pending:
            these_children = successors(node.s)
            for c in these_children:
                global ALL_STATES
                ALL_STATES.append(c)
                new_node = TREE_DISP.node_class(c)
                TREE_DISP.state_node_hash[c] = new_node
                TREE_DISP.nodes_to_undo.append(new_node)
                TREE_DISP.all_nodes.append(new_node)
                new_pending.append(new_node)
        pending = new_pending
    redraw_tree()

# Domain-dependent part of setup:
"""
The following parts of this program are specific to the
domain of simple operations on numbers.
The data component of each state is a Wrapped_int which
is an int that knows how to save itself to a file.
Saving is not really important for an int, but we do that
here for consistency with what one would want if the data
were more complicated, such as with an image."""

class Wrapped_obj:
  """Provides objects with a method to save themselves."""
  def save(self,filename):
    print "Preparing to save the data ",self
    try: 
      f = open(filename, "a")
      f.write(str(self))
      f.close()
    except: print "Couldn't write the file: ", filename
    else: print "Saved the number ", self, " to file ", filename

class Wrapped_int(int,Wrapped_obj):
  """Provides int data with a method to save themselves."""
  pass

class Wrapped_string(str,Wrapped_obj):
  """Provides string data with a method to save themselves."""
  pass

def create_initial_state():
  """(Re)initializes the counter for states, and
     create a new state, assigning it to the global
     variable INITIAL_STATE."""
  global INITIAL_STATE
  State.number = 0 # Start state numbering from 0.
  INITIAL_STATE = State(INITIAL_DATA,None,None,desc='number')

def read_data_from_file(filename):
    """Gets information from a file."""
    f = open(filename, "r")
    temp = f.read()
    f.close()
    return temp

def build_data_object(data):
  """Converts the information read in from a file to the
     form needed within an initial state."""
  return Wrapped_int(data)

def get_current_value(s):
  # This function is used by one of the operators to
  # initialize the value to be edited in a dialog
  return s.data

def f1(n,ps=""):return Wrapped_int(n+1)
def f2(n,ps=""):return Wrapped_int(n*2)
def f3(n,ps=""):return Wrapped_int(n*7)
def f4(n,ps=""):return Wrapped_int(n / 10)
def f5(n,ps=""): return Wrapped_string(num_to_string(n))
def f6(strng,ps=""): return Wrapped_string(strng.swapcase())
def f7(strng,ps=""): return Wrapped_string(str_rot(strng))

def num_to_string(n):
    """Converts, for example, 475 to 'four-seven-five' """
    if n<10: return {0:'zero',1:'one',2:'two',3:'three',4:'four',
                     5:'five',6:'six',7:'seven',8:'eight',9:'nine'}[n]
    else: return num_to_string(n/10)+'-'+num_to_string(n%10)

def str_rot(strng):
    """Moves the last character of the string to the front."""
    if len(strng)<2: return strng
    else: return strng[-1]+strng[:-1]
    
def needs_num(s):
    """Returns True if the data for state s is a Wrapped_int."""
    return isinstance(s.data, Wrapped_int)

def needs_str(s):
    """Returns True if the data for state s is a Wrapped_string."""
    return isinstance(s.data, Wrapped_string)

# Note that six of our seven operators use the default postcondition,
# which is for the child state to receive the same desc value as
# its parent.  Only op5 changes it (from 'number' to 'string')
arithmetic_operators = map(
    lambda name,fun:Operator(name,fun,precondition=needs_num,group="Arithmetic"),
    ['Add-1', 'Times-2', 'Times-7', 'Div-10'],
    [f1, f2, f3, f4])

convert_operators = [
    Operator('Convert',f5,
             precondition=needs_num,
             postcondition=lambda s: 'string',
             group="Conversion") ]
string_operators = [
    Operator('SwapCase',f6, precondition=needs_str,group="String"),
    Operator('Rotate',f7, precondition=needs_str,group="String") ]

# sample dialog-operator function:
def dof1(data, stored_param, applic_time_param_str=""):
    return Wrapped_string(stored_param)
def dof2(data, stored_param, applic_time_param_str=""):
    return Wrapped_int(stored_param)
    
dialog_operators = [
    OperatorSpecifiedWithDialog('Add text freely',dof1,group="Dialog",
                                dialog_title="Give description",
                                prompt="Type your description here:"),
    OperatorSpecifiedWithDialog('Specify an integer freely',
                                dof2,
                                group="Dialog",
                                dialog_title="Number needed",
                                prompt="Type an integer:",
                                param_op_type="Int"),
    OperatorSpecifiedWithDialog('Specify a file',dof1,group="Dialog",
                                dialog_title="Choose a file",
                                prompt="Select the file of your choice",
                                param_op_type="File"),
    OperatorSpecifiedWithDialog('Edit existing text',dof1,group="Dialog",
                                dialog_title="Revise description",
                                prompt="Update your description here:",
                                initial_value_fn=get_current_value),

    ]

  
register_operators(arithmetic_operators +
                   convert_operators +
                   string_operators +
                   dialog_operators)

def get_full_root_data():
    return INITIAL_STATE.data
  
def display_data_in_detail(detail_canvas, desc, op_seq, temp_object, the_state):
    # temp_object is ignored in this default application.
    # op_seq is not used explicitly, since desc, which embeds it in
    #  a string, is being used here instead.
    # It is important in TRAIPSE and other applications.
    detail_canvas.create_text(300,200,text=desc,anchor=CENTER)

# The following dummy function is for logging UI events.  It can be overridden
# in application modules to actually log to a file or to a server.
def log(msg):
  pass

# The following is only executed if this module is being run as the main
# program, rather than imported from another one.
if __name__ == '__main__':
  print "Setting up..."
  global INITIAL_DATA
  INITIAL_DATA = Wrapped_int(1)
  AUTO_FOCUS_LAST_NODE = False
  setup_and_run()
  print "The session is finished."



 
