'''primeXMLwriter.py

Status:
  The XML file now includes properties (title, version, authors, date),
  layout, puzzles, and rules  (as of 6:00 PM, Mon, Aug. 30, 2007).

  This file also contains the code to create processed images for
  image puzzles.  This code is working.
  
  It is currently considered the responsibility of the game engine to
  show a "dead-end" image, or no image at all when the player creates
  a sequence operators that does not have a corresponding processed image.

  Written by S. Tanimoto as part of the NSF-funded project on
  "Problem-solving methodology in collaborative design."

  Updated Oct 23, 2007.

  (C) 2007 by the University of Washington and Steven Tanimoto.
  
'''

import os, os.path
import tkFileDialog
import TStar09 as TStar
import time
import ImageP.ImagePuzzleDesigner05 as ImageP    
import MusicP.MusicPuzzleDesigner05 as MusicP    
XML_FILENAME=None
FOLDER_NAME=None
GAMES_FOLDER = "GameFiles"
PATH_PREFIX = GAMES_FOLDER+"/"

def write_prime_game_as_xml(floor_plan, image_puzzles,
                            music_puzzles, rule_base, props,
                            use_folder=False, full=False,
                            draft=False, images=True):
    # Get a filename:
    global XML_FILENAME, FOLDER_NAME, GAMES_FOLDER
    
    #if XML_FILENAME!=None:
    #    #init_filename = XML_FILENAME.split("/")[-1]
    #    init_filename = XML_FILENAME
    #else:
    init_filename = 'untitled_game.xml'
    savedir = os.getcwd()
    #os.chdir(GAMES_FOLDER)
    filename = tkFileDialog.SaveAs(filetypes=[('PRIME XML files','*.xml')],
                                   initialfile=PATH_PREFIX+init_filename,
                                   title='Export your game file').show()
    if not filename: print "Export operation cancelled."; return
    if use_folder:
        if filename[-4:]==".xml":
          FOLDER_NAME = filename[:-4]
        else:
          FOLDER_NAME = filename
        parts = FOLDER_NAME.split("/")
        FOLDER_NAME = parts[-1]
        GAMES_FOLDER = "/".join(parts[:-1])
        if not os.path.exists(GAMES_FOLDER+"/"+FOLDER_NAME):
            os.mkdir(GAMES_FOLDER+"/"+FOLDER_NAME)
        filename = GAMES_FOLDER+"/"+FOLDER_NAME+"/"+FOLDER_NAME+".xml"
        XML_FILENAME=filename
    else:
        XML_FILENAME=filename
    # Create XML:
    d = '<prime_game name="'+props[0]+'" version="'+props[1]+'" '
    d += 'authors="'+props[2]+'" \n'
    d += '  design_tools="PRIME Designer V0.3"\n'
    d += '  date="'+time.ctime()+'">\n'
    d += make_layout_xml(floor_plan, use_folder=use_folder)
    d += make_image_puzzles_xml(image_puzzles,
                                procimages=full,
                                draft=draft, images=images)
    d += make_music_puzzles_xml(music_puzzles, use_folder=use_folder)
    d += make_rule_base_xml(rule_base, use_folder=use_folder)
    d += '</prime_game>\n'
    # for testing:
    #print d; return

    fp = open(filename, 'w')
    fp.write(d)
    fp.close()

    os.chdir(savedir) # restore directory
    # Remember the name of this file to help select game for playing
    # in a future session.
    flast = open("last_game_exported.txt", "w")
    flast.write(filename)
    print "Finished exporting the game to "+filename
    print "**************************************************"

def make_layout_xml(fp, use_folder=False):
    # From the FloorPlan fp, build a string in XML.
    d = "<layout>\n"
    # If overall wallpaper has been specified, write that out.
    try:
        game_default_wallpaper = fp.overall_default_wallpaper
        if use_folder: copy_image_to_folder(game_default_wallpaper)
        if not game_default_wallpaper: raise exception
        stripped = strip(game_default_wallpaper)
    except: game_default_wallpaper = ''
    stripped = strip(game_default_wallpaper)
    d += '<overall_default_wallpaper name="'+stripped+'"></overall_default_wallpaper>'+"\n"
    # Write out the description of each room.
    rooms = fp.rooms
    for i in range(3):
        for j in range(3):
            r = rooms[i][j]
            # Note that room numbering for internal purposes is 0 through 8,
            # but users see values 1 through 9 in names and rules.
            # Game engines should, but might not, display external numbers only.
            # If the room has been given a name, write that out.
            try:
                name = fp.room_names[3*i+j]
                if not name: raise exception
            except: name = 'ROOM-'+str(3*i+j+1)
            d += '<room num="'+str(3*i+j)+'" name="'+name+'" '
            # If there is background audio for this room, write that out.
            try:
                bg_audio = fp.background_audio_filenames[3*i+j]
                if use_folder: copy_image_to_folder(bg_audio)
                if not bg_audio: raise exception
                stripped = strip(bg_audio)
            except: bg_audio = ''
            stripped = strip(bg_audio)
            d += 'background_audio="'+stripped+'" '
            # If there is default wallpaper for this room, write that out.
            try:
                rm_df_wallpaper = fp.room_default_wallpaper_filenames[3*i+j]
                if use_folder: copy_image_to_folder(rm_df_wallpaper)
                if not rm_df_wallpaper: raise exception
                stripped = strip(rm_df_wallpaper)
            except: rm_df_wallpaper = ''
            stripped = strip(rm_df_wallpaper)
            d += 'default_room_wallpaper="'+stripped+'">'+"\n"
            for k in range(4):
                w = r[k]
                d += '  <wall loc="'+['N','E','S','W'][k]+'">'+"\n"
                if w.wallpaper:
                    if use_folder: copy_image_to_folder(w.wallpaper)
                    stripped = strip(w.wallpaper) # Remove path info. 
                    d+='    <wallpaper name="'+stripped+'"></wallpaper>\n'
                if w.caption:
                    d+='    <caption text="'+w.caption+'"></caption>\n'
                if w.door and w.door > -1:
                    # get first and second room numbers from door number
                    n1 = w.door/10; n2 = w.door%10
                    this_room = 3*i+j+1
                    if n1==this_room: into = str(n2-1)
                    else: into = str(n1-1)
                    if w.code & 8 > 0: state="closed"
                    else: state="open"
                    d+='    <door num="'+str(w.door)+'" into="'+into+'" state="'+state+'"></door>\n'
                if w.puzzle and w.puzzle > -1:
                    d+='    <puzzle_inst puzzle_num="'+str(w.puzzle)+'"></puzzle_inst>\n'
                d += '  </wall>\n'
            d += '</room>\n'
    d += '</layout>\n'
    return d

def strip(full_filename):
    parts = full_filename.split("/")
    filename = parts[-1]
    return filename

def copy_image_to_folder(filename):
    global GAMES_FOLDER, FOLDER_NAME
    import shutil
    dest = GAMES_FOLDER + "/" + FOLDER_NAME + "/"
    shutil.copy(filename, dest)
    
PROC_IMAGE_RES = None
NUM_PROCIMAGES = None # total will be used to qualify progress
PROCIMAGE_NUM = None
USE_PROGRESSBAR = True
def make_image_puzzles_xml(puzzles, procimages=False, draft=False, images=True):
    global PROCIMAGE_XML, PROC_IMAGE_RES, USE_PROGRESSBAR, PROCIMAGE_NUM
    global NUM_PROCIMAGES, PROGRESS_BAR
    d=''
    if procimages:
        if draft:
            # Set the resolution for PROC IMAGES to a lower value to
            # speed the generation (and lower the storage requirement)
            PROC_IMAGE_RES = ((64, 64))
        else:
            PROC_IMAGE_RES = ((ImageP.BASEIMAGE_SIZE, ImageP.BASEIMAGE_SIZE))
        PROCIMAGE_NUM = 0 # current number reflects progress
        if USE_PROGRESSBAR:
            # Create the meter:
            show_meter()
            # simulate creating all the procimages without actually
            # generating images, in order to find out how many there will be.
            procimage_xml = generate_puzzle_images(puzzles, images=False)
            NUM_PROCIMAGES = PROCIMAGE_NUM
            print "Number of images to generate is "+str(NUM_PROCIMAGES)
            PROCIMAGE_NUM = 0 # reset the counter
        # now actually make the images, if the user requested them to be made:
        #procimage_xml = generate_puzzle_images(puzzles, images)
    num = 1
    for p in puzzles:
        print "In make_image_puzzles_xml, puzzle p is: "+str(p)
        d+='<image_puzzle num="'+str(num)+'" '+name_to_xml(p)+'>\n'
        d+=starting_file_and_msg_to_xml(p)
        d+=render_operator_set_to_xml(p)
        d+=correct_sequence_xml(p)
        if procimages:
            # now actually make the images, if the user requested them to be made:
            procimage_xml = generate_images_for_puzzle(p, num, images)
            d+=PROCIMAGE_XML
        d+='</image_puzzle>\n'
        num += 1
    if USE_PROGRESSBAR:
        PROGRESS_BAR.destroy()
    return d

def make_music_puzzles_xml(puzzles, use_folder=False):
    d=''
    num = 101 # Numbers over 100 refer to music puzzles.
    for p in puzzles:
        d+='<music_puzzle num="'+str(num)+'" '+name_to_xml(p)+'>\n'
        d+=starting_file_and_msg_to_xml(p, use_folder=use_folder)
        d+=render_operator_set_to_xml(p, music=True)
        d+=correct_sequence_xml(p, music=True)
        d+=proc_music_xml(p)
        d+='</music_puzzle>\n'
        num += 1
    return d

PROC_MUSIC_XML = ""
def proc_music_xml(p):
    global PROC_MUSIC_XML
    PROC_MUSIC_XML = ""
    # create scrambled message by applying the op_seq from the puzzle desc. to the secret message.
    #
    # Do a DF traversal of a tree rooted at the scrambled message, applying inverse ops.
    # Use an error tolerance to cut off the search. xxx
    misd = 3 # MAX_INCORRECT_SEQUENCE_DEPTH = 3 #-- how far a user can wander from
    # the correct operator sequence and still get an actual
    # representation of that sequence rather than a default image.
    import TextListListDesigner05 as TextListListDesigner
    inverse_op_functions = map(lambda op: MusicP.INVERSE[op], MusicP.TRANSFORMATION_FUNCTIONS)
    theMusicPNode = MusicP.MusicPNode(TStar.TREE_DISP.focus_node.s.data[1])
    op_sequence = theMusicPNode.build_op_sequence(p)
    message = p[2][16:]
    scrambled_message = MusicP.permute_notes(message, op_sequence)
    PROC_MUSIC_XML += '<proc_music seq="()" msg="'+scrambled_message+'"></proc_music>\n'
    #
    # Assume that all image ops are available to the puzzle solver.
    all_ops_for_puzzle = range(len(MusicP.TRANSFORMATION_FUNCTIONS))
    # Walk the tree of operator application sequences, depth-first
    # keeping track of how much of each sequence is "correct",
    # and go a certain number of levels into error territory.
    # At each node of this tree, create a procimage.

    correct_seq=list(op_sequence)
    correct_seq.reverse()
    op_set_indices = range(len(inverse_op_functions))
    gen_music_for_puzzle(scrambled_message,
                          op_set_indices, correct_seq, [], misd, 0)
    return PROC_MUSIC_XML

def gen_music_for_puzzle(current_message, op_set_indices, rest_of_correct_seq,
                          seq_so_far, misd, current_error_depth):
    print "In gen_music_for_puzzle, seq_so_far is "+str(seq_so_far)
    print "  current_error_depth is "+str(current_error_depth)
    print "  rest_of_correct_seq is "+ str(rest_of_correct_seq)
    # Here, starting_im is the scrambled image, in base-image form.
    for op in op_set_indices:
        print "  Extending the current op sequence with op "+str(op)
        new_message = get_or_create_puzzle_proc_music_incrementally(current_message,
                                                     seq_so_far, op)
        if current_error_depth + 1 < misd:
            if len(rest_of_correct_seq)>0 and op==rest_of_correct_seq[0]:
                new_rest = rest_of_correct_seq[1:]
                if current_error_depth==0: new_error_depth=0
            else:
                new_error_depth = current_error_depth+1
                new_rest = []
            gen_music_for_puzzle(new_message,
                                  op_set_indices, new_rest, seq_so_far+[op],
                                  misd, new_error_depth+1)
            
def get_or_create_puzzle_proc_music_incrementally(last_message, seq_so_far, this_op):
  global PROC_MUSIC_XML
  new_message = apply(MusicP.INVERSE_FUNCTIONS[this_op], [last_message])
  PROC_MUSIC_XML += '<proc_music seq="'+str(tuple(seq_so_far+[this_op]))+'" msg="'+new_message+'"></proc_music>\n'
  return new_message

def UPPER_CASE(strng): return strng.upper()

def make_rule_base_xml(rules, use_folder=False):
    d='<control_program>\n'
    n = 0
    for r in rules:
        default_name = "RULE_"+str(n+1)
        d+='<rule num="'+str(n)+'" name="'+default_name+'" condition_type="AND">\n'
        for component in r:
            if component[0:4]=='(Con':
                rest = component[13:] # Remove the "(Condition:) " prefix
                i = rest.index(':'); token = rest[:i]
                #print "token in condition is:"+token
                cond_type = "event"
                if token=="In room" or token=="Door closed" or token=="Door open" or\
                   token=="Score is below" or token=="Score is above" or\
                   token=="Flag set" or token=="Flag not set":
                    cond_type="predicate"
                if token=="Entering room":
                    rest=rest[15:] # consume "Entering room: "
                    token="Entering"
                elif token=="In room": rest=rest[9:]; token="In"
                elif token=="Uttering phrase":
                    rest=rest[17:] # consume "Uttering phrase: "
                    token="Uttering"
                elif token=="Door closed":
                    rest=rest[13:] # consume "Door closed: "
                    token="DoorClosed"
                elif token=="Door open":
                    rest=rest[11:] # consume "Door open: "
                    token="DoorOpen"
                elif token=="Flag set":
                    rest=rest[10:] # consume "Flag set: "
                    token="FlagSet"
                    rest=UPPER_CASE(rest)
                elif token=="Flag not set":
                    rest=rest[14:] # consume "Flag not set: "
                    token="FlagNotSet"
                    rest=UPPER_CASE(rest)
                elif token=="Score is below":
                    rest=rest[16:] # consume "Score is below: "
                    token="ScoreBelow"
                elif token=="Score is above":
                    rest=rest[16:] # consume "Score is above: "
                    token="ScoreAbove"
                elif token=="Reached room for 1st time":
                    rest=rest[27:] # consume "Reached room for 1st time: "
                    token="Reached"
                #print "Before using arg, rest="+rest
                d+=' <condition '+cond_type+'="'+token+'" arg="'
                arg = rest
                d+=arg+'"></condition>\n'
            elif component[0:4]=='(Act':
                rest = component[10:] # Remove the "(Action:) " prefix
                i = rest.index(':'); token = rest[:i]
                #print "token in action is:"+token
                if token=="Announce": rest=rest[10:] # consume "Announce: "
                elif token=="Open door": rest=rest[11:]; token="Open"    # consume "Open door: "
                elif token=="Close door": rest=rest[12:]; token="Close"    # consume "Close door: "
                elif token=="Increase score":
                    rest=rest[16:] # consume "Increase score: "
                    token="Score"
                elif token=="Set flag":
                    rest=rest[10:];
                    token="Set"
                    rest=UPPER_CASE(rest)
                elif token=="Clear flag":
                    rest=rest[12:];
                    token="Clear"
                    rest=UPPER_CASE(rest)
                elif token=="Play audio file":
                    rest=rest[17:]; token="Play"    # consume "Play audio file: "
                    if use_folder:
                        copy_image_to_folder(rest)
                        stripped = strip(rest)
                        rest = stripped
                #print "Before using arg, rest="+rest
                d+=' <action event="'+token+'" arg="'
                arg = rest
                d+=arg+'"></action>\n'
        d+='</rule>\n'
        n += 1
    d+='</control_program>\n'
    return d
def name_to_xml(p):
    name=p[0][13:]
    return 'name="'+name+'"'
def starting_file_and_msg_to_xml(p, use_folder=False):
    filename=p[1][12:]
    message=p[2][16:]
    print 'filename is:"'+filename+'"'
    if filename == "Default-music-image.gif":
        filename = "common-images/Beethoven-by-Stieler1820.jpg"
    d =  ' <message text="'+message+'"></message>\n'
    # Copy the background image.
    if use_folder:
      copy_image_to_folder(filename)
    d += ' <source role="1" file="'+strip(filename)+'"></source>\n'
    return d
def render_operator_set_to_xml(p, music=False):
    d = ''
    if music:
        import MusicP.MusicPuzzleDesigner05 as MusicP
        xforms = MusicP.TRANSFORMATION_NAMES
    else:
        xforms = ImageP.TRANSFORMATION_NAMES
    for i in range(len(xforms)):
        d += ' <operator role="'+str(i)+'" name="'+xforms[i]+'" op="T_'+str(i)+'"></operator>\n'
    return d
def correct_sequence_xml(p, music=False):
    d = ' <correct_sequence seq="'
    seq = get_correct_sequence(p, music=music)
    d += str(tuple(seq))
    d += '"></correct_sequence>\n'
    return d

def get_correct_sequence(p, music=False):
    seq = []
    for t in p[3:]: # Skip Name, File, and Msg.
        if music: i = get_index_of_music_xform(t)
        else: i = get_index_of_image_xform(t)
        seq = [i]+seq
    return seq
    
def get_index_of_image_xform(t):
    # Return the integer that identifies the transformation
    # named in string t.
    xforms = ImageP.TRANSFORMATION_NAMES
    print "In get_index_of_image_xform, transformation names are: "+str(xforms)
    print "   This transform is "+str(t)
    return xforms.index(t)
def get_index_of_music_xform(t):
    # Return the integer index of music xform t.
    # named in string t.
    import MusicP.MusicPuzzleDesigner05 as MusicP
    xforms = MusicP.TRANSFORMATION_NAMES
    return xforms.index(t)

PROCIMAGE_XML = None
INVERSE_OP_FUNCTIONS = None

def generate_puzzle_images(image_puzzles, images=True):
    # create a set of images that represent the most important
    # potential states of each image puzzle, so that game engines
    # don't have to generate them on demand.
    misd = 3 # MAX_INCORRECT_SEQUENCE_DEPTH = 3 #-- how far a user can wander from
    # the correct operator sequence and still get an actual
    # representation of that sequence rather than a default image.
    # If images==False, create only the XML.
    import TextListListDesigner05 as TextListListDesigner
    import ImageP.primeIMAGEops as ImageOps
    global INVERSE_OP_FUNCTIONS
    INVERSE_OP_FUNCTIONS = map(lambda op: ImageOps.INVERSE[op], ImageP.TRANSFORMATION_FUNCTIONS)
    global FOLDER_NAME, PROCIMAGE_XML, GAMES_FOLDER, PROC_IMAGE_RES
    PROCIMAGE_XML = ""
    theImagePNode = ImageP.ImagePNode(TStar.TREE_DISP.focus_node.s.data[1])
    n = 1
    for p in image_puzzles:
        print "In generate_puzzles_images, puzzle p is: "+str(p)
        # Get the original source filename...
        fname = TextListListDesigner.get_puzzle_filename(p)
        #fname = p[1][13:]
        print "Raw Filename is: "+fname
        # Create a new base name to use for all image files in this puzzle.
        puzzle_image_name_prefix = "Puzzle-"+str(n)+"-img-ver"
        bkey = puzzle_image_name_prefix+"--"
        # Create the scrambled version of the image.
        op_sequence = theImagePNode.build_op_sequence(p)
        message = p[2][16:]
        global PROCIMAGE_NUM, NUM_PROCIMAGES
        PROCIMAGE_NUM += 1
        if images:
            scrambled_image = theImagePNode.get_baseimage(fname, op_sequence, message).copy()
            scrambled_image.thumbnail(PROC_IMAGE_RES) # Reduces resolution in draft mode.
            
            # Save it to a file.
            #scrambled_fn = GAMES_FOLDER+"/"+FOLDER_NAME+"/"+puzzle_image_name_prefix+".jpg"
            #scrambled_image.save(scrambled_fn, "JPEG")

            im = scrambled_image
            im.save(GAMES_FOLDER+"/"+FOLDER_NAME+"/"+bkey+".jpg", "JPEG")
            if USE_PROGRESSBAR:
                PROGRESS_BAR.set((0.0 + PROCIMAGE_NUM)/NUM_PROCIMAGES, None)
        else:
            scrambled_image = None
        PROCIMAGE_XML += '<proc_image seq="()" filename="'+bkey+'.jpg"></proc_image>\n'
            
        #
        # Assume that all image ops are available to the puzzle solver.
        all_ops_for_puzzle = range(len(ImageP.TRANSFORMATION_FUNCTIONS))
        # Walk the tree of operator application sequences, depth-first
        # keeping track of how much of each sequence is "correct",
        # and go a certain number of levels into error territory.
        # At each node of this tree, create a procimage.
        # NOTE -- THE FOLLOWING NEEDS TO CHANGE TO TURN OP INTO A FN.

        # Under construction...
        #op_function_seq = map(lambda op:ImageP.TRANSFORMATION_FUNCTIONS[op], op_sequence)
        #correct_seq = map(lambda op:ImageOps.INVERSE[ImageP.TRANSFORMATION_FUNCTIONS[op]], op_sequence)
        correct_seq=list(op_sequence)
        correct_seq.reverse()
        op_set_indices = range(len(INVERSE_OP_FUNCTIONS))
        
        
        gen_images_for_puzzle(puzzle_image_name_prefix, scrambled_image,
                              op_set_indices, correct_seq, [], misd, 0, images=images)
        n += 1

# New function being added on 25 March 2009 to fix problem of procimage xml being
# not properly distributed into corresponding image puzzles.
def generate_images_for_puzzle(puzzle, num, images=True):
    import TextListListDesigner05 as TextListListDesigner
    import ImageP.primeIMAGEops as ImageOps
    global INVERSE_OP_FUNCTIONS
    global FOLDER_NAME, PROCIMAGE_XML, GAMES_FOLDER, PROC_IMAGE_RES
    misd = 3 # MAX_INCORRECT_SEQUENCE_DEPTH = 3 #-- how far a user can wander from
    PROCIMAGE_XML = ""
    theImagePNode = ImageP.ImagePNode(TStar.TREE_DISP.focus_node.s.data[1])
    print "In generate_images_for_puzzle, puzzle is: "+str(puzzle)
    # Get the original source filename...
    fname = TextListListDesigner.get_puzzle_filename(puzzle)
    #fname = p[1][13:]
    #print "Raw Filename is: "+fname
    # Create a new base name to use for all image files in this puzzle.
    puzzle_image_name_prefix = "Puzzle-"+str(num)+"-img-ver"
    bkey = puzzle_image_name_prefix+"--"
    # Create the scrambled version of the image.
    op_sequence = theImagePNode.build_op_sequence(puzzle)
    message = puzzle[2][16:]
    global PROCIMAGE_NUM, NUM_PROCIMAGES
    PROCIMAGE_NUM += 1
    if images:
        scrambled_image = theImagePNode.get_baseimage(fname, op_sequence, message).copy()
        scrambled_image.thumbnail(PROC_IMAGE_RES) # Reduces resolution in draft mode.
        
        # Save it to a file.
        #scrambled_fn = GAMES_FOLDER+"/"+FOLDER_NAME+"/"+puzzle_image_name_prefix+".jpg"
        #scrambled_image.save(scrambled_fn, "JPEG")

        im = scrambled_image
        im.save(GAMES_FOLDER+"/"+FOLDER_NAME+"/"+bkey+".jpg", "JPEG")
        if USE_PROGRESSBAR:
            PROGRESS_BAR.set((0.0 + PROCIMAGE_NUM)/NUM_PROCIMAGES, None)
    else:
        scrambled_image = None
    PROCIMAGE_XML += '<proc_image seq="()" filename="'+bkey+'.jpg"></proc_image>\n'
    #
    # Assume that all image ops are available to the puzzle solver.
    all_ops_for_puzzle = range(len(ImageP.TRANSFORMATION_FUNCTIONS))
    # Walk the tree of operator application sequences, depth-first
    # keeping track of how much of each sequence is "correct",
    # and go a certain number of levels into error territory.
    # At each node of this tree, create a procimage.
    # NOTE -- THE FOLLOWING NEEDS TO CHANGE TO TURN OP INTO A FN.

    # Under construction...
    #op_function_seq = map(lambda op:ImageP.TRANSFORMATION_FUNCTIONS[op], op_sequence)
    #correct_seq = map(lambda op:ImageOps.INVERSE[ImageP.TRANSFORMATION_FUNCTIONS[op]], op_sequence)
    correct_seq=list(op_sequence)
    correct_seq.reverse()
    op_set_indices = range(len(INVERSE_OP_FUNCTIONS))
    
    
    gen_images_for_puzzle(puzzle_image_name_prefix, scrambled_image,
                          op_set_indices, correct_seq, [], misd, 0, images=images)
    #n += 1
    
from ProgressMeter import Meter

def _demo(meter, value):
  meter.set(value)
  if value < 1.0:
      value = value + 0.005
      meter.after(50, lambda: _demo(meter, value))
  else:
      meter.set(value, 'Demo successfully finished')

PROGRESS_BAR = None
def show_meter():
    global PROGRESS_BAR
    #import Tkinter
    #root = Tkinter.Tk(className='meter demo')
    root = TStar.TK_ROOT
    PROGRESS_BAR = Meter(root, relief='ridge', bd=3)
    #PROGRESS_BAR.pack(fill='x')
    PROGRESS_BAR.grid(row=0, column=0)
    PROGRESS_BAR.set(0.0, 'Progress of game export...')
    #PROGRESS_BAR.after(1000, lambda: _demo(PROGRESS_BAR, 0.0))
    PROGRESS_BAR._canv.update_idletasks()
    #root.mainloop()


def gen_images_for_puzzle(starting_im_fn, scrambled_image, op_set_indices, rest_of_correct_seq,
                          seq_so_far, misd, current_error_depth, images=True):
    print "In gen_images_for_puzzle, seq_so_far is "+str(seq_so_far)
    print "  current_error_depth is "+str(current_error_depth)
    print "  rest_of_correct_seq is "+ str(rest_of_correct_seq)

    # Here, starting_im is the scrambled image, in base-image form.
    for op in op_set_indices:
        print "  Extending the current op sequence with op "+str(op)
        get_or_create_puzzle_procimage_incrementally(starting_im_fn, scrambled_image,
                                                     seq_so_far+[op], write=True, images=images)
        if current_error_depth + 1 < misd:
            if len(rest_of_correct_seq)>0 and op==rest_of_correct_seq[0]:
                new_rest = rest_of_correct_seq[1:]
                if current_error_depth==0: new_error_depth=0
            else:
                new_error_depth = current_error_depth+1
                new_rest = []
            gen_images_for_puzzle(starting_im_fn, scrambled_image,
                                  op_set_indices, new_rest, seq_so_far+[op],
                                  misd, new_error_depth+1, images=images)

PROCIMAGES = {}
def get_or_create_puzzle_procimage_incrementally(fn, starting_image, op_sequence, write=False, images=True):
  global PROCIMAGES, FOLDER_NAME, GAMES_FOLDER, PROCIMAGE_XML
  global USE_PROGRESSBAR, PROGRESS_BAR, PROCIMAGE_NUM, NUM_PROCIMAGES
  if op_sequence==[]:
    return starting_image # ImageP.get_baseimage(fn)
  bkey = fn+"--"+flatten(op_sequence) # key used in both hash lookup and file name.
  try:
    if images:
      im = ImageP.PROCIMAGES[bkey].copy()
      return im
    else: raise
  except:
    # Generate the image applying the last op to an existing image.
    try:
      im = get_or_create_puzzle_procimage_incrementally(fn, starting_image, op_sequence[:-1], images=images)
    except:
      print "Could not find, or could not create an image from file with name: "+filename
      return None
    # Now apply the new operator to this image.
    op = op_sequence[-1]
    PROCIMAGE_NUM += 1
    print "PROCIMAGE_NUM is now "+str(PROCIMAGE_NUM)
    if images:
      im = apply(ImageP.INVERSE_FUNCTIONS[op], [im])
      if USE_PROGRESSBAR:
          PROGRESS_BAR.set((0.0 + PROCIMAGE_NUM)/NUM_PROCIMAGES, None)
    if images:
        PROCIMAGES[bkey]=im.copy() # Save for future use
  if write:
      if images:
        im.save(GAMES_FOLDER+"/"+FOLDER_NAME+"/"+bkey+".jpg", "JPEG")
      PROCIMAGE_XML += '<proc_image seq="'+str(tuple(op_sequence))+'" filename="'+bkey+'.jpg"></proc_image>\n'
  if images: return im
  else: return None

def flatten(lst):
    # Make a hyphenated string from the elements in lst."
    return "-".join(map(lambda x:str(x), lst))

